/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019 Harald Foerster		*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/


#include <string.h>

#ifdef HAVE_SEMAPHORE
#include <semaphore.h>
#endif

#include <libswscale/swscale.h>
#include <libavutil/imgutils.h>
#include <libavutil/parseutils.h>

#include "cff_ffmpeg.h"
#include "cff_nls.h"

/*
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <stdint.h>
#include <stdio.h>
#include <libavutil/opt.h>
#include <libavutil/mathematics.h>
#include <libavutil/rational.h>
*/


#ifdef AV_PIX_FMT_FLAG_PSEUDOPAL
#define CFF_PIX_FMT_FLAG_PSEUDOPAL	AV_PIX_FMT_FLAG_PSEUDOPAL
#else
#define CFF_PIX_FMT_FLAG_PSEUDOPAL	0
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffMuxerIterate(ofmt, opaque)	av_muxer_iterate(opaque)
#else
#define CffMuxerIterate(ofmt, opaque)	av_oformat_next(ofmt)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffDemuxerIterate(ifmt, opaque)	av_demuxer_iterate(opaque)
#else
#define CffDemuxerIterate(ifmt, opaque)	av_iformat_next(ifmt)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffCodecIterate(codec, opaque)	av_codec_iterate(opaque)
#else
#define CffCodecIterate(codec, opaque)	av_codec_next(codec)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffBsfIterate(opaque)		av_bsf_iterate(opaque)
#else
#define CffBsfIterate(opaque)		av_bsf_next(opaque)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffRegisterAll()		{;}
#else
#define CffRegisterAll()		av_register_all()
#endif


#define kOpenFileGetFirstFrame	0
#define kOpenFileSeekLastGop	5
#define kOpenFileGetLastFrame	6

#define kUndo			0
#define kRedo			1

/* Prototypes */
static void	cff_ffmpeg_video_undo_clear(cff_undo_t **list);
static int	cff_ffmpeg_video_cut_add_dts(cff_video_t *video, cff_dts_t *src);
static int	cff_ffmpeg_video_cut_remove_dts(cff_video_t *video, cff_dts_t *src);
static AVStream	*cff_ffmpeg_get_stream_from_index(cff_video_t *video, int index);

/* Global */
cff_video_t gVideo;

/* 'opaque' is a pointer for the iteration state */
struct CFFCodecNext
{
	const AVCodec	*codec;
	void		*opaque;
};


/*
	For 'cff_ffmpeg_log_callback' and
	'cff_file_info_callback' (av_dump_format)
*/

#define kCffLogBufferSize	1024

typedef struct
{
	char		*str;
	size_t		max;
	size_t		len;
	int		prev;

	void			*data;
	func_log_callback_t	*func;

} cff_logbuffer_t;

static cff_logbuffer_t *gLogBufferPtr = NULL;

/* Semaphore */

#ifdef HAVE_SEMAPHORE
static sem_t gLogSemaphore;
#define CFF_SEMAPHORE_INIT()	sem_init(&gLogSemaphore, 0, 1)
#define CFF_SEMAPHORE_LOCK()	sem_wait(&gLogSemaphore)
#define CFF_SEMAPHORE_UNLOCK()	sem_post(&gLogSemaphore)
#else
#define CFF_SEMAPHORE_INIT()
#define CFF_SEMAPHORE_LOCK()
#define CFF_SEMAPHORE_UNLOCK()
#endif

/* Loglevel */

typedef struct
{
	const char	*name;
	int		value;

} LoglevelTab;

static const LoglevelTab loglevel_tab[] =
{
	{ "quiet"  , AV_LOG_QUIET   },
	{ "panic"  , AV_LOG_PANIC   },
	{ "fatal"  , AV_LOG_FATAL   },
	{ "error"  , AV_LOG_ERROR   },
	{ "warning", AV_LOG_WARNING },
	{ "info"   , AV_LOG_INFO    },
	{ "verbose", AV_LOG_VERBOSE },
	{ "debug"  , AV_LOG_DEBUG   },
	{ NULL }

}; /* static const LoglevelTab loglevel_tab[] = */

static int gLoglevel;


/* Codecs */

enum
{
	kLIBSTART,
	kLIBAVUTIL = kLIBSTART,
	kLIBAVCODEC,
	kLIBAVFORMAT,
	kLIBSWSCALE,
	kLIBEND
};

typedef struct
{
	const char	*name;
	unsigned int	version_int;
	unsigned int	(*versionFunc)(void);

} AVLibTab;

static const AVLibTab avlib_tab[] =
{
   	{ "avutil",
		LIBAVUTIL_VERSION_INT,
		avutil_version
	},

   	{ "avcodec",
		LIBAVCODEC_VERSION_INT,
		avcodec_version
	},

   	{ "avformat",
		LIBAVFORMAT_VERSION_INT,
		avformat_version
	},

   	{ "swscale",
		LIBSWSCALE_VERSION_INT,
		swscale_version
	}

}; /* static const AVLibTab avlib_tab[] */


static void *cff_ffmpeg_memdup(const void *src, int sz)
{
	void *ptr;

	if (sz <= 0)
	{
		return NULL;
	}

	ptr = malloc(sz);

	if (ptr == NULL)
	{
		return NULL;
	}

	return memcpy(ptr, src, sz);

} /* static void *cff_ffmpeg_memdup(const void*, int) */

static char *cff_ffmpeg_strdup(const char *str, int slen)
{
	int  sz;
	char *ptr;

	if (slen < 0)
	{
		slen = strlen(str);

		/* slen plus '\0' */
		sz = slen + 1;
	}

	else
	{
		sz = slen;

		if (str[slen] != '\n')
		{
			/* slen plus '\0' */
			sz++;
		}
	}

	ptr = cff_ffmpeg_memdup(str, sz);

	if (ptr != NULL)
	{
		ptr[slen] = '\0';
	}

	return ptr;

} /* static char *cff_ffmpeg_strdup(const char*, int) */

static char *cff_ffmpeg_string_truncate(char *str, size_t slen)
{
	if (str != NULL)
	{
		char *ptr;

		/* Include terminating '\0' */
		ptr = realloc(str, slen + 1);

		if (ptr != NULL)
		{
			/* Terminate string */
			ptr[slen] = '\0';

			return ptr;
		}

	} /* if (str != NULL) */

	return str;

} /* static char *cff_ffmpeg_string_truncate(char*, size_t) */

static char *cff_ffmpeg_printf_concat(char *str, size_t *slen,
					size_t *max, const char *fmt, ...)
{
	int	n;
	size_t  bufsz, sz;
	va_list args;

	va_start(args, fmt);
	n = vsnprintf(NULL, 0, fmt, args);
	va_end(args);

	if (n <= 0)
	{
		n = 1024;
	}

	bufsz = *slen + n + 1;
	sz    = *max;

	if (sz == 0)
	{
		sz = bufsz;
	}

	while (bufsz > sz)
	{
		sz <<= 1;
	}

	if (str == NULL || sz > *max)
	{
		str = realloc(str, sz);
	}

	if (str != NULL)
	{
		*max = sz;

		sz -= *slen;

		va_start(args, fmt);
		n  = vsnprintf(&str[*slen], sz, fmt, args);
		va_end(args);

		if (n > 0 && n < sz)
		{
			*slen += n;
		}
	}

	return str;

} /* static char *cff_ffmpeg_printf_concat(char *str, size_t*, size_t*,
							const char*, ...) */


/* FFmpeg => cmdutil.h */

char *cff_ffmpeg_get_libs_info(void)
{
	int		lib, slen;
	const AVLibTab	*avlib;
	char		buf[4096];

	avlib = avlib_tab;
	slen  = 0;
	lib   = kLIBSTART;

	do
	{
		int		n;
		size_t		sz;
		unsigned int	version;

		version = avlib->versionFunc();

		sz = sizeof(buf) - slen;
		n  = snprintf(&buf[slen], sz,
			"lib%-9s %2d.%3d.%2d / %2d.%3d.%2d\n",
			avlib->name,
			avlib->version_int >> 16,		/* major */
			(avlib->version_int >> 8) & 0xff,	/* minor */
			avlib->version_int & 0xff,		/* micro */
			version >> 16,				/* major */
			(version >> 8) & 0xff,			/* minor */
			version & 0xff);			/* micro */

		if (n > 0 && n < sz)
		{
			slen += n;
		}

		avlib++;
	}
	while (++lib < kLIBEND);

	/* remove last '\n' */
	return cff_ffmpeg_strdup(buf, slen - 1);

} /* char *cff_ffmpeg_get_libs_info(void) */


/* Formats */

struct FormatInfo
{
	const char	*name;
	const char	*long_name;
	unsigned int	encode;
	unsigned int	decode;
};

struct FormatInfo *cff_ffmpeg_get_format_entry(struct FormatInfo **buf,
				size_t *num, size_t *max, const char *name)
{
	int			n;
	size_t			sz;
	struct FormatInfo	*ptr;

	ptr = *buf;
	sz  = *max;

	if (*num >= sz)
	{
		sz  = *num + 128;
		ptr = realloc(ptr, sizeof(struct FormatInfo) * sz);

		if (ptr == NULL)
		{
			return NULL;
		}

		*buf = ptr;
		*max = sz;

	} /* if (*num >= sz) */

	n = *num;

	while (--n >= 0)
	{
		int i = strcmp(ptr->name, name);

		if (i == 0)
		{
			return ptr;
		}

		if (i > 0)
		{
			memmove(&ptr[1], ptr,
				(n + 1) * sizeof(struct FormatInfo));
			break;
		}

		ptr++;

	} /* while (--n >= 0) */

	(*num)++;

	return ptr;

} /* struct FormatInfo *cff_ffmpeg_get_format_entry(struct FormatInfo**,
					size_t*, size_t*, const char*) */

char *cff_ffmpeg_get_formats_info(void)
{
	size_t			n, num, max, bufsz, slen;
	const AVInputFormat	*ifmt;
	const AVOutputFormat	*ofmt;
	char			*str;
#ifdef HAVE_FFMPEG_ITERATE
	void			*opaque;
#endif

	struct FormatInfo *ptr, *buf;

	max  = 0;
	num  = 0;
 	n    = 0;
	buf  = NULL;

#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ofmt = NULL;
#endif

	while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL)
	{
		ptr = cff_ffmpeg_get_format_entry(&buf, &num, &max, ofmt->name);

		if (ptr == NULL)
		{
			break;
		}

		if (num > n)
		{
			n = num;

			ptr->name	= ofmt->name;
			ptr->long_name	= ofmt->long_name;
			ptr->encode	= 'E';
			ptr->decode	= '.';
		}

	} /* while ((ofmt = CffMuxerIterate(ofmt, &onext)) != NULL) */


#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ifmt = NULL;
#endif

	while ((ifmt = CffDemuxerIterate(ifmt, &opaque)) != NULL)
	{
		ptr = cff_ffmpeg_get_format_entry(&buf, &num, &max, ifmt->name);

		if (ptr == NULL)
		{
			break;
		}

		ptr->decode = 'D';

		if (num > n)
		{
			n = num;

			ptr->name	= ifmt->name;
			ptr->long_name	= ifmt->long_name;
			ptr->encode	= '.';
		}

	} /* while ((ifmt = CffDemuxerIterate(ofmt, &inext)) != NULL) */


	ptr = buf;

	if (ptr == NULL)
	{
		return NULL;
	}

	bufsz = 4096;
	slen  = 0;
	str   = cff_ffmpeg_printf_concat(NULL, &slen, &bufsz,
			" D. = Demuxing supported\n"
			" .E = Muxing supported\n"
			" --\n");
	if (str == NULL)
	{
		return str;
	}

	do
	{
		char *s;

		s = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
			" %c%c %-20s %s\n",
			(char) ptr->decode,
			(char) ptr->encode,
			ptr->name,
			ptr->long_name != NULL ? ptr->long_name : "");

		if (s == NULL)
		{
			break;
		}

		str = s;
		ptr++;
	}
	while (--num != 0);

	free(buf);

	/* remove last '\n' */
	return cff_ffmpeg_string_truncate(str, slen - 1);

} /* char *cff_ffmpeg_get_formats_info(void) */



/* Codecs */

static char cff_ffmpeg_get_media_type_char(enum AVMediaType type)
{
	switch (type)
	{
		case AVMEDIA_TYPE_VIDEO:	return 'V';
		case AVMEDIA_TYPE_AUDIO:	return 'A';
		case AVMEDIA_TYPE_DATA:		return 'D';
		case AVMEDIA_TYPE_SUBTITLE:	return 'S';
		case AVMEDIA_TYPE_ATTACHMENT:	return 'T';
		default:			break;
	}

	return '?';

} /* static char cff_ffmpeg_get_media_type_char(enum AVMediaType) */

static int cff_ffmpeg_compare_codec_desc(const void *a, const void *b)
{
	const AVCodecDescriptor * const *da = a;
	const AVCodecDescriptor * const *db = b;

	return ((*da)->type != (*db)->type) ?
			(*da)->type - (*db)->type :
				strcmp((*da)->name, (*db)->name);

} /* static int cff_ffmpeg_compare_codec_desc(const void*, const void*) */

static const AVCodecDescriptor **cff_ffmpeg_get_codecs_sorted(void)
{
	unsigned int			nb_codecs, i;
	const AVCodecDescriptor		*desc, **codecs;

	nb_codecs	= 0;
	desc		= NULL;

	while ((desc = avcodec_descriptor_next(desc)) != NULL)
	{
		nb_codecs++;
	}

	if (nb_codecs == 0)
	{
		return NULL;
	}

	codecs = av_calloc(nb_codecs + 1, sizeof(*codecs));

	if (codecs == NULL)
	{
		return NULL;
	}

	codecs[nb_codecs] = NULL;

	i	= 0;
	desc	= NULL;

	while ((desc = avcodec_descriptor_next(desc)) != NULL)
	{
		codecs[i++] = desc;
	}

	qsort(codecs, nb_codecs, sizeof(*codecs), cff_ffmpeg_compare_codec_desc);

	return codecs;

} /* static const AVCodecDescriptor **cff_ffmpeg_get_codecs_sorted(void) */

static int cff_ffmpeg_next_codec_for_id(enum AVCodecID id,
				struct CFFCodecNext *prev, int is_encoder)
{
	while ((prev->codec = CffCodecIterate(prev->codec,
						&(prev->opaque))) != NULL)
	{
		if ((prev->codec)->id == id && ((is_encoder == 0) ?
			av_codec_is_decoder(prev->codec) :
					av_codec_is_encoder(prev->codec)))
		{
			return 0;
		}
	}

	return -1;

} /* static int cff_ffmpeg_next_codec_for_id(enum AVCodecID, struct CFFCodecNext*, int) */

static char *cff_ffmpeg_print_codecs_for_id(char *str, size_t *slen, size_t *bufsz, enum AVCodecID id, int is_encoder)
{
	struct CFFCodecNext codec;

	str = cff_ffmpeg_printf_concat(str, slen, bufsz,
			" (%s: ", is_encoder ? "encoders" : "decoders");
	if (str == NULL)
	{
		return str;
	}

	codec.codec  = NULL;
	codec.opaque = NULL;

	while (cff_ffmpeg_next_codec_for_id(id, &codec, is_encoder) == 0)
	{
		str = cff_ffmpeg_printf_concat(str, slen, bufsz,
						"%s ", codec.codec->name);
		if (str == NULL)
		{
			return str;
		}
	}

	return cff_ffmpeg_printf_concat(str, slen, bufsz, "%c", ')');

} /* static char *cff_ffmpeg_print_codecs_for_id(char*, size_t*, size_t*, enum AVCodecID, int) */

char *cff_ffmpeg_get_codecs_info(void)
{
	size_t			bufsz, slen;
	const AVCodecDescriptor *desc, **codecs, **codecs_orig;
	char			*str;

	bufsz = 4096;
	slen  = 0;
	str   = cff_ffmpeg_printf_concat(NULL, &slen, &bufsz,
			" D..... = Decoding supported\n"
			" .E.... = Encoding supported\n"
			" ..V... = Video codec\n"
			" ..A... = Audio codec\n"
			" ..S... = Subtitle codec\n"
			" ...I.. = Intra frame-only codec\n"
			" ....L. = Lossy compression\n"
			" .....S = Lossless compression\n"
			" -------\n");

	if (str == NULL)
	{
		return str;
	}

	codecs = cff_ffmpeg_get_codecs_sorted();

	if (codecs == NULL)
	{
		return NULL;
	}

	codecs_orig = codecs;

	while ((desc = *(codecs++)) != NULL)
	{
		char			*ptr, decoder, encoder;
		struct CFFCodecNext	codec;

		decoder = encoder = '.';

		if (avcodec_find_decoder(desc->id))
		{
			decoder = 'D';
		}

		if (avcodec_find_encoder(desc->id))
		{
			encoder = 'E';
		}

		if (decoder == '.' && encoder == '.')
		{
			continue;
		}

		ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
			" %c%c%c%c%c%c %-20s %s",
			(char) decoder,
			(char) encoder,
			cff_ffmpeg_get_media_type_char(desc->type),
			(desc->props & AV_CODEC_PROP_INTRA_ONLY)? 'I' : '.',
			(desc->props & AV_CODEC_PROP_LOSSY)	? 'L' : '.',
			(desc->props & AV_CODEC_PROP_LOSSLESS)	? 'S' : '.',
			desc->name,
			desc->long_name ? desc->long_name : "");

		if (ptr == NULL)
		{
			break;
		}

		str = ptr;

		/*
			Print decoders/encoders when there's more than
			one or their names are different from codec name
		*/

		codec.codec  = NULL;
		codec.opaque = NULL;

		while (cff_ffmpeg_next_codec_for_id(desc->id, &codec, 0) == 0)
		{
			if (strcmp(codec.codec->name, desc->name) != 0)
			{
				ptr = cff_ffmpeg_print_codecs_for_id(str, &slen,
								&bufsz, desc->id, 0);
				if (ptr != NULL)
				{
					str = ptr;
				}

				break;
			}
		}


		codec.codec  = NULL;
		codec.opaque = NULL;

		while (cff_ffmpeg_next_codec_for_id(desc->id, &codec, 1) == 0)
		{
			if (strcmp(codec.codec->name, desc->name) != 0)
			{
				ptr = cff_ffmpeg_print_codecs_for_id(str, &slen,
								&bufsz, desc->id, 1);
				if (ptr != NULL)
				{
					str = ptr;
				}

				break;
			}
		}

		ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz, "%c", '\n');

		if (ptr == NULL)
		{
			break;
		}

		str = ptr;

	} /* while ((desc = *(codecs++)) != NULL) */

	av_free(codecs_orig);

	return cff_ffmpeg_string_truncate(str, slen - 1);

} /* char *cff_ffmpeg_get_codecs_info(void) */

static char *cff_ffmpeg_get_bitstream_codecs(char *str, size_t *slen, size_t *bufsz, const enum AVCodecID *id_ptr)
{
	const char		*end;
	enum AVCodecID		codec_id;

	end = "";

	while ((codec_id = *(id_ptr++)) != AV_CODEC_ID_NONE)
	{
		char	*ptr;
		AVCodec	*codec;

		codec = avcodec_find_decoder(codec_id);

		if (codec != NULL)
		{
			if (end[0] == '\0')
			{
				end = "(";
			}

			else if (end[0] == '(')
			{
				end = ", ";
			}

			ptr = cff_ffmpeg_printf_concat(str, slen, bufsz,
						"%s%s", end, codec->name);
			if (ptr == NULL)
			{
				break;
			}

			str = ptr;

		} /* if (codec != NULL) */

	} /* while ((codec_id = *(codec_ptr++)) != AV_CODEC_ID_NONE) */

	if (end[0] == '(')
	{
		end = ")";
	}

	else if (end[0] == ',')
	{
		end = ")";
	}

	return cff_ffmpeg_printf_concat(str, slen, bufsz, "%s\n", end);

} /* static char *cff_ffmpeg_get_bitstream_codecs(char*, size_t*, size_t*, const enum AVCodecID*) */

char *cff_ffmpeg_get_bitstream_filters_info(void)
{
	size_t			bufsz, slen;
	char			*str;
	void			*opaque;
	const AVBitStreamFilter	*bsf;

	opaque = NULL;
	str    = NULL;
	bufsz  = 4096;
	slen   = 0;

	while ((bsf = CffBsfIterate(&opaque)) != NULL)
	{
		char *ptr;

		ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
							"%s ", bsf->name);
		if (ptr == NULL)
		{
			break;
		}

		str = ptr;

		if (bsf->codec_ids == NULL)
		{
			ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
							"%s\n", "(all codecs)");
			if (ptr == NULL)
			{
				break;
			}

			str = ptr;
		}

		else
		{
			ptr = cff_ffmpeg_get_bitstream_codecs(str, &slen, &bufsz, bsf->codec_ids);

			if (ptr == NULL)
			{
				break;
			}

			str = ptr;

		} /* if (id_ptr == NULL) .. else */

	} /* while ((bsf = CffBsfIterate(&opaque)) != NULL) */

	/* remove last '\n' */
	return cff_ffmpeg_string_truncate(str, slen - 1);

} /* char *cff_ffmpeg_get_bitstream_filters_info(void) */

cff_fmtlist_t *cff_ffmpeg_get_format_names(cff_video_t *video)
{
	int			i, n, nb_codecs;
	unsigned int		num;
	const AVOutputFormat	*ofmt;
	cff_fmtlist_t		buf[512];
	const AVCodecDescriptor	*codec_descriptors[128], *codecs_found[128];
#ifdef HAVE_FFMPEG_ITERATE
	void			*opaque;
#endif

	if (video == NULL || video->inFormatCtx == NULL ||
					video->nb_streams == 0)
	{
		return NULL;
	}

	i = 0;
	nb_codecs = 0;

	do
	{
		AVStream *is = cff_ffmpeg_get_stream_from_index(video, i);

		if (is == NULL)
		{
			continue;
		}

		codec_descriptors[nb_codecs] =
			avcodec_descriptor_get((is->codecpar)->codec_id);

		if (codec_descriptors[nb_codecs] != NULL)
		{
			const AVCodecDescriptor	**d;

			d = codec_descriptors;
			n = nb_codecs;

			while (--n >= 0)
			{
				if (*(d++) == codec_descriptors[nb_codecs])
				{
					break;
				}
			}

			if (n < 0)
			{
				nb_codecs++;

				if (nb_codecs >= (sizeof(codec_descriptors) /
						sizeof(AVCodecDescriptor*)) - 1)
				{
					break;
				}
			}

		} /* if (codec_descriptors[nb_codecs] != NULL) */
	}
	while (++i < video->nb_streams);

	if (nb_codecs == 0)
	{
		return NULL;
	}

	codec_descriptors[nb_codecs] = NULL;


	num = 0;

#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ofmt = NULL;
#endif

	while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL)
	{
		const AVCodecDescriptor	*desc, **desc_ptr, **found;
		cff_fmtlist_t		*ptr;

		desc_ptr = codec_descriptors;
		found    = codecs_found;

		while ((desc = *(desc_ptr++)) != NULL)
		{
			if (av_codec_get_tag(ofmt->codec_tag, desc->id) != 0)
			{
				*(found++) = desc;
			}

			else if (ofmt->video_codec == desc->id ||
					ofmt->audio_codec == desc->id ||
						ofmt->subtitle_codec == desc->id)
			{
				*(found++) = desc;
			}

			else if (strcmp(ofmt->name, "mpegts") == 0 &&
				(ofmt->name == (video->inFormatCtx->iformat)->name ||
				strcmp(ofmt->name, (video->inFormatCtx->iformat)->name) == 0))
			{
				/* 'mpegts' does not show all codecs (e.g. 'hevc') */
				*(found++) = desc;
			}

		} /* while ((desc = *(desc_ptr++)) != NULL) */

		if (found == codecs_found)
		{
			continue;
		}

		*found	= NULL;
		ptr	= buf;
		n	= num;
		i	= -1;

		while (--n >= 0)
		{
			i = strcmp(ptr->name, ofmt->name);

			if (i == 0)
			{
				break;
			}

			if (i > 0)
			{
				memmove(&ptr[1], ptr,
					(n + 1) * sizeof(cff_fmtlist_t));
				break;
			}

			ptr++;

		} /* while (--n >= 0) */

		if (i != 0)
		{
			const char		*s1, *s2;
			size_t			slen;
			const AVCodecDescriptor	*d;
			char			sbuf[128];

			s1 = s2 = "";

			found	= codecs_found;
			sbuf[0]	= '(';
			slen	= sizeof(char);

			while ((d = *(found++)) != NULL)
			{
				const char end[] = ")";

				if (*found == NULL)
				{
					s2 = end;
				}

				n = snprintf(&sbuf[slen], sizeof(buf) - slen,
							"%s%s%s", s1, d->name, s2);
				if (n > 0)
				{
					slen += n;

					if (slen >= sizeof(sbuf))
					{
						slen = sizeof(sbuf) -
								sizeof(char);
						break;
					}

					if (s2 != end)
					{
						s1 = ", ";
					}
				}

			} /* while ((d = *(found++)) != NULL) */


			ptr->name   = ofmt->name;
			ptr->codecs = NULL;

			if (slen > sizeof(char))
			{
				ptr->codecs = cff_ffmpeg_memdup(sbuf,
							slen + sizeof(char));
			}

			num++;

			if (num >= (sizeof(buf) / sizeof(cff_fmtlist_t)) - 1)
			{
				break;
			}

		} /* if (i != 0) */

	} /* while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL) */

	if (num == 0)
	{
		return NULL;
	}

	buf[num].name   = NULL;
	buf[num].codecs = NULL;

	return cff_ffmpeg_memdup(buf, sizeof(cff_fmtlist_t) * (num + 1));

} /* cff_fmtlist_t *cff_ffmpeg_get_format_names(cff_video_t*) */

cff_bsflist_t *cff_ffmpeg_get_bitstream_filter_names(cff_video_t *video, int media_type)
{
	int			i, nb_codecs, has_media;
	unsigned int		num;
	void			*opaque;
	const AVBitStreamFilter	*bsf;
	AVCodec			*codec;
	cff_bsflist_t		*ptr, buf[512];
	enum AVCodecID		codec_ids[128];

	if (video == NULL || video->inFormatCtx == NULL ||
					video->nb_streams == 0)
	{
		return NULL;
	}

	i = 0;
	nb_codecs = 0;
	has_media = 0;

	do
	{
		AVStream		*is;
		AVCodecParameters	*codecpar;
		
		is = cff_ffmpeg_get_stream_from_index(video, i);

		if (is == NULL)
		{
			continue;
		}

		codecpar = is->codecpar;

		if (codecpar->codec_type == AVMEDIA_TYPE_VIDEO)
		{
			if (media_type != kCFFMediaVideo)
			{
				continue;
			}
		}

		else if (codecpar->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			if (media_type != kCFFMediaAudio)
			{
				continue;
			}
		}

		else
		{
			if (media_type != kCFFMediaOther)
			{
				continue;
			}
		}

		has_media = 1;
		codec = avcodec_find_decoder(codecpar->codec_id);

		if (codec != NULL)
		{
			codec_ids[nb_codecs] = codecpar->codec_id;
			nb_codecs++;

			if (nb_codecs >= (sizeof(codec_ids) /
						sizeof(enum AVCodecID)) - 1)
			{
				break;
			}

		} /* if (codec != NULL) */
	}
	while (++i < video->nb_streams);

	if (has_media == 0)
	{
		return NULL;
	}

	codec_ids[nb_codecs] = AV_CODEC_ID_NONE;


	num	= 0;
	ptr	= buf;
	opaque	= NULL;

	while ((bsf = CffBsfIterate(&opaque)) != NULL)
	{
		const char		*name;
		const enum AVCodecID	*id_ptr;

		name		= NULL;
		ptr->codecs	= NULL;
		id_ptr		= bsf->codec_ids;

		if (id_ptr == NULL)
		{
			/* All codecs */
			name = bsf->name;
		}

		else
		{
			enum AVCodecID codec_id;

			while ((codec_id = *(id_ptr++)) != AV_CODEC_ID_NONE)
			{
				enum AVCodecID		cid;
				const enum AVCodecID	*cid_ptr;

				cid_ptr = codec_ids;

				while ((cid = *(cid_ptr++)) != AV_CODEC_ID_NONE)
				{
					if (cid == codec_id)
					{
						break;
					}
				}

				if (cid != AV_CODEC_ID_NONE)
				{
					break;
				}

			} /* while ((codec_id = *(codec_ptr++)) != AV_CODEC_ID_NONE) */

			if (codec_id != AV_CODEC_ID_NONE)
			{
				char	*str;
				size_t	bufsz, slen;

				slen  = 0;
				bufsz = 1024;
				str   = cff_ffmpeg_get_bitstream_codecs(NULL, &slen, &bufsz, bsf->codec_ids);

				/* remove last '\n' */
				ptr->codecs = cff_ffmpeg_string_truncate(str, slen - 1);

				name = bsf->name;
			}

		} /* if (id_ptr == NULL) .. else */

		if (name != NULL)
		{
			ptr->name = name;

			ptr++;
			num++;

			if (num >= (sizeof(buf) / sizeof(cff_bsflist_t)) - 1)
			{
				break;
			}
		}

	} /* while ((bsf = CffBsfIterate(&opaque)) != NULL) */

	if (num == 0)
	{
		return NULL;
	}

	ptr->name   = NULL;
	ptr->codecs = NULL;

	return cff_ffmpeg_memdup(buf, sizeof(cff_bsflist_t) * (num + 1));

} /* cff_bsflist_t *cff_ffmpeg_get_bitstream_filter_names(cff_video_t*, int) */

int cff_ffmpeg_get_image_format_names(char ***long_name, char ***format_name)
{
	int			num;
	const AVOutputFormat	*ofmt;
	const char		*namebuf[512], *formatbuf[512];
#ifdef HAVE_FFMPEG_ITERATE
	void			*opaque;
#endif

	num = 0;

#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ofmt = NULL;
#endif

	while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL)
	{
		if (ofmt->mime_type != NULL &&
			strncmp(ofmt->mime_type, "image/", 6) == 0)
		{
			enum AVCodecID codec_id = ofmt->video_codec;

			if (codec_id != AV_CODEC_ID_NONE)
			{
				AVCodec	*codec = avcodec_find_encoder(codec_id);

				if (codec != NULL)
				{
					namebuf[num] = ofmt->long_name;

					if (namebuf[num] == NULL)
					{
						namebuf[num] = ofmt->name;
					}

					formatbuf[num] = ofmt->name;

					num++;

					if (num >= 510)
					{
						break;
					}
				}

			} /* if (codec_id != AV_CODEC_ID_NONE) */

		} /* if (ofmt->mime_type != NULL &&
			strncmp(ofmt->mime_type, "image/", 6) == 0) */

	} /* while ((ofmt = CffMuxerIterate(ofmt, &next)) != NULL) */

	if (num == 0)
	{
		if (long_name != NULL)
		{
			*long_name = NULL;
		}

		if (format_name != NULL)
		{
			*format_name = NULL;
		}

		return 0;
	}

	namebuf[num]   = NULL;
	formatbuf[num] = NULL;
	num++;

	if (long_name != NULL)
	{
		*long_name = cff_ffmpeg_memdup(namebuf, sizeof(char*) * num);
	}

	if (format_name != NULL)
	{
		*format_name = cff_ffmpeg_memdup(formatbuf, sizeof(char*) * num);
	}

	return num - 1;

} /* int cff_ffmpeg_get_image_format_names(char***, char***) */

char* cff_ffmpeg_get_outformat_extension(const char *format_name, int *slen)
{
	unsigned int	n;
	char		*str;
	AVOutputFormat	*format;

	if (format_name == NULL)
	{
		return NULL;
	}

	format = av_guess_format(format_name, NULL, NULL);

	if (format == NULL)
	{
		return NULL;
	}

	if (format->extensions == NULL)
	{
		return NULL;
	}

	if ((n = strcspn(format->extensions, ",")) == 0)
	{
		return NULL;
	}

	/* Plus '.' and '\0' */
	str = malloc(n + 2);

	if (str != NULL)
	{
		if (slen != NULL)
		{
			*slen = n + 1;
		}

		memcpy(&str[1], format->extensions, n);
		str[0]     = '.';
		str[n + 1] = '\0';
	}

	return str;

} /* char *cff_ffmpeg_get_outformat_extension(const char*, int*) */

static void cff_ffmpeg_dts2time(uint64_t dts, AVStream *is, unsigned int buf[4])
{
	unsigned int hour, min, sec;

	dts = av_rescale_q(dts, is->time_base, AV_TIME_BASE_Q);

	sec  = dts / AV_TIME_BASE;
	min  = sec / 60;
	hour = min / 60;

	buf[0] = hour;
	buf[1] = min % 60;
	buf[2] = sec % 60;
	buf[3] = (dts % AV_TIME_BASE) / 1000;

	if (buf[3] < 999 && (dts % AV_TIME_BASE) >= AV_TIME_BASE / 2)
	{
		buf[3]++;
	}

} /* static void cff_ffmpeg_dts2time(uint64_t, unsigned int[4]) */

static int64_t cff_ffmpeg_sec2dts(int sec, cff_video_t *video)
{
	int64_t		dts, dts_mod;
	AVStream	*is;

	is = (video->inFormatCtx->streams[video->vframe.stream_index]);

	dts =  av_rescale_q((int64_t) sec * video->vframe.frame_duration,
			(AVRational) {1, video->vframe.frame_duration}, is->time_base);

	dts_mod = dts % video->vframe.frame_duration;

	if (dts_mod == 0)
	{
		return dts;
	}

	if (sec > 0)
	{
		/* round up */
		return dts + video->vframe.frame_duration - dts_mod;
	}

	/* round down */
	return dts - video->vframe.frame_duration - dts_mod;

} /* static int64_t cff_ffmpeg_sec2dts(int, cff_video_t*) */

static const char *cff_ffmpeg_get_pictype_str(const AVFrame *frame)
{
	static const char *picture_type[] =
		{ "?", "I", "P", "B", "S", "SI", "SP", "BI" };

	static char pictype_buf[5];

	int key_frame, pict_type;

	key_frame = 0;
	pict_type = 0;

	if (frame != NULL)
	{
		if (frame->key_frame)
		{
			strcpy(pictype_buf, "K/");
			key_frame = 2;
		}

		if (frame->pict_type <= AV_PICTURE_TYPE_BI)
		{
			pict_type = frame->pict_type;
		}
	}

	strcpy(&pictype_buf[key_frame], picture_type[pict_type]);

	return pictype_buf;

} /* static const char *cff_ffmpeg_get_pictype_str(const AVFrame*) */

int cff_ffmpeg_get_current_loglevel_index(void)
{
	if (gLogBufferPtr != NULL)
	{
		int			loglevel;
		const LoglevelTab	*ptr;

		loglevel = av_log_get_level();

		ptr = loglevel_tab;

		do
		{
			if (loglevel == ptr->value)
			{
				return ptr - loglevel_tab;
			}

			ptr++;
		}
		while (ptr->name != NULL);

	} /* if (gLogBufferPtr != NULL) */

	return -1;

} /* int cff_ffmpeg_get_current_loglevel_index(void) */

static const LoglevelTab *cff_ffmpeg_get_loglevel_ptr(const char *logstr)
{
	if (logstr != NULL)
	{
		const LoglevelTab *p = loglevel_tab;

		do
		{
			if (strcmp(logstr, p->name) == 0)
			{
				return p;
			}

			p++;
		}
		while (p->name != NULL);

	} /* if (logstr != NULL) */

	return NULL;

} /* static const LoglevelTab *cff_ffmpeg_get_loglevel_ptr(const char*) */

const char *cff_ffmpeg_get_loglevel_name(int logindex)
{
	if (logindex < 0 ||
		logindex >= sizeof(loglevel_tab) / sizeof(LoglevelTab))
	{
		return NULL;
	}

	return loglevel_tab[logindex].name;

} /* const char *cff_ffmpeg_get_loglevel_name(int) */

static int cff_ffmpeg_log_callback_print(cff_logbuffer_t *logbuf)
{
	if ((*logbuf->func)(logbuf->str, logbuf->len, logbuf->data) != 0)
	{
		/* Restore previous level */
		cff_ffmpeg_set_loglevel(NULL, 0, NULL, NULL);
		return -1;
	}

	return 0;

} /* static int cff_ffmpeg_log_callback_print(cff_logbuffer_t*) */

static void cff_ffmpeg_log_callback_flush(void)
{
	cff_logbuffer_t *logbuf = gLogBufferPtr;

	if (logbuf != NULL)
	{
		int retv = 0;

		if (logbuf->func != NULL)
		{
			if (logbuf->len > 0 && logbuf->prev > 0)
			{
				if (logbuf->prev > 1)
				{
					size_t	len;
					char	*str;

					len = 0;
					str = cff_ffmpeg_printf_concat(logbuf->str,
						&len,
						&(logbuf->max),
						"Last message repeated %d times\n",
						logbuf->prev);

					if (str != NULL)
					{
						logbuf->str = str;
						logbuf->len = len;
					}
				}

				retv = cff_ffmpeg_log_callback_print(logbuf);

			} /* if (logbuf->len > 0 && logbuf->prev > 0) */

		} /* if (logbuf->func != NULL) */

		if (retv == 0)
		{
			logbuf->len  = 0;
			logbuf->prev = 0;
		}

	} /* if (logbuf != NULL) */

} /* static void cff_ffmpeg_log_callback_flush(void) */

static void cff_ffmpeg_log_callback(void *ptr, int level, const char *fmt, va_list vl)
{
	/* FFmpeg => libavutil/log.c */

	int	n, len, sz;
	AVClass	*avc;
	char	*str, buf[kCffLogBufferSize];

	if (level > av_log_get_level())
	{
		return;
	}

	len = 0;
	avc = (ptr == NULL) ? NULL : *(AVClass **) ptr;

	if (avc != NULL)
	{
		static const char null_str[] = "NULL";
		static const char avcfmt[]   = "[%s @ %p] ";

		const char *item_name;

		if (avc->parent_log_context_offset > 0 &&
					avc->category != AV_CLASS_CATEGORY_NA)
		{
			AVClass **parent = *(AVClass ***)
				(((uint8_t *) avc) + avc->parent_log_context_offset);

			if (parent != NULL && *parent != NULL)
			{
				item_name = (*(*parent)->item_name)(parent);

				if (item_name == NULL)
				{
					item_name = null_str;
				}

				sz = snprintf(buf, sizeof(buf),
						avcfmt, item_name, parent);
				if (sz > 0)
				{
					if (sz >= sizeof(buf))
					{
						sz = sizeof(buf) - sizeof(char);
					}

					len = sz;
				}
			}

		} /* if (avc->parent_log_context_offset > 0 &&
					avc->category != AV_CLASS_CATEGORY_NA) */

		item_name = (*avc->item_name)(ptr);

		if (item_name == NULL)
		{
			item_name = null_str;
		}

		sz = sizeof(buf) - len;
		n  = snprintf(&buf[len], sz, avcfmt, item_name, ptr);

		if (n > 0)
		{
			if (n >= sz)
			{
				n = sz - sizeof(char);
			}

			len += n;
		}

	} /* if (avc != NULL) */

	sz = sizeof(buf) - len;
	n  = vsnprintf(&buf[len], sz, fmt, vl);

	if (n > 0)
	{
		if (n >= sz)
		{
			n = sz - sizeof(char);
		}

		len += n;
	}

	if (len > 0)
	{
		cff_logbuffer_t *logbuf;

		CFF_SEMAPHORE_LOCK();

		logbuf = gLogBufferPtr;

		if (logbuf == NULL)
		{
			CFF_SEMAPHORE_UNLOCK();
			return;
		}

		if (logbuf->func != NULL && logbuf->str != NULL)
		{
			if (len == logbuf->len &&
				strcmp(buf, logbuf->str) == 0)
			{
				logbuf->prev++;
				CFF_SEMAPHORE_UNLOCK();
				return;
			}

			cff_ffmpeg_log_callback_flush();

			if (logbuf != gLogBufferPtr)
			{
				CFF_SEMAPHORE_UNLOCK();
				return;
			}

		} /* if (logbuf->func != NULL && logbuf->str != NULL) */

		str = cff_ffmpeg_printf_concat(logbuf->str,
					&(logbuf->len),
					&(logbuf->max),
					"%s",
					buf);
		if (str != NULL)
		{
			logbuf->str = str;

			if (logbuf->func != NULL)
			{
				cff_ffmpeg_log_callback_print(logbuf);
			}
		}

		CFF_SEMAPHORE_UNLOCK();

	} /* if (len > 0) */

} /* static void cff_ffmpeg_log_callback(void*, int, const char*, va_list) */

int cff_ffmpeg_set_loglevel(const char *logname, int lock, func_log_callback_t func, void *user_data)
{
	cff_logbuffer_t		*logbuf;
	const LoglevelTab	*ptr;

#ifdef HAVE_SEMAPHORE
	if (lock)
	{
		CFF_SEMAPHORE_LOCK();
	}
#endif
	if (func == NULL ||
		(ptr = cff_ffmpeg_get_loglevel_ptr(logname)) == NULL ||
						ptr->value == AV_LOG_QUIET)
	{
		if ((logbuf = gLogBufferPtr) != NULL)
		{
			if (logbuf->str != NULL)
			{
				free(logbuf->str);
			}

			free(logbuf);
			gLogBufferPtr = NULL;
		}

		av_log_set_level(gLoglevel);
		av_log_set_callback(av_log_default_callback);

#ifdef HAVE_SEMAPHORE
		if (lock)
		{
			CFF_SEMAPHORE_UNLOCK();
		}
#endif
		return -1;
	}

	if ((logbuf = gLogBufferPtr) == NULL)
	{
		logbuf = malloc(sizeof(cff_logbuffer_t));

		if (logbuf == NULL)
		{
#ifdef HAVE_SEMAPHORE
			if (lock)
			{
				CFF_SEMAPHORE_UNLOCK();
			}
#endif
			return -1;
		}

		gLogBufferPtr = memset(logbuf, '\0', sizeof(cff_logbuffer_t));
		logbuf->max   = 512;
	}

	logbuf->data = user_data;
	logbuf->func = func;

	av_log_set_level(ptr->value);
	av_log_set_callback(cff_ffmpeg_log_callback);

#ifdef HAVE_SEMAPHORE
	if (lock)
	{
		CFF_SEMAPHORE_UNLOCK();
	}
#endif
	return 0;

} /* int cff_ffmpeg_set_loglevel(const char*, int, func_log_callback_t, void*) */

int cff_ffmpeg_init(const char *logname)
{
	int loglevel;

	if (logname == NULL)
	{
		loglevel = AV_LOG_QUIET;
	}

	else
	{
		const LoglevelTab *ptr;

		ptr = cff_ffmpeg_get_loglevel_ptr(logname);

		if (ptr == NULL)
		{
			return -1;
		}

		loglevel = ptr->value;
	}

	gLoglevel = loglevel;

	memset(&gVideo, '\0', sizeof(cff_video_t));
	CffRegisterAll();
	av_log_set_level(loglevel);
	gVideo.vframe.pict_type = cff_ffmpeg_get_pictype_str(NULL);

	CFF_SEMAPHORE_INIT();

	return 0;

} /* int cff_ffmpeg_init(const char *loglevel) */

static enum AVCodecID cff_ffmpeg_is_video_stream(cff_video_t *video, int index)
{
	enum AVCodecID	codec_id;
	AVStream	*is;

	codec_id = AV_CODEC_ID_NONE;

	is = (video->inFormatCtx)->streams[index];

	if ((is->codecpar)->codec_type == AVMEDIA_TYPE_VIDEO)
	{
		video->vframe.stream_index = index;
		codec_id = (is->codecpar)->codec_id;
	}

	return codec_id;

} /* static enum AVCodecID cff_ffmpeg_is_video_stream(cff_video_t *video, int index) */

static enum AVCodecID cff_ffmpeg_get_program_video_stream(cff_video_t *video, int index)
{
	if (index < (video->inFormatCtx)->nb_programs &&
			(video->inFormatCtx)->programs != NULL)
	{
		unsigned int		nb_idx, *stream_index;
		const AVProgram		*prg;

		prg = (video->inFormatCtx)->programs[index];

		if (prg == NULL)
		{
			return AV_CODEC_ID_NONE;
		}

		stream_index = prg->stream_index;

		if (stream_index == NULL)
		{
			return AV_CODEC_ID_NONE;
		}

		nb_idx = prg->nb_stream_indexes;

		while (--nb_idx >= 0)
		{
			enum AVCodecID codec_id =
				cff_ffmpeg_is_video_stream(video,
						stream_index[nb_idx]);

			if (codec_id != AV_CODEC_ID_NONE)
			{
				video->nb_streams = prg->nb_stream_indexes;
				return codec_id;
			}
		}

	} /* if (index < (video->inFormatCtx)->nb_programs &&
				(video->inFormatCtx)->programs != NULL) */

	return AV_CODEC_ID_NONE;

} /* static enum AVCodecID cff_ffmpeg_get_program_video_stream(cff_video_t*, int) */

static int cff_ffmpeg_read_packet(cff_video_t *video, AVPacket *packet, const char **error)
{
	int retv;

	retv = av_read_frame(video->inFormatCtx, packet);

	if (retv < 0)
	{
		AVIOContext *pb = (video->inFormatCtx)->pb;

		if (avio_feof(pb) != 0 ||
			avio_tell(pb) >= video->vframe.io_seek_end)
		{
			*error = NULL;
		}

		else
		{
			*error = gettext_nop("'av_read_frame()' failed!");
		}

	} /* if (retv < 0) */

	return retv;

} /* static int cff_ffmpeg_read_packet(cff_video_t*, AVPacket*, const char**) */

static const char *cff_ffmpeg_get_frame_dts(cff_video_t *video)
{
	/*
		Get first or last frame dts,
		min and max and gop_size,
		and max frame duration
	*/

	int		gop_size, retv;
	unsigned int	i;
	const char	*str;
	AVFrame		*frame;
	AVPacket	*pktptr, packet;
	int64_t		dts_next, dts_drain[20];

	str      = NULL;
	i        = 0;
	gop_size = 0;
	dts_next = AV_NOPTS_VALUE;
	frame    = video->vframe.avframe;
	pktptr   = &packet;

	av_init_packet(pktptr);

	if (video->vframe.frame_width > kOpenFileGetFirstFrame)
	{
		gop_size++;
	}

	do
	{
		if (pktptr != NULL)
		{
			retv = cff_ffmpeg_read_packet(video, pktptr, &str);

			if (retv < 0)
			{
				/* EOF or Error */

				if (str == NULL)
				{
					/* EOF */
					pktptr = NULL;
				}

				else
				{
					if (video->vframe.frame_width != kOpenFileGetFirstFrame)
					{
						str = NULL;
					}

					break;
				}
			}

			else if (packet.stream_index != video->vframe.stream_index)
			{
				/* Another stream */
				continue;
			}

		} /* if (pktptr != NULL) */

		if (pktptr != NULL || retv != 0)
		{
			retv = avcodec_send_packet(video->vframe.ic, pktptr);

			if (retv != 0)
			{
				/* Failed */
				if (retv == AVERROR(EAGAIN))
				{
					retv = 0;
				}

				if (video->vframe.frame_duration == 0 &&
					pktptr != NULL &&
					! (pktptr->flags & AV_PKT_FLAG_KEY))
				{
					/* First frame(s) is no keyframe */
					retv = 0;
				}

				continue;
			}

		} /* if (pktptr != NULL || retv != 0) */

		retv = avcodec_receive_frame(video->vframe.ic, frame);

		if (retv != 0)
		{
			/* Failed */
			if (retv == AVERROR(EAGAIN))
			{
				retv = 0;
			}

			continue;
		}

		if (video->vframe.frame_height == 0)
		{
			int64_t frame_duration;

			frame_duration = av_rescale_q_rnd(video->vframe.ic->ticks_per_frame,
				(video->vframe.ic)->time_base,
				(video->inFormatCtx)->streams[video->vframe.stream_index]->time_base,
				AV_ROUND_DOWN);

			if (frame_duration > 0)
			{
				video->vframe.frame_duration = frame_duration;
				video->vframe.frame_height   = 1;
			}

			else
			{
				frame_duration = frame->pkt_duration;

				if (frame_duration == 0 && dts_next != AV_NOPTS_VALUE)
				{
					frame_duration = frame->pkt_dts - dts_next;
				}

				if (frame_duration > video->vframe.frame_duration)
				{
					/* Duration (may differ from frame to frame */
					video->vframe.frame_duration = frame_duration;
				}
			}

		} /* if (video->vframe.frame_height == 0) */

		if (video->vframe.frame_width < kOpenFileGetLastFrame)
		{
			if (frame->pkt_dts != AV_NOPTS_VALUE)
			{
				dts_next = frame->pkt_dts + frame->pkt_duration;

				if (video->vframe.dts_start == AV_NOPTS_VALUE)
				{
					/* First frame */
					video->vframe.dts_start = frame->pkt_dts;
				}
			}

			if (frame->key_frame != 0)
			{
				if (gop_size > 0)
				{
					if (video->vframe.min_gop_size == 0 ||
							video->vframe.min_gop_size > gop_size)
					{
						/* Min gop size */
						video->vframe.min_gop_size = gop_size;
					}

					if (gop_size > video->vframe.max_gop_size)
					{
						/* Max gop size */
						video->vframe.max_gop_size = gop_size;
					}

					if (pktptr == NULL)
					{
						goto _last_frame;
					}

					/* Break */
					retv = -1;
					continue;
				}

				else
				{
					gop_size++;
				}
			}

			else if (gop_size > 0)
			{
				gop_size++;
			}

			if (pktptr == NULL)
			{
				goto _last_frame;
			}
		}

		else
		{
_last_frame:
			if (frame->pkt_dts != AV_NOPTS_VALUE)
			{
				/* Last frame */
				video->vframe.dts_stop = frame->pkt_dts;
				dts_next = frame->pkt_dts + frame->pkt_duration;
			}

			else if (pktptr == NULL && dts_next != AV_NOPTS_VALUE && i < 20)
			{
				/* Last frame */
				video->vframe.dts_stop = dts_next;
				dts_drain[i] = dts_next;
				i++;
				dts_next += frame->pkt_duration;
			}

		} /* if (video->vframe.frame_width < kOpenFileGetLastFrame) .. else */
	}
	while (av_packet_unref(&packet), retv == 0);

	if (i > 0)
	{
		video->vframe.drain = malloc(sizeof(cff_drain_t) +
						((i - 1) * sizeof(int64_t)));
		if (video->vframe.drain != NULL)
		{
			(video->vframe.drain)->num = i;
			(video->vframe.drain)->idx = 0;
			memcpy((video->vframe.drain)->pkt_dts, dts_drain,
							i * sizeof(int64_t));
		}
	}

	return str;

} /* static const char *cff_ffmpeg_get_frame_dts(cff_video_t*) */

int cff_ffmpeg_file_open_cb(cff_video_t *video, double *progress, const char **error)
{
	const char *str;

	if (video->vframe.frame_width == kOpenFileSeekLastGop)
	{
		int64_t dts_seek = video->vframe.dts_stop;

		for (;;)
		{
			int retv;

			dts_seek -= video->vframe.max_gop_size * video->vframe.frame_duration;

			retv = av_seek_frame(video->inFormatCtx,
					video->vframe.stream_index, dts_seek, 0);
			if (retv == 0)
			{
				str = NULL;
				break;
			}

			if (dts_seek <= video->vframe.dts_start)
			{
				str = gettext_nop("Cannot seek to last frame!");
				break;
			}

		} /* for (;;) */
	}

	else
	{
		/* Get first or last frame */

		int64_t dts_stop = video->vframe.dts_stop;

		str = cff_ffmpeg_get_frame_dts(video);

		if (str == NULL)
		{
			static const char cannot_get_frame_duration_str[] =
				gettext_nop("Cannot get the frame duration!");

			if (video->vframe.frame_duration == 0)
			{
				str = cannot_get_frame_duration_str;
			}

			else if (video->vframe.frame_width == kOpenFileSeekLastGop - 1)
			{
				if (video->vframe.min_gop_size == 0 ||
						video->vframe.max_gop_size == 0)
				{
					str = gettext_nop("Cannot get the gop size!");
				}
			}

			else if (video->vframe.frame_width == kOpenFileGetLastFrame)
			{
				if (video->vframe.frame_duration == 0)
				{
					str = cannot_get_frame_duration_str;
				}
			}
		}

		else if (video->vframe.frame_width == kOpenFileGetLastFrame)
		{
			if (dts_stop != video->vframe.dts_stop)
			{
				/* Not EOF, but 'dts_stop' has changed */
				str = NULL;
			}

		} /* if (str == NULL) .. else */

	} /* if (video->vframe.frame_width == kOpenFileSeekLastGop) .. else */

	if (error != NULL)
	{
		*error = str;
	}

	if (str == NULL)
	{
		if (video->vframe.frame_width < kOpenFileGetLastFrame)
		{
			video->vframe.frame_width++;

			if (progress != NULL)
			{
				*progress = (double) video->vframe.frame_width /
							(double) kOpenFileGetLastFrame;
			}

			return 0;
		}
	}

	else
	{
		/* Failure */
		avcodec_free_context(&(video->vframe.ic));

	} /* if (str == NULL) .. else */

	cff_ffmpeg_log_callback_flush();

	return 1;

} /* int cff_ffmpeg_file_open_cb(cff_video_t*, double*, const char**) */

static const char *cff_ffmpeg_find_codec(cff_video_t *video,
				enum AVCodecID codec_id,
				func_cff_callback_t func, void *user_data)
{
	int		retv;
	const char	*str;
	AVCodec         *codec;
	AVCodecContext	*ic;
	AVStream	*is;
	AVDictionary	*opts;

	codec = avcodec_find_decoder(codec_id);

	if (codec == NULL)
	{
		return gettext_nop("No video codec found!");
	}

	ic = avcodec_alloc_context3(NULL);
	video->vframe.ic = ic;

	if (ic == NULL)
	{
		return gettext_nop("Couldn't alloc codec context!");
	}

	ic->flags = AV_CODEC_CAP_DR1 | AV_CODEC_CAP_DELAY;

	is = (video->inFormatCtx)->streams[video->vframe.stream_index];

	avcodec_parameters_to_context(ic, is->codecpar);
	ic->pkt_timebase = is->time_base;

	opts = NULL;
	av_dict_set(&opts, "threads", "auto", 0);
	av_dict_set(&opts, "refcounted_frames", "1", 0);
	av_dict_set(&opts, "flags", "low_delay", 0);

	retv = avcodec_open2(ic, codec, &opts);

	av_dict_free(&opts);

	if (retv < 0)
	{
		avcodec_free_context(&(video->vframe.ic));
		return gettext_nop("Couldn't open codec!");
	}

	video->vframe.avframe = av_frame_alloc();

	if (video->vframe.avframe == NULL)
	{
		avcodec_free_context(&(video->vframe.ic));
		return gettext_nop("Couldn't alloc a frame!");
	}


	/* Gather some informations */

	if (is->start_time == AV_NOPTS_VALUE || is->duration == AV_NOPTS_VALUE)
	{
		video->vframe.stream_duration =
			av_rescale_q((video->inFormatCtx)->duration,
					AV_TIME_BASE_Q, is->time_base);
		video->vframe.dts_start =
			av_rescale_q((video->inFormatCtx)->start_time,
					AV_TIME_BASE_Q, is->time_base);
	}

	else
	{
		video->vframe.stream_duration  = is->duration;
		video->vframe.dts_start = is->start_time;
	}

	video->vframe.dts_stop = video->vframe.dts_start + video->vframe.stream_duration;

	video->vframe.io_seek_start = avio_tell((video->inFormatCtx)->pb);
	video->fsize = avio_size((video->inFormatCtx)->pb);

	is->discard = AVDISCARD_DEFAULT;

	/* Get first and last frame, gop_size and frame duration */

	video->vframe.frame_width = kOpenFileGetFirstFrame;
	video->vframe.dts_start   = AV_NOPTS_VALUE;

	if (func == NULL)
	{
		while ((retv = cff_ffmpeg_file_open_cb(video, NULL, &str)) == 0) {;}
	}

	else
	{
		retv = (*func)(user_data);
		str  = NULL;
	}

	av_frame_free(&video->vframe.avframe);
	video->vframe.frame_width  = 0;
	video->vframe.frame_height = 0;

	if (retv == 0 || video->vframe.ic == NULL)
	{
		/* Error, or user canceled */
		if (video->vframe.ic != NULL)
		{
			avcodec_free_context(&(video->vframe.ic));
		}

		return str;
	}

	video->vframe.io_seek_end = avio_tell((video->inFormatCtx)->pb);

	video->vframe.min_gop_size *= video->vframe.frame_duration;
	video->vframe.max_gop_size *= video->vframe.frame_duration;

	video->vframe.current.frame	= video->vframe.dts_stop;
	video->vframe.current.packet	= video->vframe.dts_stop;
	video->vframe.stream_duration	= video->vframe.dts_stop - video->vframe.dts_start;
	video->vframe.frame_total	= video->vframe.stream_duration / video->vframe.frame_duration;

	/* Duration (h:m:s) */
	cff_ffmpeg_dts2time(video->vframe.stream_duration, is, video->vframe.time_total);

	return NULL;

} /* static const char *cff_ffmpeg_find_codec(cff_video_t*, enum AVCodecID, func_cff_callback_t, void*) */

const char *cff_ffmpeg_file_open(cff_video_t *video, const char *fname,
					func_cff_callback_t func, void *user_data)
{
	int		i, retv;
	enum AVCodecID	codec_id;
	AVFormatContext	*inFormatCtx;
	AVDictionary	*opts;

	if (fname == NULL || fname[0] == '\0')
	{
		return gettext_nop("Missing the file name!");
	}

	cff_ffmpeg_file_close(video);

	video->fname = cff_ffmpeg_strdup((char *) fname, -1);

	if (video->fname == NULL)
	{
		return gettext_nop("Cannot copy the filename!");
	}

	inFormatCtx = avformat_alloc_context();
	video->inFormatCtx = inFormatCtx;

	if (inFormatCtx == NULL)
	{
		return gettext_nop("Cannot allocate input context!");
	}

	opts = NULL;
	av_dict_set(&opts, "fflags", "genpts", 0);
	av_dict_set(&opts, "avoid_negative_ts", "make_zero", 0);

	/* Open video file */
	retv = avformat_open_input(&inFormatCtx, video->fname, NULL, &opts);

	av_dict_free(&opts);

	if (retv != 0)
	{
		av_dict_free(&opts);
		avformat_free_context(inFormatCtx);
		video->inFormatCtx = NULL;
	 	return gettext_nop("Couldn't open the input file!");
	}

	/* Retrieve stream information */

	if (avformat_find_stream_info(video->inFormatCtx, NULL) != 0)
	{
		avformat_close_input(&inFormatCtx);
	 	return gettext_nop("Couldn't find stream information!");
	}


	video->nb_streams = inFormatCtx->nb_streams;

	if (video->nb_streams == 0)
	{
		return gettext_nop("No video stream found!");
	}

	codec_id = AV_CODEC_ID_NONE;
	i = inFormatCtx->nb_programs;

	if (i > 0)
	{
		unsigned int	*ptr;
		AVProgram	**program;

		program = inFormatCtx->programs;

		if (program != NULL &&
			(ptr = av_malloc(i * sizeof(int))) != NULL)
		{
			int index;

			video->program.index = ptr;
			index = 0;

			do
			{
				codec_id = cff_ffmpeg_get_program_video_stream(video, index);

				if (codec_id != AV_CODEC_ID_NONE)
				{
					*(ptr++) = index;
					video->program.num++;
				}
			}
			while (++index < i);

			if (video->program.num == 0)
			{
				/* No video stream found */
				av_freep(&(video->program.index));
			}

			else
			{
				/* Open first program in the list */
				i = video->program.index[0];
				codec_id = cff_ffmpeg_get_program_video_stream(video, i);
			}

		} /* if (program != NULL &&
				(ptr = av_malloc(i * sizeof(int))) != NULL) */
	} /* if (i > 0) */

	if (codec_id == AV_CODEC_ID_NONE)
	{
		i = 0;

		do
		{
			codec_id = cff_ffmpeg_is_video_stream(video, i);
		}
		while (codec_id == AV_CODEC_ID_NONE &&
					++i < inFormatCtx->nb_streams);

		if (codec_id == AV_CODEC_ID_NONE)
		{
			return gettext_nop("No video codec found!");
		}

	} /* if (codec_id == AV_CODEC_ID_NONE) */

	i = 0;

	do
	{
		AVStream *is = (video->inFormatCtx)->streams[i];

		is->discard = AVDISCARD_ALL;
	}
	while (++i < inFormatCtx->nb_streams);

	return cff_ffmpeg_find_codec(video, codec_id, func, user_data);

} /* const char *cff_ffmpeg_file_open(cff_video_t*, const char*, func_cff_callback_t, void*) */

const char *cff_ffmpeg_program_open(cff_video_t *video, unsigned int index,
				func_cff_callback_t func, void *user_data)
{
	if (index < video->program.num && video->program.index != NULL)
	{
		int		i, prev_index;
		enum AVCodecID	codec_id;

		prev_index = video->vframe.stream_index;
		video->program.selected = index;

		if (video->vframe.ic)
		{
			avcodec_free_context(&(video->vframe.ic));
		}

		if (video->vframe.drain)
		{
			free(video->vframe.drain);
		}

		av_freep(&(video->vframe.rgb24.data));
		av_frame_free(&(video->vframe.avframe));

		cff_ffmpeg_video_cut_list_clear(video);

		cff_ffmpeg_video_undo_clear(video->vframe.undo);
		cff_ffmpeg_video_undo_clear(video->vframe.redo);

		memset(&(video->vframe), '\0', sizeof(cff_vframe_t));

		video->vframe.pict_type = cff_ffmpeg_get_pictype_str(NULL);

		avio_flush((video->inFormatCtx)->pb);
		avformat_flush(video->inFormatCtx);
		av_seek_frame(video->inFormatCtx, -1, -1, AVSEEK_FLAG_ANY);

		i = video->program.index[index];
		codec_id = cff_ffmpeg_get_program_video_stream(video, i);

		if (codec_id != AV_CODEC_ID_NONE)
		{
			const char	*str;
			AVStream	*is;

			is = (video->inFormatCtx)->streams[prev_index];
			is->discard = AVDISCARD_ALL;

			str = cff_ffmpeg_find_codec(video,
						codec_id, func, user_data);
			if (str != NULL)
			{
				/* failed */

				is = (video->inFormatCtx)->streams[prev_index];
				is->discard = AVDISCARD_ALL;
			}

			return str;

		} /* if (codec_id != AV_CODEC_ID_NONE) */

	} /* if (index < video->program.num && video->program.index != NULL) */

	return gettext_nop("No video codec found!");

} /* const char *cff_ffmpeg_program_open(cff_video_t*, unsigned int, func_cff_callback_t, void*) */

const char *cff_ffmpeg_get_program_name(cff_video_t *video, unsigned int index, int *prgnum)
{
	if (index < video->program.num && video->program.index != NULL)
	{
		unsigned int	i;
		const AVProgram	*prg;

		i   = video->program.index[index];
		prg = (video->inFormatCtx)->programs[i];

		if (prg != NULL)
		{
			AVDictionaryEntry *entry;

			if (prgnum != NULL)
			{
				*prgnum = prg->program_num;
			}

			entry = av_dict_get(prg->metadata,
						"service_name", NULL, 0);
			if (entry != NULL)
			{
				return entry->value;
			}

		} /* if (prg != NULL) */

	} /* if (index < video->program.num && video->program.index != NULL) */

	if (prgnum != NULL)
	{
		*prgnum = -1;
	}

	return NULL;

} /* const char *cff_ffmpeg_get_program_number(cff_video_t*, unsigned int, int*) */

void cff_ffmpeg_file_close(cff_video_t *video)
{
	if (video->fname)
	{
		/* malloc'ed */
		free(video->fname);
	}

	if (video->vframe.ic)
	{
		avcodec_free_context(&(video->vframe.ic));
	}

	if (video->vframe.drain)
	{
		/* malloc'ed */
		free(video->vframe.drain);
	}

	if (video->inFormatCtx)
	{
		avformat_close_input(&(video->inFormatCtx));
	}

	if (video->cut.list)
	{
		/* av_malloc'ed */
		av_free(video->cut.list);
	}

	if (video->program.index)
	{
		/* av_malloc'ed */
		av_free(video->program.index);
	}

	cff_ffmpeg_video_undo_clear(video->vframe.undo);
	cff_ffmpeg_video_undo_clear(video->vframe.redo);

	/* av_malloc'ed */
	av_freep(&(video->vframe.rgb24.data));
	av_frame_free(&(video->vframe.avframe));

	memset(video, '\0', sizeof(cff_video_t));

	video->vframe.pict_type = cff_ffmpeg_get_pictype_str(NULL);

} /* void cff_ffmpeg_file_close(cff_video_t*) */

static char *cff_ffmpeg_file_info_io(AVFormatContext *ctx,
					const char *fname, int is_output)
{
	if (ctx != NULL)
	{
		int		loglevel;
		cff_logbuffer_t	info, *old;

		old = gLogBufferPtr;
		gLogBufferPtr = memset(&info, '\0',
					sizeof(cff_logbuffer_t));

		av_log_set_callback(cff_ffmpeg_log_callback);
		loglevel = av_log_get_level();
		av_log_set_level(AV_LOG_DEBUG);

		info.max = 2048;
		av_dump_format(ctx, 0, fname, is_output);

		av_log_set_level(loglevel);

		if (old == NULL)
		{
			av_log_set_callback(av_log_default_callback);
		}

		gLogBufferPtr = old;

		if (info.str != NULL)
		{
			if (info.str[0] != '\0')
			{
				/* Remove last '\n' */
				return cff_ffmpeg_string_truncate(info.str, info.len - 1);
			}

			free(info.str);
		}

	} /* if (ctx != NULL) */

	return NULL;

} /* static char *cff_ffmpeg_file_info_io(AVFormatContext*, const char*, int) */

char *cff_ffmpeg_file_info(cff_video_t *video)
{
	return cff_ffmpeg_file_info_io(video->inFormatCtx, video->fname, 0);

} /* char *cff_ffmpeg_file_info(cff_video_t*) */

static void cff_ffmpeg_scale_dimension(cff_video_t *video, int *width, int *height)
{
	int		h, w, scale;
	double		aspect;
	AVRational	sar;

	/* from 'ffplay.c' */

	sar = av_guess_sample_aspect_ratio(video->inFormatCtx,
		(video->inFormatCtx)->streams[video->vframe.stream_index],
		video->vframe.avframe);

	if (sar.den == 0)
	{
		aspect = 1.0;
	}

	else
	{
		aspect = av_q2d(sar);

		if (aspect <= 0.0)
		{
			aspect = 1.0;
		}
	}

	aspect *= (double) video->vframe.avframe->width /
				(double) video->vframe.avframe->height;

	scale	= 0;
	w	= *width;
	h	= *height;

	if (w != video->vframe.avframe->width)
	{
		/* Scale the width */
		scale = 0x01;
	}

	if (h != video->vframe.avframe->height)
	{
		/* Scale the height (or width and height) */
		scale |= 0x02;
	}

	if (scale == 0 && sar.num != sar.den)
	{
		/* Scale only the width */
		scale = 0x01;
	}

	if (scale & 0x01)
	{
		h = lrint((double) w / aspect) & ~1;
	}

	if (scale & 0x02)
	{
		w = lrint((double) *height * aspect) & ~1;
	}

	if (scale == 0x03)
	{
		if (w > *width)
		{
			w = *width;
			h = lrint((double) w / aspect) & ~1;
		}

		else if (h > *height)
		{
			h = *height;
			w = lrint((double) h * aspect) & ~1;
		}

	} /* if (scale == 0x03) */

	*width  = w;
	*height = h;

} /* static void cff_ffmpeg_scale_dimension(cff_video_t*, int*, int*) */

static int cff_ffmpeg_scale_frame(AVFrame *frame, int width, int height,
	enum AVPixelFormat pixfmt, int align, uint8_t *data[4], int linesize[4])
{
	int			retv;
	struct SwsContext	*sws_context;

	/*
		'av_image_alloc' may return with an error without
		freeing the allocated memory (on wrong align) !
	*/

	data[0] = NULL;

	if (av_image_alloc(data, linesize, width, height, pixfmt, align) < 0)
	{
		return -1;
	}

	sws_context = sws_getContext(frame->width, frame->height,
					frame->format,
					width, height,
					pixfmt,
					SWS_BICUBIC, NULL, NULL, NULL);
	if (sws_context == NULL)
	{
		return -1;
	}

	retv = sws_scale(sws_context,
			(const uint8_t* const*) frame->data, frame->linesize,
			0, frame->height,
			data, linesize);

	sws_freeContext(sws_context);

	return retv;

} /* static int cff_ffmpeg_scale_frame(AVFrame*, int, int,
				enum AVPixelFormat, int, uin8t_t[4]*, int[4]) */

int cff_ffmpeg_frame2rgb24(cff_video_t *video, int width, int height)
{
	int		h, w, retv;
	AVFrame		*frame;

	struct
	{
		uint8_t	*data[4];
		int	linesize[4];

	} rgb;

	frame = video->vframe.avframe;

	if (frame == NULL)
	{
		return -1;
	}

	w = width;
	h = height;
	cff_ffmpeg_scale_dimension(video, &w, &h);

	retv = cff_ffmpeg_scale_frame(frame,
				w, h, AV_PIX_FMT_RGB24,
					1, rgb.data, rgb.linesize);
	if (retv > 0)
	{
		av_freep(&(video->vframe.rgb24));

		video->vframe.rgb24.data   = rgb.data[0];
		video->vframe.rgb24.width  = w;
		video->vframe.rgb24.height = h;
	}

	else
	{
		av_freep(&(rgb.data[0]));
	}

	return retv;

} /* int cff_ffmpeg_frame2rgb24(cff_video_t*, int, int) */

static int cff_ffmpeg_seek(cff_video_t *video, int64_t dts)
{
	int	any, flags;
	int64_t	dts_seek;

	/*
		Keep this order (else 'avi' with 'dts_start' don't work):
		 1) flags = 0
		 2) flags = AVSEEK_FLAG_FRAME
		 3) flags = AVSEEK_FLAG_BYTE
		 4) flags = AVSEEK_FLAG_ANY
	*/

	flags = 0;
	any = 0;
	dts_seek = dts;

	for (;;)
	{
		int retv;

		if (dts <= video->vframe.dts_start)
		{
			/* Seek first frame */

			if (flags == 0)
			{
				if (dts_seek != video->vframe.dts_start - video->vframe.max_gop_size)
				{
					/* Rewind - for some files necessary */
					avio_seek((video->inFormatCtx)->pb,
						video->vframe.io_seek_start, SEEK_SET);

					/* Get first frame - works with 'mp4' and 'matroska' */
					dts_seek = video->vframe.dts_start - video->vframe.max_gop_size;
				}

				else
				{
					/* Get first frame - works with 'avi' and 'matroska' */
					flags = AVSEEK_FLAG_FRAME; 
					dts_seek = 0;
				}
			}

			else if (flags == AVSEEK_FLAG_FRAME)
			{
				/* Get first frame - works with 'mpeg-ts' and 'mpeg-2ps' */
				flags = AVSEEK_FLAG_BYTE;
				dts_seek = video->vframe.io_seek_start;
			}

			else if (flags == AVSEEK_FLAG_BYTE)
			{
				/* Get first frame - last chance */
				flags = AVSEEK_FLAG_ANY;
				dts_seek = dts;
			}

			else
			{
				/* Failed */
				return -1;
			}

		} /* if (dts <= video->vframe.dts_start) */

		retv = av_seek_frame(video->inFormatCtx,
				video->vframe.stream_index,
				dts_seek, flags);

		if (retv >= 0)
		{
			int64_t pos;

			if (dts > video->vframe.dts_start)
			{
				break;
			}

			pos = avio_tell((video->inFormatCtx)->pb);

			if (pos <= video->vframe.io_seek_start)
			{
				break;
			}
		}

		else if (flags == AVSEEK_FLAG_ANY || dts <= video->vframe.dts_start)
		{
			if (flags == AVSEEK_FLAG_ANY)
			{
				if (any)
				{
					return -1;
				}

				/* Getting first frame failed - try second frame */
				dts += video->vframe.frame_duration;
				dts_seek = dts;
				any = 1;
			}
		}

		else
		{
			flags = AVSEEK_FLAG_ANY;
			any = 1;
		}

	} /* for (;;) */

	return 0;

} /* static int cff_ffmpeg_seek(cff_video_t*, int64_t) */

static int cff_ffmpeg_get_packet_with_dts(cff_video_t *video, int64_t dts, AVPacket *packet)
{
	/* Get the packet with the given timestamp */

	int	n;
	int64_t	dts_sub;
	
	n = 5;
	dts_sub = video->vframe.frame_duration;

	do
	{
		int64_t	dts_last, dts_seek;

		/*
			Move a few frames back, else the frame
			we are looking for might be skipped (???)
		*/

		if (dts == AV_NOPTS_VALUE)
		{
			/* May happen if seeking the first frame */
			dts_seek = video->vframe.dts_start;
			n = 0;
		}

		else
		{
			dts_sub <<= 2;
			dts_seek = dts - dts_sub;
		}

		if (dts_seek <= video->vframe.dts_start)
		{
			if (cff_ffmpeg_seek(video, dts_seek) != 0)
			{
				return -1;
			}

			n = 0;
		}

		else  if (av_seek_frame(video->inFormatCtx,
					video->vframe.stream_index,
					dts_seek,
					AVSEEK_FLAG_ANY) < 0)
		{
			continue;
		}

		dts_last = dts - video->vframe.frame_duration;

		do
		{
			if (av_read_frame(video->inFormatCtx, packet) < 0)
			{
				break;
			}

			if (packet->stream_index == video->vframe.stream_index)
			{
				if (dts == packet->dts)
				{
					return 0;
				}

				dts_last = packet->dts;
			}

			av_packet_unref(packet);
		}
		while (dts > dts_last);
	}
	while (--n > 0);

	return -1;

} /* static int cff_ffmpeg_get_packet_with_dts(cff_video_t*, int64_t, AVPacket*) */

static int cff_ffmpeg_get_frame(cff_video_t *video, int64_t dts, int width, int height, enum cff_FrameFlags frame_flags)
{
	int		retv, rewind, num_last, dts_error;
	unsigned int	drain_mode;
	int64_t		diff, pkt_dts, pkt_last, pkt_next, iframe_last, dts_seek, gop_size;
	AVFrame		*frame;
	AVPacket	packet;
	cff_packet_t	*pkt_ptr;

	struct
	{
		AVFrame	*frame;
		int64_t	dts;
	} last;


	if (dts < video->vframe.dts_start)
	{
		dts = video->vframe.dts_start;
	}

	else if (dts > video->vframe.dts_stop)
	{
		dts = video->vframe.dts_stop;
	}

	diff = dts - video->vframe.current.frame;

	if (diff == 0)
	{
		/* Nothing to do */
		return -1;
	}


	pkt_next = dts;

	if (diff > 0)
	{
		if (diff == 1)
		{
			/* Next frame */
			pkt_next--;
			pkt_next += video->vframe.frame_duration;

			if (pkt_next > video->vframe.dts_stop)
			{
				pkt_next = video->vframe.dts_stop;
			}
		}

		else if (diff < video->vframe.frame_duration)
		{
			/* Clicked at same position */
			return -1;
		}
	}

	else
	{
		for (;;)
		{
			if (diff == -1)
			{
				/* Previous frame */
				pkt_next++;
				pkt_next -= video->vframe.frame_duration;

				if (pkt_next < video->vframe.dts_start)
				{
					pkt_next = video->vframe.dts_start;
				}

				break;
			}

			if (diff > -video->vframe.frame_duration)
			{
				/* Clicked at same position */
				return -1;
			}

			if (diff > -video->vframe.frame_duration * 2)
			{
				/* Clicked near same position */
				diff = -1;
			}

			else
			{
				break;
			}

		} /* for (;;) */

	} /* if (diff > 0) .. else */


	if (pkt_next == video->vframe.dts_stop &&
		video->vframe.current.frame > video->vframe.dts_stop)
	{
		/* may happen if frame duration differs */
		return -1;
	}

	gop_size    = video->vframe.max_gop_size;
	pkt_last    = AV_NOPTS_VALUE;
	iframe_last = AV_NOPTS_VALUE;
	dts_seek    = dts;
	rewind      = 0;
	num_last    = 0;
	dts_error   = 0;
	drain_mode  = 0;
	last.frame  = NULL;

	/*
		Do not seek if going to the next frame
		and all picture types or P-Frames should
		be displayed
	*/

	if ((frame_flags != 0 && (frame_flags & kPFrameFlag) == 0) ||
			diff != 1 || video->vframe.packet[0].pos == -1)
	{
		if (pkt_next < video->vframe.dts_stop - gop_size)
		{
			gop_size = video->vframe.min_gop_size;
		}

		if (pkt_next > video->vframe.dts_start)
		{
			dts_seek -= gop_size;
		}

		rewind = 1;

	} /* if ((frame_flags != 0 && (frame_flags & kPFrameFlag) == 0) ||
				diff != 1 || video->vframe.packet[0].pos == -1) */

	frame = av_frame_alloc();

	if (frame == NULL)
	{
		return -1;
	}

	av_init_packet(&packet);


	/* Start */

	do
	{
		int		i;
		AVRational	sar;
		AVFrame		*tmp_frame;

		if (rewind > 0)
		{
			if (num_last > 20 ||
				dts_seek < video->vframe.dts_start - gop_size)
			{
				retv = -1;
				break;
			}

			retv = cff_ffmpeg_seek(video, dts_seek);

			if (retv != 0)
			{
				break;
			}

			dts_seek -= gop_size;

			drain_mode = 0;

			if (video->vframe.drain != NULL)
			{
				(video->vframe.drain)->idx = 0;
			}

			avcodec_flush_buffers(video->vframe.ic);
			rewind = -1;

		} /* if (rewind > 0) */

		if (drain_mode || (retv = av_read_frame(video->inFormatCtx, &packet)) != 0)
		{
			if (num_last < 3 && pkt_next + video->vframe.max_gop_size >= video->vframe.dts_stop)
			{
				/* Reading last packed failed */

				retv = 0;

				if (drain_mode == 0)
				{
					if (num_last == 0 && last.frame == NULL)
					{
						if (diff == 1 && video->vframe.drain != NULL &&
							(video->vframe.drain)->num >= (video->vframe.drain)->idx)
						{
							drain_mode = (video->vframe.drain)->idx;
						}

						else
						{
							num_last++;
							rewind = 1;
							continue;
						}
					}

					if (drain_mode == 0 && avcodec_send_packet(video->vframe.ic, NULL) == 0)
					{
						drain_mode = 1;
					}

				} /* if (drain_mode == 0) */

				if (drain_mode &&
					avcodec_receive_frame(video->vframe.ic, frame) == 0)
				{
					goto _got_frame;
				}

				drain_mode = 0;

				if (video->vframe.drain != NULL)
				{
					(video->vframe.drain)->idx = drain_mode;
				}

				if (last.frame != NULL)
				{
					if (last.dts != video->vframe.current.frame)
					{
						goto _got_picture;
					}

					if (diff > 0)
					{
						retv = -1;
						continue;
					}
				}

				if (num_last > 1)
				{
					/* Try once again, stop if not found */
					dts -= video->vframe.frame_duration;
				}

				/* Clear last packet */
				pkt_last = AV_NOPTS_VALUE;

				/* Store the last frame */
				pkt_next = video->vframe.dts_stop;

				/* Try again */
				num_last++;
				rewind = 1;

			} /* if (num_last < 3 && (pkt_next == video->vframe.dts_stop ||
				video->vframe.current.frame + gop_size >= video->vframe.dts_stop)) */

			/* finish */
			continue;

		} /* if (drain_mode || (retv = av_read_frame(video->inFormatCtx, &packet)) != 0) */

		if (packet.stream_index != video->vframe.stream_index)
		{
			continue;
		}

		if (packet.dts != AV_NOPTS_VALUE)
		{
			int same_packet = 0;

			if (pkt_last == AV_NOPTS_VALUE)
			{
				pkt_last = packet.dts;
			}

			else if (pkt_last == packet.dts)
			{
				num_last++;

				if (num_last > video->vframe.max_gop_size / video->vframe.frame_duration)
				{
					retv = -1;
					continue;
				}

				if (frame_flags == 0 &&
					pkt_next != packet.dts &&
						dts_seek > video->vframe.dts_start)
				{
					same_packet = 1;
				}
			}

			if (rewind < 0)
			{
				rewind = 1;

				if (dts_seek > video->vframe.dts_start)
				{
					int64_t distance = packet.dts + video->vframe.max_gop_size * 8;

					if (gop_size > video->vframe.max_gop_size * 8 &&
							(dts_seek > distance || dts_seek < distance))
					{
						dts_error = 1;
					}

					if (packet.dts > dts)
					{
						/*
							The first packet dts after seeking should
							be less or equal than the requested dts!
						*/

						if (dts > distance)
						{
							if (diff > 0)
							{
								/*
									The packet dts should be
									less than the requested dts!
								*/

								if (dts < video->vframe.dts_stop)
								{
									dts_seek += gop_size * 2;
								}

								else if (num_last > 1)
								{
									dts_seek += gop_size * 2;
								}
							}

							else if (pkt_next > video->vframe.dts_start)
							{
								/*
									diff < 0
									The requested dts should be
									less than the packet dts!
								*/

								if (dts_seek > (video->vframe.dts_start << 1))
								{
									dts_seek >>= 1;
									continue;
								}

								if (dts_seek > video->vframe.dts_start + gop_size)
								{
									dts_seek -= gop_size * 2;
								}

							} /* if (diff > 0) .. else */

						} /* if (dts > distance) */

						if (same_packet)
						{
							if (gop_size < video->vframe.max_gop_size)
							{
								gop_size = video->vframe.max_gop_size;
							}

							else if (dts_seek > video->vframe.dts_start + gop_size)
							{
								if (num_last & 1)
								{
									gop_size <<= num_last;
								}

								else
								{
									gop_size <<= 1;
								}
							}

						} /* if (same_packet) */

						continue;

					} /* if (packet.dts > dts) */

					if ((video->vframe.ic)->codec_id == AV_CODEC_ID_HEVC)
					{
						/*
							The first packet after seeking should
							be a Key-frame for propper decoding!
						*/

			 			if ( ! (packet.flags & AV_PKT_FLAG_KEY))
						{
							/* Skip this frame */
							rewind = -1;
							continue;
						}

						if (iframe_last == packet.dts)
						{
							/* Same Key-frame, try again */
							num_last++;
							rewind = -1;
							continue;
						}

						iframe_last = packet.dts;
					}

				} /* if (dts_seek > video->vframe.dts_start) */

				if (same_packet)
				{
					continue;
				}

				rewind = 0;

			} /* if (rewind < 0) */

		} /* if (packet.dts != AV_NOPTS_VALUE) */


		/*
			Store the packet dts, may differ from the
			displayed frame
			If only Key-frame and/or I-frame is selected,
			store only the packet dts of Key-frames
		*/

		if (frame_flags == 0 || (frame_flags & kPFrameFlag) ||
					(packet.flags & AV_PKT_FLAG_KEY))
		{
			if (num_last < 2 &&
				packet.dts == AV_NOPTS_VALUE &&
					pkt_next > video->vframe.dts_start)
			{
				if (dts_seek > video->vframe.dts_start + video->vframe.max_gop_size)
				{
					/*
						The packet dts is invalid. If this
						is the first packet and the minimum
						gop size is selected, change it to
						the maximum gop size and try again!
					*/

					if (gop_size < video->vframe.max_gop_size)
					{
						gop_size = video->vframe.max_gop_size;
						dts_seek = dts - gop_size;
					}

					rewind = 1;
				}

				num_last++;
				continue;

			} /* if (num_last < 2 && .. */

			if (frame_flags == kPFrameFlag &&
				(packet.flags & AV_PKT_FLAG_KEY) &&
							packet.dts > dts)
			{
				num_last++;

				if (num_last > 7)
				{
					/*
						No P-frame found!
						If not searching the first
						frame after 'Open' break,
						else display the first frame
					*/

					if (video->vframe.avframe != NULL)
					{
						retv = -1;
						continue;
					}

					frame_flags = 0;
					num_last = 0;
					pkt_last = AV_NOPTS_VALUE;
					dts_seek = dts;
					rewind = 1;
					continue;

				} /* if (num_last > 7) */

			} /* if (frame_flags == kPFrameFlag &&
				(packet.flags & AV_PKT_FLAG_KEY) &&
							packet.dts > dts) */

			if (packet.dts != AV_NOPTS_VALUE ||
						pkt_next == video->vframe.dts_start)
			{
				memmove(&video->vframe.packet[1], video->vframe.packet,
					(kPacketArrayEntries - 1) * sizeof(cff_packet_t));
				video->vframe.packet[0].dts = packet.dts;
				video->vframe.packet[0].pos = packet.pos;
			}

		} /* if (frame_flags == 0 || (frame_flags & kPFrameFlag) ||
						(packet.flags & AV_PKT_FLAG_KEY)) */


		/* Decode */

		if (avcodec_send_packet(video->vframe.ic, &packet) != 0)
		{
			continue;
		}

		if (avcodec_receive_frame(video->vframe.ic, frame) != 0)
		{
			if (num_last < 3 && packet.dts != AV_NOPTS_VALUE &&
				((diff == -1) ? packet.dts >= pkt_next : packet.dts >= dts))
			{
				/*
					Rewind if next frame is
					requested and receiving it failed
				*/

				if (diff == 1)
				{
					if (dts <= video->vframe.dts_start + gop_size)
					{
						dts_seek = video->vframe.dts_start;
					}

					num_last++;
					rewind = 1;
				}

				else if (diff < 0 || frame_flags == 0)
				{
					if (dts_seek < video->vframe.dts_start)
					{
						dts_seek = video->vframe.dts_start;
					}

					num_last++;
					rewind = 1;
				}

			} /* if (num_last < 3 && packet.dts != AV_NOPTS_VALUE &&
				((diff == -1) ? packet.dts >= pkt_next : packet.dts >= dts)) */

			continue;

		} /* if (avcodec_receive_frame(video->vframe.ic, frame) != 0) */

_got_frame:
		pkt_dts = frame->pkt_dts;

		if (pkt_dts == AV_NOPTS_VALUE &&
			drain_mode && video->vframe.drain != NULL &&
				(video->vframe.drain)->num >= drain_mode)
		{
			pkt_dts = (video->vframe.drain)->pkt_dts[drain_mode - 1];
			drain_mode++;
			(video->vframe.drain)->idx = drain_mode;
		}

		if (pkt_dts == AV_NOPTS_VALUE ||
			pkt_dts < video->vframe.dts_start)
		{
			/* Discard invalid packets */
			continue;
		}

		if (frame_flags != 0)
		{
			enum cff_FrameFlags tmp_flags = 0;

			if (frame->key_frame != 0)
			{
				tmp_flags |= kKeyFrameFlag;
			}

			if (frame->pict_type == AV_PICTURE_TYPE_I)
			{
				tmp_flags |= kIFrameFlag;
			}

			if (frame->pict_type == AV_PICTURE_TYPE_P)
			{
				tmp_flags |= kPFrameFlag;
			}

			if ((frame_flags & tmp_flags) == 0)
			{
				/* Not the wanted picture type */
				continue;
			}

			if (dts > pkt_dts)
			{
				if (dts != video->vframe.dts_start)
				{
					if (last.frame == NULL)
					{
						last.frame = av_frame_clone(frame);
					}

					else
					{
						av_frame_unref(last.frame);
						av_frame_ref(last.frame, frame);
					}

					last.dts = pkt_dts;
				}

				if (diff > 0 || dts_error == 0)
				{
					continue;
				}

			} /* if (dts > pkt_dts) */

			if (diff < 0 && dts != video->vframe.dts_start)
			{
				if (pkt_dts > dts)
				{
					if (last.frame != NULL)
					{
						goto _got_picture;
					}

					rewind = 1;
					continue;
				}

				if (pkt_dts >= video->vframe.current.frame)
				{
					rewind = 1;
					continue;
				}

			} /* if (diff < 0 && dts != video->vframe.dts_start) */

		} /* if (frame_flags != 0) */

		else
		{
			if (dts > pkt_dts)
			{
				if (diff == -1 || pkt_next == video->vframe.dts_stop)
				{
					if (last.frame == NULL)
					{
						last.frame = av_frame_clone(frame);
						last.dts = pkt_dts;
					}

					else if (pkt_dts > last.dts)
					{
						av_frame_unref(last.frame);
						av_frame_ref(last.frame, frame);
						last.dts = pkt_dts;
					}
				}

				if (diff > 0 || dts_error == 0)
				{
					continue;
				}

			} /* if (dts > pkt_dts) */

			if (diff < 0 && pkt_dts >= video->vframe.current.frame)
			{
				if (last.frame != NULL)
				{
					goto _got_picture;
				}

				if (diff == -1 || num_last < 3)
				{
					num_last++;
					rewind = 1;
					continue;
				}
			}

		} /* if (frame_flags != 0) .. else */

		/* Picture found */

		if (last.frame != NULL)
		{
			av_frame_free(&last.frame);
		}

		/* Success */
_got_picture:
		if (last.frame != NULL)
		{
			av_frame_free(&frame);
			frame = last.frame;
			pkt_dts = last.dts;
		}

		/* Get the packet dts belonging to this frame */

		pkt_ptr = video->vframe.packet;
		i = kPacketArrayEntries;

		do
		{
			if (pkt_ptr->pos == frame->pkt_pos)
			{
				/* Packet found */
				video->vframe.current.packet = pkt_ptr->dts;
				break;
			}
		}
		while (++pkt_ptr, --i > 0);

		if (last.frame != NULL)
		{
			last.frame = NULL;
			video->vframe.packet[0].pos = -1;
		}

		if (i == 0)
		{
			/* No packet found */

			if (num_last < 5 &&
				pkt_next > video->vframe.dts_start &&
					pkt_next < video->vframe.dts_stop)
			{
				/* Try again */
				num_last++;
				rewind = 1;
				continue;
			}

			/* Packet and frame dts are equal (may incorrect) */
			video->vframe.current.packet = pkt_dts;

		} /* if (i == 0) */

		video->vframe.current.frame = pkt_dts;


		/* Take this frame */

		tmp_frame = video->vframe.avframe;
		video->vframe.avframe = frame;

		retv = cff_ffmpeg_frame2rgb24(video, width, height);

		if (retv <= 0)
		{
			video->vframe.avframe = tmp_frame;
			retv = -1;
			continue;
		}

		av_frame_free(&tmp_frame);

		video->vframe.pict_type =
				cff_ffmpeg_get_pictype_str(frame);

		video->vframe.frame_width  = frame->width;
		video->vframe.frame_height = frame->height;

		sar = av_guess_sample_aspect_ratio(video->inFormatCtx,
			(video->inFormatCtx)->streams[video->vframe.stream_index],
			frame);

		if (sar.num == 0)
		{
			double aspect;

			aspect = (double) video->vframe.avframe->width /
					(double) video->vframe.avframe->height;
			sar = av_d2q(aspect, 1024 * 1024);
			video->vframe.frame_aspect[0] = sar.num;
			video->vframe.frame_aspect[1] = sar.den;
		}

		else if (sar.den == 0)
		{
			video->vframe.frame_aspect[0] = 1;
			video->vframe.frame_aspect[1] = 1;
		}

		else
		{
			av_reduce((int *) &video->vframe.frame_aspect[0],
				(int *) &video->vframe.frame_aspect[1],
				frame->width * (int64_t) sar.num,
				frame->height * (int64_t) sar.den,
				1024 * 1024);
		}

		dts = pkt_dts - video->vframe.dts_start;

		if (dts < 0)
		{
			dts = -dts;
		}

		cff_ffmpeg_dts2time(dts,
			video->inFormatCtx->streams[video->vframe.stream_index],
						video->vframe.time_current);

		video->vframe.frame_current = dts / video->vframe.frame_duration;
	}
	while (av_packet_unref(&packet), retv == 0);

	if (last.frame != NULL)
	{
		av_frame_free(&last.frame);
	}

	if (retv < 0)
	{
		av_frame_free(&frame);
		video->vframe.packet[0].pos = -1;

		retv = -1;

	} /* if (retv < 0) */

	cff_ffmpeg_log_callback_flush();

	return retv;

} /* static int cff_ffmpeg_get_frame(cff_video_t*, int64_t, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_next(cff_video_t *video, int next, int width, int height, enum cff_FrameFlags frame_flags)
{
	int		i;
	int64_t		dts, diff, t;
	cff_dts_t	*ptr;

	switch (next)
	{
		case kVideoFramePrev05Sec:
			dts = cff_ffmpeg_sec2dts(-5, video);
			break;
		case kVideoFramePrevHalfSec:
			dts = cff_ffmpeg_sec2dts(-1, video) / 2;
			break;
		case kVideoFramePrevSingle:
			dts = -1;
			break;
		case kVideoFrameNextSingle:
			dts = 1;
			break;
		case kVideoFrameNextHalfSec:
			dts = cff_ffmpeg_sec2dts(1, video) / 2;
			break;
		case kVideoFrameNext05Sec:
			dts = cff_ffmpeg_sec2dts(5, video);
			break;
		case kVideoFramePrev10Min:
			dts = cff_ffmpeg_sec2dts(-(10 * 60), video);
			break;
		case kVideoFramePrev05Min:
			dts = cff_ffmpeg_sec2dts(-(5 * 60), video);
			break;
		case kVideoFramePrev01Min:
			dts = cff_ffmpeg_sec2dts(-(1 * 60), video);
			break;
		case kVideoFrameNext01Min:
			dts = cff_ffmpeg_sec2dts(1 * 60, video);
			break;
		case kVideoFrameNext05Min:
			dts = cff_ffmpeg_sec2dts(5 * 60, video);
			break;
		case kVideoFrameNext10Min:
			dts = cff_ffmpeg_sec2dts(10 * 60, video);
			break;
		case kVideoFrameNextCut:
		case kVideoFrameEnd:
			dts = video->vframe.dts_stop;
			break;
		default: /* kVideoPrevCut and kVideoFrameStart */
			dts = video->vframe.dts_start;
			break;

	}; /* switch (next) */

	if (next == kVideoFramePrevCut)
	{
		i = video->cut.num;

		if (i > 0 && video->cut.list != NULL)
		{
			ptr = &video->cut.list[i];

			do
			{
				ptr--;
				t = ptr->frame;

				if (video->vframe.current.frame > t)
				{
					frame_flags = 0;
					dts = t;
					break;
				}
			}
			while (--i > 0);

		} /* if (ptr != NULL) */
	}

	else if (next == kVideoFrameNextCut)
	{
		i = video->cut.num;

		if (i > 0 && video->cut.list != NULL)
		{
			ptr = video->cut.list;

			do
			{
				t = ptr->frame;
				ptr++;

				if (t > video->vframe.current.frame)
				{
					frame_flags = 0;
					dts = t;
					break;
				}
			}
			while (--i > 0);

		} /* if (ptr != NULL) */
	}

	else if (next != kVideoFrameStart && next != kVideoFrameEnd)
	{
		dts += video->vframe.current.frame;

	} /* if (next == kVideoFrameCutPrev) ..
			else if (next == kVideoFrameCutNext) .. else if .. ) */


	if (dts < video->vframe.dts_start)
	{
		dts = video->vframe.dts_start;
	}

	else if (dts > video->vframe.dts_stop)
	{
		dts = video->vframe.dts_stop;
	}

	diff = dts - video->vframe.current.frame;

	if (diff == 0)
	{
		return -1;
	}

	if (diff > 0)
	{
		if (diff < video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame + 1;
		}
	}

	else
	{
		if (diff > -video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame - 1;
		}

	} /* if (diff > 0) .. else */

	return cff_ffmpeg_get_frame(video, dts, width, height, frame_flags);

} /* int cff_ffmpeg_get_frame_next(cff_video_t*, int, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_slider(cff_video_t *video, double next, int width, int height, enum cff_FrameFlags frame_flags)
{
	int64_t dts;
 
	dts  = next * (double) video->vframe.stream_duration;
	dts += video->vframe.dts_start;

	return cff_ffmpeg_get_frame(video, dts, width, height, frame_flags);

} /* int cff_ffmpeg_get_frame_slider(cff_video_t*, double, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_cutpoint(cff_video_t *video, unsigned int next, int width, int height, enum cff_FrameFlags frame_flags)
{
	int64_t dts, diff;

	if (next >= video->cut.num || video->cut.list == NULL)
	{
		return -1;
	}

	dts  = video->cut.list[next].frame;
	diff = dts - video->vframe.current.frame;

	if (diff == 0)
	{
		return -1;
	}

	if (diff > 0)
	{
		if (diff < video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame + 1;
		}
	}

	else
	{
		if (diff > -video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame - 1;
		}

	} /* if (diff > 0) .. else */

	return cff_ffmpeg_get_frame(video, dts, width, height, 0);

} /* int cff_ffmpeg_get_frame_cutpoint(cff_video_t*, unsigned int, int, int, enum cff_FrameFlags) */

static void cff_ffmpeg_video_undo_clear(cff_undo_t **list)
{
	int		n;
	cff_undo_t	*ptr;

	n = kUndoArrayEntries;

	do
	{
		ptr = *list;

		if (ptr == NULL)
		{
			break;
		}

		free(ptr);
		*(list++) = NULL;
	}
	while (--n > 0);

} /* static void cff_ffmpeg_video_undo_clear(cff_undo_t**) */

static int cff_ffmpeg_video_undo_redo(cff_video_t *video, int action)
{
	int		retv;
	unsigned int	n;
	cff_undo_t	*cut, **src, **dst;

	int (*func)(cff_video_t*, cff_dts_t*);

	if (action == kUndo)
	{
		src = video->vframe.undo;
		dst = video->vframe.redo;
	}

	else
	{
		src = video->vframe.redo;
		dst = video->vframe.undo;
	}

	cut = src[0];

	if (cut->action == kCutpointAdd)
	{
		cut->action = kCutpointRemove;
		func = cff_ffmpeg_video_cut_remove_dts;
	}

	else
	{
		cut->action = kCutpointAdd;
		func = cff_ffmpeg_video_cut_add_dts;
	}

	n = 0;

	do
	{
		/* Add or remove cutpoints */
		retv = (*func)(video, &(cut->dts[n]));

		if (retv < 0)
		{
			return retv;
		}
	}
	while (++n < cut->num);

	/* Move one entry down */
	memmove(&src[0], &src[1],
		sizeof(cff_undo_t*) * (kUndoArrayEntries - 1));

	/* Clear last entry */
	src[kUndoArrayEntries - 1] = NULL;

	if (dst[kUndoArrayEntries - 1] != NULL)
	{
		/* Free last entry (this shouldn't happen!) */
		free(dst[kUndoArrayEntries - 1]);
	}

	/* Move one entry up */
	memmove(&dst[1], &dst[0],
		sizeof(cff_undo_t*) * (kUndoArrayEntries - 1));

	/* Copy new entry */
	dst[0] = cut;

	/* Clear 'dirty' if list is empty */
	video->dirty = (video->cut.num > 0) ? 1 : 0;

	return retv;

} /* static int cff_ffmpeg_video_undo_redo(cff_video_t*, int) */

int cff_ffmpeg_video_undo(cff_video_t *video)
{
	if (video->vframe.undo[0] == NULL)
	{
		return -1;
	}

	return cff_ffmpeg_video_undo_redo(video, kUndo);

} /* int cff_ffmpeg_video_undo(cff_video_t*) */

int cff_ffmpeg_video_redo(cff_video_t *video)
{
	if (video->vframe.redo[0] == NULL)
	{
		return -1;
	}

	return cff_ffmpeg_video_undo_redo(video, kRedo);

} /* int cff_ffmpeg_video_redo(cff_video_t*) */

static void cff_ffmpeg_video_undo_add(cff_video_t *video, cff_undo_t *cut)
{
	if (cut != NULL)
	{
		cff_undo_t **undo = video->vframe.undo;

		if (undo[kUndoArrayEntries - 1] != NULL)
		{
			/* Free last entry */
			free(undo[kUndoArrayEntries - 1]);
		}

		/* Move one entry up */
		memmove(&undo[1], &undo[0],
				sizeof(cff_undo_t*) * (kUndoArrayEntries - 1));

		undo[0] = cut;

		cff_ffmpeg_video_undo_clear(video->vframe.redo);

		/* Clear 'dirty' if list is empty */
		video->dirty = (video->cut.num > 0) ? 1 : 0;

	} /* if (cut != NULL) */

} /* static void cff_ffmpeg_video_undo_add(cff_video_t*, cff_undo_t*) */

static void cff_ffmpeg_video_undo_add_current(cff_video_t *video, enum cff_UndoAction action)
{
	cff_undo_t undo;

	undo.action	= action;
	undo.num	= 1;
	undo.dts[0]	= video->vframe.current;

	cff_ffmpeg_video_undo_add(video,
			cff_ffmpeg_memdup(&undo, sizeof(cff_undo_t)));

} /* static void cff_ffmpeg_video_undo_add_current(cff_video_t*, enum cff_UndoAction) */

unsigned int cff_ffmpeg_get_cut_frame_number(cff_video_t *video, unsigned int index)
{
	if (index < video->cut.num && video->cut.list != NULL)
	{
		return (video->cut.list[index].frame -
					video->vframe.dts_start) /
						video->vframe.frame_duration;
	}

	return (unsigned int) -1;

} /* int cff_ffmpeg_get_cut_frame_number(cff_video_t*, unsigned int) */

static int cff_ffmpeg_video_cut_find_dts(const cff_video_t *video, const cff_dts_t *src, int both)
{
	/*
		Return the index if
		  both
		    frame and packet
		  else
		    either frame or packet
		do exist in the list.	
	*/

	int		i, match;
	cff_dts_t	*ptr;

	i   = video->cut.num;
	ptr = video->cut.list;

	if (video->vframe.ic == NULL || i <= 0 || ptr == NULL)
	{
		return -1;
	}

	match = (both == 0) ? 1 : 2;

	i--;
	ptr += i;

	do
	{
		int found = 0;

		if (src->frame == ptr->frame)
		{
			found++;
		}

		if (src->packet == ptr->packet)
		{
			found++;
		}

		if (found >= match)
		{
			break;
		}

		ptr--;
	}
	while (--i >= 0);

	return i;

} /* static int cff_ffmpeg_video_cut_find_dts(const cff_video_t*, const cff_dts_t*, int) */

int cff_ffmpeg_video_cut_find_current(const cff_video_t *video)
{
	return cff_ffmpeg_video_cut_find_dts(video, &(video->vframe.current), 1);

} /* int cff_ffmpeg_video_cut_find_current(const cff_video_t*) */

static int cff_ffmpeg_video_cut_add_dts(cff_video_t *video, cff_dts_t *src)
{
	int		i;
	unsigned int	n;
	cff_dts_t	*dst;

	if (video->vframe.ic == NULL ||
		src->frame < video->vframe.dts_start ||
			src->frame > video->vframe.dts_stop)
	{
		/* Invalid */
		return -1;
	}

	/*
		With no B-frames the DTS should be
		equal, else the frame DTS should be
		greater (or equal) than the packet DTS!
	*/

	if (src->packet != AV_NOPTS_VALUE &&
			src->packet > video->vframe.dts_start)
	{
		if ((video->vframe.ic)->has_b_frames == 0)
		{
			if (src->frame != src->packet)
			{
				return -1;
			}
		}

		else if (src->frame < src->packet)
		{
			return -1;
		}

	} /* if (src->packet != AV_NOPTS_VALUE &&
			src->packet > video->vframe.dts_start) */

	i = cff_ffmpeg_video_cut_find_dts(video, src, 0);

	if (i >= 0)
	{
		/* Entry already exists */
		return -1;
	}

	i   = video->cut.num;
	n   = video->cut.max;
	dst = video->cut.list;

	if (i >= n)
	{
		n  += 10;
		dst = av_realloc(dst, n * sizeof(cff_dts_t));

		if (dst == NULL)
		{
			/* Out of memory */
			return -1;
		}

		video->cut.max  = n;
		video->cut.list = dst;
	}

	i   = video->cut.num;
	dst = video->cut.list;

	while (--i >= 0)
	{
		if (src->frame < dst->frame)
		{
			/* Insert - entries one position up (entries == i + 1) */
			memmove(&dst[1], &dst[0], (i + 1) * sizeof(cff_dts_t));
			break;
		}

		dst++;
	}

	*dst = *src;
	video->cut.num++;

	return video->cut.num;

} /* static int cff_ffmpeg_video_cut_add_dts(cff_video_t*, cff_dts_t*) */

int cff_ffmpeg_video_cut_add(cff_video_t *video)
{
	int retv;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

	retv = cff_ffmpeg_video_cut_add_dts(video, &(video->vframe.current));

	if (retv > 0)
	{
		cff_ffmpeg_video_undo_add_current(video, kCutpointAdd);
	}

	return retv;

} /* int cff_ffmpeg_video_cut_add(cff_video_t*) */

static int cff_ffmpeg_video_cut_remove_dts(cff_video_t *video, cff_dts_t *src)
{
	int i;

	i = cff_ffmpeg_video_cut_find_dts(video, src, 1);

	if (i >= 0)
	{
		int		n;
		cff_dts_t	*ptr;

		ptr = &video->cut.list[i];
		n   = video->cut.num - i - 1;

		if (n > 0)
		{
			/* Remove - upper entries one position down */
			memmove(&ptr[0], &ptr[1], n * sizeof(cff_dts_t));
		}

		i = video->cut.num - 1;
		video->cut.num = i;

		/* Clear 'dirty' if list is empty */
		video->dirty = (i > 0) ? 1 : 0;

	} /* if (i >= 0) */

	return i;

} /* static int cff_ffmpeg_video_cut_remove_dts(cff_video_t*, cff_dts_t*) */

int cff_ffmpeg_video_cut_remove(cff_video_t *video)
{
	int retv;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

	retv = cff_ffmpeg_video_cut_remove_dts(video, &(video->vframe.current));

	if (retv >= 0)
	{
		cff_ffmpeg_video_undo_add_current(video, kCutpointRemove);
	}

	return retv;

} /* int cff_ffmpeg_video_cut_remove(cff_video_t*) */

void cff_ffmpeg_video_cut_list_clear(cff_video_t *video)
{
	if (video->cut.list != NULL)
	{
		if (video->cut.num > 0)
		{
			cff_undo_t *undo;

			undo = malloc(sizeof(cff_undo_t) +
				((video->cut.num - 1) * sizeof(cff_dts_t)));

			if (undo != NULL)
			{
				unsigned int i;

				undo->action	= kCutpointRemove;
				undo->num	= video->cut.num;

				i = 0;

				do
				{
					undo->dts[i] = video->cut.list[i];
				}
				while (++i < video->cut.num);

				cff_ffmpeg_video_undo_add(video, undo);

			} /* if (undo != NULL) */

		} /* if (video->cut.num > 0) */

		av_freep(&(video->cut.list));

	} /* if (video->cut.list != NULL) */

	video->cut.num = 0;
	video->cut.max = 0;
	video->dirty   = 0;

} /* void cff_ffmpeg_video_cut_list_clear(cff_video_t*) */

static int64_t cff_ffmpeg_video_cut_put_dts(cff_video_t *video, int64_t dts)
{
	AVStream *is = (video->inFormatCtx)->streams[video->vframe.stream_index];

	if (dts != AV_NOPTS_VALUE)
	{
		dts = av_rescale_q(dts, is->time_base, AV_TIME_BASE_Q);
	}

	return dts;

} /* static int64_t cff_ffmpeg_video_cut_put_dts(cff_video_t*, int64_t) */

int cff_ffmpeg_video_cut_list_save(cff_video_t *video, const char *fname)
{
	int		i;
	FILE		*fd;
	cff_dts_t	*dts;

	i   = video->cut.num;
	dts = video->cut.list;

	if (video->vframe.ic == NULL || i <= 0 || dts == NULL)
	{
		return -1;
	}

#ifdef BUILD_MINGW
	fd = av_fopen_utf8(fname, "w");
#else
	fd = fopen(fname, "w");
#endif

	if (fd == NULL)
	{
		return -1;
	}

	video->dirty = 0;

	fputs("#\n# Cut List (generated from 'CutterFF')\n#\n\n", fd);

	do
	{
		int64_t frame, packet;

		frame  = cff_ffmpeg_video_cut_put_dts(video, dts->frame);
		packet = cff_ffmpeg_video_cut_put_dts(video, dts->packet);

		fprintf(fd, "%lld %lld\n", (long long) frame, (long long) packet);
	}
	while (dts++, --i > 0);

	fclose(fd);

	return 0;

} /* int cff_ffmpeg_video_cut_list_save(cff_video_t*, const char*) */

static char *cff_ffmpeg_video_cut_get_dts(cff_video_t *video, char *line, int64_t *ptr)
{
	char		*end;
	size_t		i;
	int64_t		dts;
	AVStream	*is;

	if ((i = strcspn(line, " \t#\n\r")) == 0)
	{
		/* Empty line or no second entry */
		return (char *) -1;
	}

	line[i]	= '\0';

	dts = (int64_t) strtoll(line, &end, 10);

	if (end[0] != '\0')
	{
		return NULL;
	}

	is = (video->inFormatCtx)->streams[video->vframe.stream_index];

	if (dts != AV_NOPTS_VALUE)
	{
		dts = av_rescale_q(dts, AV_TIME_BASE_Q, is->time_base);
	}

	*ptr = dts;

	return &end[1];

} /* static char *cff_ffmpeg_video_cut_get_dts(cff_video_t*, char*, int64_t*) */

int cff_ffmpeg_video_cut_list_load(cff_video_t *video, const char *fname)
{
	int		retv, num;
	unsigned int	max;
	char		*line, buf[2048];
	FILE		*fd;
	cff_undo_t	*undo, *ptr;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

#ifdef BUILD_MINGW
	fd = av_fopen_utf8(fname, "r");
#else
	fd = fopen(fname, "r");
#endif

	if (fd == NULL)
	{
		return -1;
	}

	max  = 0;
	undo = NULL;

	num  = video->cut.num;
	retv = 0;

	while ((line = fgets(buf, sizeof(buf), fd)) != NULL)
	{
		AVPacket	packet;
		cff_dts_t	list;

		line += strspn(line, " \t");

		line = cff_ffmpeg_video_cut_get_dts(video, line, &(list.frame));

		if (line == (char *) - 1)
		{
			/* Empty line */
			continue;
		}

		if (line == NULL)
		{
			/* DTS not found */
			retv++;
			continue;
		}

		if (cff_ffmpeg_get_packet_with_dts(video, list.frame, &packet) < 0)
		{
			int i = -1;

			if (video->vframe.drain != NULL)
			{
				i = (video->vframe.drain)->num;

				while (--i >= 0)
				{
					if ((video->vframe.drain)->pkt_dts[i] == list.frame)
					{
						break;
					}
				}
			}

			if (i < 0)
			{
				/* DTS is not existing */
				retv++;
				continue;
			}
		}

		else
		{
			av_packet_unref(&packet);
		}

		line = cff_ffmpeg_video_cut_get_dts(video, line, &(list.packet));

		if (line == NULL || line == (char *) - 1)
		{
			/* DTS not found or no second entry */
			retv++;
			continue;
		}

		if (cff_ffmpeg_get_packet_with_dts(video, list.packet, &packet) < 0)
		{
			/* DTS is not existing */
			retv++;
			continue;
		}

		av_packet_unref(&packet);

		if (cff_ffmpeg_video_cut_add_dts(video, &list) < 0)
		{
			/* Frame already stored */
			retv++;
			continue;
		}

		if (undo == NULL || undo->num >= max)
		{
			max += 10;

			ptr = realloc(undo, sizeof(cff_undo_t) +
					((max - 1) * sizeof(cff_dts_t)));

			if (ptr == NULL)
			{
				continue;
			}

			if (undo == NULL)
			{
				ptr->action = kCutpointAdd;
				ptr->num    = 0;
			}

			undo = ptr;
		}

		if (undo != NULL)
		{
			undo->dts[undo->num] = list;
			undo->num++;
		}

	} /* while ((line = fgets(buf, sizeof(buf), fd)) != NULL) */

	fclose(fd);

	video->vframe.packet[0].pos = -1;

	if (undo == NULL)
	{
		if (video->cut.num > num)
		{
			video->dirty = 1;
		}
	}

	else
	{
		if (max > undo->num)
		{
			/* Free unused memory */

			ptr = realloc(undo, sizeof(cff_undo_t) +
				((undo->num - 1) * sizeof(cff_dts_t)));

			if (ptr != NULL)
			{
				undo = ptr;
			}
		}

		cff_ffmpeg_video_undo_add(video, undo);

	} /* if (undo == NULL) .. else */

	if (num == 0)
	{
		video->dirty = 0;
	}

	return retv;

} /* int cff_ffmpeg_video_cut_list_load(cff_video_t*, const char*) */

static int cff_ffmpeg_stream_info_to_buf(AVStream *s, char *buf, int buf_sz)
{
	static const char* const fmt_string[] =
	{
		"%d] %s: %s%s%s",
		"%d] %s: %s%s (%s)",
		"%d] %s: (%s) %s%s",
		"%d] %s: (%s) %s (%s)"
	};

	int			n;
	const char		*codec_type, *language, *codec_name, *profile;
	AVCodecParameters	*codecpar;
	AVDictionaryEntry	*dict;

	dict = av_dict_get(s->metadata, "language", NULL, 0);

	if (dict == NULL)
	{
		language = "";
		n = 0;
	}

	else
	{
		language = dict->value;
		n = 2;
	}

	codecpar = s->codecpar;

	codec_type = av_get_media_type_string(codecpar->codec_type);
	codec_name = avcodec_get_name(codecpar->codec_id);
	profile    = avcodec_profile_name(codecpar->codec_id, codecpar->profile);

	if (codec_type == NULL)
	{
		codec_type = "unknown";
	}

	if (profile == NULL)
	{
		profile	= "";
	}

	else
	{
		n++;
	}

	return snprintf(buf, buf_sz, fmt_string[n],
		s->index, codec_type, language, codec_name, profile);

} /* static int cff_ffmpeg_stream_info_to_buf(AVStream*, char*, int) */

static AVStream *cff_ffmpeg_get_stream_from_index(cff_video_t *video, int index)
{
	int i = index;

	if (video->program.index != NULL)
	{
		const AVProgram	*prg;

		prg = (video->inFormatCtx)->programs[video->program.selected];

		if (index >= prg->nb_stream_indexes)
		{
			return NULL;
		}

		i = prg->stream_index[index];
	}

	else if (index >= (video->inFormatCtx)->nb_streams)
	{
		return NULL;
	}

	return (video->inFormatCtx)->streams[i];

} /* static AVStream *cff_ffmpeg_get_stream_from_index(cff_video_t*, int) */

int cff_ffmpeg_get_stream_info(cff_video_t *video, int index, int *stream_idx, char *buf, int buf_sz)
{
	AVStream *is = cff_ffmpeg_get_stream_from_index(video, index);

	if (is == NULL)
	{
		return -1;
	}

	if (stream_idx != NULL)
	{
		*stream_idx = is->index;
	}

	return cff_ffmpeg_stream_info_to_buf(is, buf, buf_sz);

} /* int cff_ffmpeg_get_stream_info(cff_video_t*, int, int*, char*, int) */

static void cff_ffmpeg_copy_free(cff_video_t *video)
{
	if (video->copy.outFormatCtx != NULL)
	{
		if ((video->copy.outFormatCtx)->pb != NULL)
		{
			/* Close the output file */
			avio_closep(&(video->copy.outFormatCtx->pb));
		}

		avformat_free_context(video->copy.outFormatCtx);

		if (video->copy.streamData != NULL)
		{
			int			i;
			struct StreamData	*st;

			st = video->copy.streamData;
			i  = 0;

			do
			{
				if (st->oStream == NULL)
				{
					/* no more output streams */
					break;
				}

				if ((st->iStream)->index == video->vframe.stream_index)
				{
					(st->iStream)->discard = AVDISCARD_DEFAULT;
				}

				else
				{
					(st->iStream)->discard = AVDISCARD_ALL;
				}

				av_bsf_free(&(st->bsf_ctx));
			}
			while (st++, ++i < video->copy.nb_streams);

			av_free(video->copy.streamData);

		} /* if (video->copy.streamData != NULL) */

	} /* if (video->copy.outFormatCtx != NULL) */

	memset(&(video->copy), '\0', sizeof(cff_copy_t));

	/* Back to last position */
	video->vframe.packet[0].pos = -1;

	cff_ffmpeg_log_callback_flush();

} /* static void cff_ffmpeg_copy_free(cff_video_t*) */

int cff_ffmpeg_copy_cb(cff_video_t *video, double *progress, const char **error)
{
	/* Read frames and copy them to the output file */

	int		i, end, next, retv;
	const char	*str;
	AVPacket	packet;

	av_init_packet(&packet);

	str	= NULL;
	end	= 0;
	next	= 0;
	retv	= 0;
	i	= 100;

	if (video->copy.cutidx == 0)
	{
		/* Start */
		next = 1;
	}

	do
	{
		int			n;
		uint64_t		dts, pts, sz;
		struct StreamData	*st;

		if (next != 0)
		{
			if (video->copy.eof)
			{
				goto _stop;
			}

			/* Get next start frame */

			next = 0;

			if (video->copy.cutidx >= video->cut.num)
			{
				goto _stop;
			}

			video->copy.dts_next = video->cut.list[video->copy.cutidx].packet;
			video->copy.cutidx++;

			if (video->copy.dts_next >= video->vframe.dts_stop)
			{
				goto _stop;
			}

			retv = cff_ffmpeg_get_packet_with_dts(video,
						video->copy.dts_next, &packet);
			if (retv != 0)
			{
				str = gettext_nop("'av_seek_frame()' failed!");
				break;
			}

			/* Get next stop frame */

			if (video->copy.cutidx >= video->cut.num)
			{
				/* Copy till end */
				video->copy.dts_next = AV_NOPTS_VALUE;
			}

			else
			{
				/* Next stop */
				video->copy.dts_next = video->cut.list[video->copy.cutidx].packet;
				video->copy.cutidx++;

				if (video->copy.dts_next >= video->vframe.dts_stop)
				{
					/* Copy till end */
					video->copy.dts_next = AV_NOPTS_VALUE;
				}
			}

			st = video->copy.streamData;
			n  = video->copy.nb_streams;

			do
			{
				if (st->ts_from_duration == 0)
				{
					st->got_offset = 0;
				}

				st++;
			}
			while (--n > 0);
		}

		else
		{
			if (video->copy.eof)
			{
				goto _stop;
			}

			retv = cff_ffmpeg_read_packet(video, &packet, &str);

			if (retv < 0)
			{
_stop:
				video->copy.eof = 1;

				st = video->copy.streamData;
				n  = video->copy.nb_streams;

				do
				{
					if (st->bsf_ctx != NULL && st->bsf_flush != -1)
					{
						goto _bsf;
					}

					st++;
				}
				while (--n > 0);

				retv = av_write_trailer(video->copy.outFormatCtx);

				if (retv < 0 && str == NULL)
				{
					str = gettext_nop("Trailer write failed!");
				}

				end = 1;
				break;

			} /* if (retv < 0) */

			if (packet.stream_index == video->vframe.stream_index &&
					video->copy.dts_next != AV_NOPTS_VALUE &&
						packet.dts > video->copy.dts_next)
			{
				next = 1;
				continue;
			}

		} /* if (next != 0) .. else */

		st = video->copy.streamData;
		n  = video->copy.nb_streams;

		do
		{
			if ((st->iStream)->index == packet.stream_index)
			{
				break;
			}

			st++;
		}
		while (--n > 0);

		if (n == 0)
		{
			continue;
		}

		packet.stream_index = (st->oStream)->index;

		if (st->got_offset == 0)
		{
			st->got_offset = 1;

			if (st->got_start == 0)
			{
				st->got_start = 1;

				dts = (packet.dts == AV_NOPTS_VALUE) ? 0 : packet.dts;

				if (video->copy.ts_keep_start ==
							st->ts_from_duration)
				{
					/*
						Both '0' or '1'
						Either timestamps starts from zero or
						'offset' plus 'duration' gives timestamp
						(which starts from first packet)
					*/

					st->dts_offset = dts;
				}

				else
				{
					/*
						Either 'ts_keep_start != 0'
						  Timestamps starts from first packet
						or 'ts_from_duration != 0'
						  Don't matter ('pkt_dts' not needed)
					*/

					st->pkt_dts = dts;
				}

				if ((st->oStream->codecpar)->codec_type ==
							AVMEDIA_TYPE_VIDEO &&
					(packet.flags & AV_PKT_FLAG_KEY) &&
					strcmp((video->copy.outFormatCtx->oformat)->name, "mp4") == 0)
				{
					/* Format 'MP4' */
					packet.pts = packet.dts;
				}
			}

			else if (packet.dts != AV_NOPTS_VALUE)
			{
				/* 'ts_from_duration == 0' */
				st->dts_offset = packet.dts - st->pkt_dts;

			} /* if (st->got_start == 0) .. else */

		} /* if (st->got_offset == 0) */

		/* PTS */

		pts = packet.pts;

		if (pts == AV_NOPTS_VALUE)
		{
			if (st->ts_from_duration)
			{
				pts = st->duration + st->dts_offset;
			}
		}

		else
		{
			if (st->ts_from_duration)
			{
				pts  = st->duration + st->dts_offset;
				pts += packet.pts - packet.dts;
			}

			else
			{
				pts -= st->dts_offset;
			}

		} /* if (pts == AV_NOPTS_VALUE) .. else */

		packet.pts = pts;

		/* DTS */

		if (st->ts_from_duration)
		{
			dts = st->duration + st->dts_offset;
		}

		else
		{
			if (packet.dts == AV_NOPTS_VALUE)
			{
				dts = st->pkt_dts;
			}

			else
			{
				dts = packet.dts - st->dts_offset;
			}

			st->pkt_dts = dts + packet.duration;

		} /* if (st->ts_from_duration) .. else */

		if ((st->oStream->codecpar)->codec_type == AVMEDIA_TYPE_AUDIO &&
			st->ts_from_duration == 0 && packet.dts == AV_NOPTS_VALUE)
		{
			int duration;

			duration = av_get_audio_frame_duration2((st->iStream)->codecpar, packet.size);

			if (duration == 0)
			{
				duration = (st->iStream->codecpar)->frame_size;
			}

			dts = av_rescale_delta((st->iStream)->time_base,
					dts,
					(AVRational) {1,
						(st->iStream->codecpar)->sample_rate},
					duration,
					&(st->delta_last),
					(st->oStream)->time_base);

		} /* if ((st->oStream->codecpar)->codec_type == AVMEDIA_TYPE_AUDIO && ... */

		packet.dts = dts;

		/* BSF */

		if (st->bsf_ctx != NULL)
		{
_bsf:
			if (st->bsf_flush == 0)
			{
				AVPacket *pktptr = &packet;

				if (video->copy.eof)
				{
					pktptr = NULL;
					st->bsf_flush = 1;
				}

				if (av_bsf_send_packet(st->bsf_ctx, pktptr) < 0)
				{
					if (video->copy.eof)
					{
						st->bsf_flush = -1;
					}

					continue;
				}

			} /* if (st->bsf_flush == 0) */

			if (av_bsf_receive_packet(st->bsf_ctx, &packet) < 0)
			{
				if (video->copy.eof)
				{
					st->bsf_flush = -1;
				}

				continue;
			}

			if (st->bsf_init == 0)
			{
				if (avcodec_parameters_copy((st->oStream)->codecpar,
								(st->bsf_ctx)->par_out) < 0)
				{
					continue;
				}

				st->bsf_init = 1;
			}

		} /* if (st->bsf_ctx != NULL) */

		st->duration += packet.duration;
		sz = packet.size;

		if (st->ts_rescale != 0)
		{
			av_packet_rescale_ts(&packet,
					(st->iStream)->time_base,
						(st->oStream)->time_base);
		}

		retv = av_interleaved_write_frame(video->copy.outFormatCtx, &packet);

		if (retv < 0)
		{
			i = 0;
			str = gettext_nop("'av_interleaved_write_frame()' failed!");
			continue;
		}

		st->nbBytesWritten += sz;
		st->nbPackets++;
	}
	while (av_packet_unref(&packet), next != 0 || --i > 0);

	if (error != NULL)
	{
		*error = str;
	}

	if (retv == 0 && progress != NULL)
	{
		if (end)
		{
			*progress = 1.0;
		}

		else
		{
			*progress = (double)
				avio_tell((video->inFormatCtx)->pb) /
							(double) video->fsize;
		}
	}

	return end | retv;

} /* int cff_ffmpeg_copy_cb(cff_video_t*, double*, const char**) */

const char *cff_ffmpeg_copy_open(cff_video_t *video, const cff_copy_param_t *param)
{
	int			retv;
	unsigned int		i, duration[4];
	char			*str;
	const char		*error, *short_name, *mime_type;
	size_t			bufsz, slen, sz;
	int64_t			streamSize, totalSize;
	double			diffSize;
	AVFormatContext		*inFormatCtx, *outFormatCtx;
	AVOutputFormat		*format;
	AVProgram		*program;
	AVDictionary		*opts;
	struct StreamData	*st;

	if (param->info != NULL)
	{
		*param->info = NULL;
	}

	if (param->fname == NULL || param->fname[0] == '\0')
	{
		return gettext_nop("Missing the file name!");
	}

	if (param->stream_idx == NULL || param->nb_streams == 0 ||
				param->nb_streams > video->nb_streams)
	{
		return gettext_nop("Wrong stream index!");
	}

	video->copy.nb_streams = param->nb_streams;

	sz = sizeof(struct StreamData) * param->nb_streams;
	video->copy.streamData = av_mallocz(sz);

	if (video->copy.streamData == NULL)
	{
		return gettext_nop("Could not allocate memory!");
	}


	/* Open output file */

	inFormatCtx = video->inFormatCtx;
	short_name  = param->format_name;
	mime_type   = NULL;

	if (short_name == NULL)
	{
		short_name = param->fname;
		mime_type  = (inFormatCtx->iformat)->mime_type;
	}

	i = 1;

	do
	{
		if (i == 0 && strcmp((inFormatCtx->iformat)->name, "mpeg") == 0)
		{
			short_name = "vob";
		}

		format = av_guess_format(short_name, param->fname, mime_type);
	}
	while (format == NULL && param->format_name == NULL && --i == 0);

	retv = avformat_alloc_output_context2(&(video->copy.outFormatCtx),
						format, NULL, param->fname);

	if (retv != 0 || video->copy.outFormatCtx == NULL)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("Could not open output context!");
	}

	outFormatCtx = video->copy.outFormatCtx;
	program = (AVProgram *) inFormatCtx->programs;

	if (program != NULL)
	{
		program = av_find_program_from_stream(inFormatCtx, NULL,
						video->vframe.stream_index);
		if (program != NULL)
		{
			AVProgram *dst = av_new_program(outFormatCtx, program->id);

			if (dst != NULL)
			{
				dst->stream_index = av_malloc(sizeof(unsigned int) * param->nb_streams);

				if (dst->stream_index == NULL)
				{
					av_free(dst);
					av_freep(&(outFormatCtx->programs));
					outFormatCtx->nb_programs = 0;
					program = NULL;
				}

				else
				{
					dst->nb_stream_indexes = param->nb_streams;
					dst->program_num = program->program_num;
					dst->pmt_pid = program->pmt_pid;
					dst->pcr_pid = program->pcr_pid;

					av_dict_copy(&(dst->metadata), program->metadata, 0);

					program = dst;

				} /* if (dst->stream_index == NULL) .. else */

			} /* if (dst != NULL) .. else */

		} /* if (program != NULL) */

	} /* if (program != NULL) */

	st = video->copy.streamData;
	i  = 0;

	do
	{
		int			index;
		enum AVCodecID		codec_id;
		enum AVMediaType	codec_type;
		AVCodec			*codec;
		AVCodecParameters	*oCodecPar;
		const cff_media_t	*media;

		index = param->stream_idx[i];

		if (index >= (inFormatCtx)->nb_streams)
		{
			cff_ffmpeg_copy_free(video);
			return gettext_nop("Stream Index greater than number of streams!");
		}

		st->iStream = inFormatCtx->streams[index];

		codec_id = (st->iStream->codecpar)->codec_id;
		codec = avcodec_find_decoder(codec_id);

		if (codec != NULL)
		{
			codec_id = codec->id;
		}

		st->oStream = avformat_new_stream(outFormatCtx, codec);

		if (st->oStream == NULL)
		{
			cff_ffmpeg_copy_free(video);
			return gettext_nop("Could not alloc stream!");
		}

		oCodecPar = (st->oStream)->codecpar;

		retv = avcodec_parameters_copy(oCodecPar, (st->iStream)->codecpar);

		if (retv < 0)
		{
			cff_ffmpeg_copy_free(video);
			return gettext_nop("Could not copy codec parameters!");
		}

		if (av_codec_get_id((outFormatCtx->oformat)->codec_tag,
				(st->iStream->codecpar)->codec_tag) != codec_id)
		{
			unsigned int tag;

			if (av_codec_get_tag2((outFormatCtx->oformat)->codec_tag,
									codec_id, &tag) != 0)
			{
				oCodecPar->codec_tag = tag;
			}
		}

		/*
			Copy the frame rates before calling
			'avformat_transfer_internal_stream_timing_info'.
			See 'ffmpeg/libavformat/utils.c'
		*/

		(st->oStream)->avg_frame_rate = (st->iStream)->avg_frame_rate;
		(st->oStream)->r_frame_rate   = (st->iStream)->r_frame_rate;

		if (avformat_transfer_internal_stream_timing_info(outFormatCtx->oformat,
				st->oStream, st->iStream, AVFMT_TBCF_AUTO) < 0)
		{
			cff_ffmpeg_copy_free(video);
			return gettext_nop("Could not transfer internal stream timing info!");
		}

		if ((st->oStream)->time_base.num <= 0 || (st->oStream)->time_base.den <= 0)
		{
			(st->oStream)->time_base =
				av_add_q(av_stream_get_codec_timebase(st->oStream),
								(AVRational) {0, 1});
		}

		(st->iStream)->discard = AVDISCARD_DEFAULT;

		if (program != NULL && program->stream_index != NULL)
		{
			program->stream_index[i] = (st->oStream)->index;
		}

		(st->oStream)->id = (st->iStream)->id;
		(st->oStream)->disposition = (st->iStream)->disposition;

		/* Clear duration and bitrate */
		(st->oStream)->duration = 0;
		oCodecPar->bit_rate = 0;

		if ((st->iStream)->side_data != NULL && (st->iStream)->nb_side_data != 0)
		{
			int nsd = 0;

			do
			{
				uint8_t			*dst;
				const AVPacketSideData	*src;

				src = &(st->iStream->side_data[nsd]);
				dst = av_stream_new_side_data(st->oStream, src->type, src->size);

				if (dst == NULL)
				{
					cff_ffmpeg_copy_free(video);
					return gettext_nop("Could not copy side data!");
				}

				memcpy(dst, src->data, src->size);
			}
			while (++nsd < (st->iStream)->nb_side_data);

		} /* if ((st->iStream)->side_data != NULL && (st->iStream)->nb_side_data != 0) */

		codec_type = oCodecPar->codec_type;

		if (codec_type == AVMEDIA_TYPE_VIDEO)
		{
			AVRational sar;

			sar = (st->iStream)->sample_aspect_ratio;
			oCodecPar->sample_aspect_ratio = sar;
			(st->oStream)->sample_aspect_ratio = sar;

			media = &(param->media[kCFFMediaVideo]);
		}

		else if (codec_type == AVMEDIA_TYPE_AUDIO)
		{
			if (oCodecPar->codec_id == AV_CODEC_ID_MP3)
			{
				if (oCodecPar->block_align == 1 ||
					oCodecPar->block_align == 576 ||
						oCodecPar->block_align == 1152)
				{
					oCodecPar->block_align = 0;
				}
			}

			else if (oCodecPar->codec_id == AV_CODEC_ID_AC3)
			{
				oCodecPar->block_align = 0;
			}

			media = &(param->media[kCFFMediaAudio]);
		}

		else
		{
			media = &(param->media[kCFFMediaOther]);
		}

		if (media->bsf_name != NULL)
		{
			const AVBitStreamFilter *bsf;

			bsf = av_bsf_get_by_name(media->bsf_name);

			if (bsf != NULL)
			{
				enum AVCodecID		codec_bsf;
				const enum AVCodecID	*id_ptr;

				codec_bsf = codec_id;
				id_ptr = bsf->codec_ids;

				if (id_ptr != NULL)
				{
					while ((codec_bsf = *(id_ptr++)) != codec_id)
					{
						if (codec_bsf == AV_CODEC_ID_NONE)
						{
							break;
						}
					}
				}

				if (codec_bsf == codec_id && av_bsf_alloc(bsf, &(st->bsf_ctx)) == 0)
				{
					if (avcodec_parameters_copy((st->bsf_ctx)->par_in, oCodecPar) < 0)
					{
						cff_ffmpeg_copy_free(video);
						return gettext_nop("Error copying codec parameters to the bitstream filter!");
					}

					(st->bsf_ctx)->time_base_in = (st->oStream)->time_base;

					if (av_bsf_init(st->bsf_ctx) < 0)
					{
						cff_ffmpeg_copy_free(video);
						return gettext_nop("Error initializing the bitstream filter!");
					}

					if (avcodec_parameters_copy(oCodecPar, (st->bsf_ctx)->par_out) < 0)
					{
						cff_ffmpeg_copy_free(video);
						return gettext_nop("Error copying bitstream filter parameters to the codec!");
					}

					(st->oStream)->time_base = (st->bsf_ctx)->time_base_out;

				} /* if (codec_bsf == codec_id && av_bsf_alloc(bsf, &(st->bsf_ctx)) == 0) */

			} /* if (bsf != NULL) */

		} /* if (media->bsf_name != NULL) */

		if (media->ts_from_duration)
		{
			/* Compute timestamps from duration if selected */
			st->ts_from_duration = 1;
		}

		/* Copy stream metadata */
		av_dict_copy(&((st->oStream)->metadata),
					(st->iStream)->metadata, 0);
		st++;
	}
	while (++i < param->nb_streams);

	retv = avio_open(&(outFormatCtx->pb), param->fname, AVIO_FLAG_WRITE);

	if (retv < 0)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("Could not open ouput file!");
	}

	av_dict_copy(&(outFormatCtx->metadata), inFormatCtx->metadata, 0);

	opts = NULL;

	if (param->shortest)
	{
		av_dict_set(&opts, "fflags", "shortest", 0);
	}

	retv = avformat_write_header(outFormatCtx, &opts);

	av_dict_free(&opts);

	if (retv < 0)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("Header write failed!");
	}

	st = video->copy.streamData;
	i  = 0;

	do
	{
		/* Test whether rescaling is needed */
		if ((st->oStream)->time_base.num !=
			(st->iStream)->time_base.num ||
				(st->oStream)->time_base.den !=
					(st->iStream)->time_base.den)
		{
			st->ts_rescale = 1;
		}

		st++;
	}
	while (++i < param->nb_streams);

	if (video->cut.list == NULL || video->cut.num == 0)
	{
		/* First frame */
		retv = cff_ffmpeg_seek(video, video->vframe.dts_start);

		if (retv != 0)
		{
			cff_ffmpeg_copy_free(video);
			return gettext_nop("'av_seek_frame()' failed!");
		}

		/* Copy all */
		video->copy.dts_next = AV_NOPTS_VALUE;
		video->copy.cutidx   = (unsigned int) -1;
	}

	if (param->ts_keep_start &&
		strcmp((outFormatCtx->oformat)->name, "avi") != 0)
	{
		/*
			Keep original start timestamps if selected
			'AVI' starts from zero
		*/

		video->copy.ts_keep_start = 1;
	}

	if (param->func == NULL)
	{
		while ((retv = cff_ffmpeg_copy_cb(video, NULL, &error)) == 0) {;}
	}

	else
	{
		error = NULL;
		retv  = (*param->func)(param->user_data);
	}

	if (retv <= 0 || error != NULL)
	{
		/* Error, or user canceled */
		cff_ffmpeg_copy_free(video);
		return error;
	}

	if (param->info == NULL)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("File successfully copied!");
	}

	bufsz = 4096;
	slen  = 0;
	str   = cff_ffmpeg_printf_concat(NULL, &slen, &bufsz,
				"Output: %s\n\n", param->fname);
	streamSize = 0;

	st = video->copy.streamData;
	i  = 0;

	do
	{
		char buf[256];

		streamSize += st->nbBytesWritten;

		if (cff_ffmpeg_stream_info_to_buf(st->oStream,
							buf, sizeof(buf)) < 0)
		{
			continue;
		}

		/* 'st->duration' uses time base from input stream */
		cff_ffmpeg_dts2time(st->duration, st->iStream, duration);

		str = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
					"   Stream #%s"
					" => Packets:%u Size:%1.0fkB "
					"Duration: %02u:%02u:%02u.%03u\n",
					buf,
					st->nbPackets,
					(double) st->nbBytesWritten / 1024.0,
					duration[0],
					duration[1],
					duration[2],
					duration[3]);
	}
	while (st++, ++i < video->copy.nb_streams);

	totalSize = avio_size((video->copy.outFormatCtx)->pb);

	if (totalSize < 0)
	{
		totalSize = avio_tell((video->copy.outFormatCtx)->pb);

		if (totalSize < 0)
		{
			totalSize = 0;
		}
	}

	if (totalSize > streamSize)
	{
		diffSize = (double) (totalSize - streamSize);
	}

	else
	{
		diffSize = 0.0;
	}

	str = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
		"\n"
		"Size:%1.0fkB Global headers:%1.0fkB "
		"Muxing overhead:%1.2f%%",
			(double) totalSize / 1024.0,
			diffSize / 1024.0,
			(totalSize == 0 || streamSize == 0) ?
			0.0 : 100.0 * (diffSize / (double) streamSize));

	cff_ffmpeg_copy_free(video);

	*(param->info) = cff_ffmpeg_string_truncate(str, slen);

	return NULL;

} /* const char *cff_ffmpeg_copy_open(cff_video_t*, const cff_copy_param_t*) */

const char *cff_ffmpeg_save_image(cff_video_t *video, const cff_image_param_t *param)
{
	int				i, align, retv;
	const char			*error, *mime_type;
	enum AVCodecID			codec_id;
	enum AVPixelFormat		pixfmt;
	const enum AVPixelFormat	*pix_ptr;
	const AVPixFmtDescriptor	*pix_desc;
	AVCodec         		*codec;
	AVStream			*os;
	AVOutputFormat			*format;
	AVCodecContext			*oc;
	AVFormatContext			*outFormatCtx;
	AVFrame				*frame;
	AVDictionary			*opts;
	AVPacket			packet;

	if (param->info != NULL)
	{
		*param->info = NULL;
	}

	if (param->fname == NULL || param->fname[0] == '\0')
	{
		return gettext_nop("Missing the file name!");
	}

	i = 1;
	mime_type = NULL;

	do
	{
		if (i == 0)
		{
			const char *ext;

			if (param->format_name != NULL)
			{
				break;
			}

			ext = strrchr(param->fname, '.');

			if (ext == NULL ||
				(strcmp(&ext[1], "jpg") != 0 &&
					strcmp(&ext[1], "jpeg") != 0))
			{
				break;
			}

			mime_type = "image/jpeg";
		}

		format = av_guess_format(param->format_name, param->fname, mime_type);
	}
	while (format == NULL && --i == 0);

	oc = NULL;

	retv = avformat_alloc_output_context2(&outFormatCtx,
						format, NULL, param->fname);

	if (retv != 0 || outFormatCtx == NULL)
	{
		error = gettext_nop("Could not open output context!");
		goto _done;
	}

	codec_id = format->video_codec;
	codec = avcodec_find_encoder(codec_id);

	if (codec == NULL)
	{
		error = gettext_nop("No image codec found!");
		goto _done;
	}


	oc = avcodec_alloc_context3(NULL);

	if (oc == NULL)
	{
		error = gettext_nop("Couldn't alloc codec context!");
		goto _done;
	}

	if (param->keep_dimension)
	{
		oc->width  = (video->vframe.ic)->width;
		oc->height = (video->vframe.ic)->height;
	}

	else
	{
		oc->width  = param->width;
		oc->height = param->height;
	}

	if (param->keep_ratio)
	{
		cff_ffmpeg_scale_dimension(video, &(oc->width), &(oc->height));
	}

	oc->time_base	= (AVRational) {1, 1};
	pixfmt		= AV_PIX_FMT_NONE;
	pix_ptr		= codec->pix_fmts;

	if (pix_ptr != NULL)
	{
		while ((pixfmt = *(pix_ptr++)) != AV_PIX_FMT_NONE)
		{
			if (param->gray == 0)
			{
				if (pixfmt == (video->vframe.ic)->pix_fmt)
				{
					break;
				}
			}

			else
			{
				pix_desc = av_pix_fmt_desc_get(pixfmt);

				if (strncmp(pix_desc->name, "gray", 4) == 0)
				{
					break;
				}
			}

		} /* while ((pixfmt = *(pix_ptr++)) != AV_PIX_FMT_NONE) */

	} /* if (pix_ptr != NULL) */

	if (pixfmt == AV_PIX_FMT_NONE)
	{
		if (codec->pix_fmts != NULL)
		{
			/* 'avcodec_default_get_format' may return an invalid format */
			pixfmt = codec->pix_fmts[0];
		}

		if (pixfmt == AV_PIX_FMT_NONE)
		{
			pixfmt = avcodec_default_get_format(oc, codec->pix_fmts);
		}

	} /* if (pixfmt == AV_PIX_FMT_NONE) */

	oc->pix_fmt = pixfmt;

	opts = NULL;

	if (param->gray)
	{
		av_dict_set(&opts, "flags", "gray", 0);
	}

	/* Side Aspect Ratio */
	av_dict_set(&opts, "aspect", "1:1", 0);

	retv = avcodec_open2(oc, codec, &opts);

	av_dict_free(&opts);

	if (retv < 0)
	{
		error = gettext_nop("Couldn't open codec!");
		goto _done;
	}

	os = avformat_new_stream(outFormatCtx, codec);

	if (os == NULL)
	{
		error = gettext_nop("Could not alloc stream!");
		goto _done;
	}

	avcodec_parameters_from_context(os->codecpar, oc);

	retv = avio_open(&(outFormatCtx->pb),
				param->fname, AVIO_FLAG_WRITE);
	if (retv < 0)
	{
		error = gettext_nop("Could not open ouput file!");
		goto _done;
	}

	retv = avformat_write_header(outFormatCtx, NULL);

	if (retv < 0)
	{
		error = gettext_nop("Header write failed!");
		goto _done;
	}


	if (pixfmt == (video->vframe.ic)->pix_fmt &&
		oc->width == (video->vframe.ic)->width &&
			oc->height == (video->vframe.ic)->height)
	{
		pixfmt = AV_PIX_FMT_NONE;
		frame  = av_frame_clone(video->vframe.avframe);
	}

	else
	{
		frame = av_frame_alloc();
	}

	if (frame == NULL)
	{
		error = gettext_nop("Couldn't alloc a frame!");
		goto _done;
	}

	if (pixfmt != AV_PIX_FMT_NONE)
	{
		align = 1;
		pix_desc = av_pix_fmt_desc_get(oc->pix_fmt);

		if (pix_desc != NULL &&
			(pix_desc->flags & (AV_PIX_FMT_FLAG_PAL |
						CFF_PIX_FMT_FLAG_PSEUDOPAL)))
		{
			/* Has a palette */
			align = 4;
		}

		retv = cff_ffmpeg_scale_frame(video->vframe.avframe,
					oc->width, oc->height, oc->pix_fmt,
						align, frame->data, frame->linesize);
		if (retv < 0)
		{
			av_frame_free(&frame);
			error = gettext_nop("Could not convert the image!");
			goto _done;
		}

		frame->width     = oc->width;
		frame->height    = oc->height;
		frame->format    = oc->pix_fmt;

	} /* if (pixfmt != AV_PIX_FMT_NONE) */

	frame->pts       = 0;
	frame->pkt_dts   = 0;
	frame->key_frame = 1;

	retv = avcodec_send_frame(oc, frame);
	av_frame_free(&frame);

	if (retv < 0)
	{
		error = gettext_nop("Could not write the frame to the codec!");
		goto _done;
	}

	av_init_packet(&packet);
	retv = avcodec_receive_packet(oc, &packet);

	if (retv < 0)
	{
		if (retv == AVERROR(EAGAIN))
		{
			avcodec_send_frame(oc, NULL);
			retv = avcodec_receive_packet(oc, &packet);
		}

		if (retv < 0)
		{
			error = gettext_nop("Could not read the packet from the codec!");
			goto _done;
		}

	} /* if (retv < 0) */

	retv = av_write_frame(outFormatCtx, &packet);
	av_packet_unref(&packet);

	if (retv < 0)
	{
		error = gettext_nop("'av_write_frame()' failed!");
		goto _done;
	}

	retv = av_write_trailer(outFormatCtx);

	if (retv < 0)
	{
		error = gettext_nop("Trailer write failed!");
		goto _done;
	}

	if (param->info == NULL)
	{
		error = gettext_nop("Image successfully saved!");
	}

	else
	{
		*(param->info) = cff_ffmpeg_file_info_io(outFormatCtx,
							param->fname, 1);
		error = NULL;
	}

_done:
	if (oc)
	{
		avcodec_free_context(&oc);
	}

	if (outFormatCtx != NULL)
	{
		if (outFormatCtx->pb != NULL)
		{
			/* Close the output file */
			avio_closep(&(outFormatCtx->pb));
		}

		avformat_free_context(outFormatCtx);

	} /* if (outFormatCtx != NULL) */

	cff_ffmpeg_log_callback_flush();

	return error;

} /* const char *cff_ffmpeg_save_image(cff_video_t*, const cff_image_param_t*) */

/* EOF */
