/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019 Harald Foerster		*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#ifndef CFF_FFMPEG__H
#define CFF_FFMPEG__H

#include <stdint.h>
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>


/*
	Callback for 'cff_ffmpeg_file_open' or 'cff_ffmpeg_program_open'
	call from your function 'cff_ffmpeg_read_keyframes' till
	the return value is zero.
*/

typedef int (func_cff_callback_t)(void *user_data);

/*
	Callback for 'cff_ffmpeg_log_callback'
	If the return value is not zero, the callback will be stopped.
*/

typedef int (func_log_callback_t)(const char *str, int len, void *user_data);


#define kCFFMediaVideo	0
#define kCFFMediaAudio	1
#define kCFFMediaOther	2
#define kCFFMediaTotal	3
#define kCFFMediaAll	3


typedef struct
{
	const char	*name;
	char		*codecs;

} cff_fmtlist_t;

typedef struct
{
	const char	*name;
	char		*codecs;

} cff_bsflist_t;

enum cff_FrameFlags
{
	kKeyFrameFlag	= 1 << 0,
	kIFrameFlag	= 1 << 1,
	kPFrameFlag	= 1 << 2
};

enum
{
	kVideoFramePrevCut,
	kVideoFramePrev05Sec,
	kVideoFramePrevHalfSec,
	kVideoFramePrevSingle,
	kVideoFrameNextSingle,
	kVideoFrameNextHalfSec,
	kVideoFrameNext05Sec,
	kVideoFrameNextCut,

	kVideoFrameStart,
	kVideoFramePrev10Min,
	kVideoFramePrev05Min,
	kVideoFramePrev01Min,
	kVideoFrameNext01Min,
	kVideoFrameNext05Min,
	kVideoFrameNext10Min,
	kVideoFrameEnd,

	kVideoFrameShiftAdd = kVideoFrameStart
};

struct StreamData
{
	AVStream	*iStream;
	AVStream	*oStream;
	AVBSFContext	*bsf_ctx;
	int		bsf_init;
	int		bsf_flush;
	int		ts_from_duration;
	int		ts_rescale;
	unsigned int	nbPackets;
	uint64_t	nbBytesWritten;
	int64_t		duration;
	int64_t		pkt_dts;
	int64_t		dts_offset;
	int64_t		delta_last;
	int		got_start;
	int		got_offset;
};

typedef struct
{
	AVFormatContext		*outFormatCtx;
	int64_t			dts_next;	/* next stop */
	unsigned int		cutidx;		/* current index of cuttings */
	int			nb_streams;	/* number of streams */
	int			ts_keep_start;	/* keep original start timestamps */
	int			eof;		/* end of file or copying */
	struct StreamData	*streamData;

} cff_copy_t;

/*
	Packet and frame dts may differ, find the packet
	dts by the matching packet and frame pos.
*/

#define kPacketArrayEntries	50

typedef struct
{
	int64_t	dts;
	int64_t	pos;

} cff_packet_t;

typedef struct
{
	int64_t		frame;
	int64_t		packet;

} cff_dts_t;

typedef struct
{
	unsigned int	num;		/* number of entries */
	unsigned int	idx;		/* index of next dts (plus one) */
	int64_t		pkt_dts[1];	/* timestamp of flushed frames */

} cff_drain_t;

#define kUndoArrayEntries	20

enum cff_UndoAction
{
	kCutpointAdd,
	kCutpointRemove
};

typedef struct
{
	enum cff_UndoAction	action;
	unsigned int		num;	/* number of cuttings */
	cff_dts_t		dts[1];	/* undo or redo */

} cff_undo_t;

typedef struct
{
	uint8_t		*data;
	int		width;
	int		height;

} cff_rgb24_t;

typedef struct
{
	AVCodecContext	*ic;
	AVFrame		*avframe;	/* decoded frame */
	cff_rgb24_t	rgb24;		/* rgb24 picture */
	int64_t		io_seek_start;	/* first frame (avio_tell) */
	int64_t		io_seek_end;	/* end of file (avio_tell) */
	unsigned int	stream_index;	/* index of video stream */
	unsigned int	min_gop_size;	/* min gop size (dts duration) */
	unsigned int	max_gop_size;	/* max gop size (dts duration) */
	int64_t		stream_duration;/* duration of stream */
	int64_t		frame_duration;	/* duration per frame */
	cff_dts_t	current;	/* current position */
	int64_t		dts_start;	/* start time */
	int64_t		dts_stop;	/* stop time */
	cff_drain_t	*drain;		/* timestamps in drain mode */

	/* packet 'dts' and 'pos' values */
	cff_packet_t	packet[kPacketArrayEntries];

	cff_undo_t	*undo[kUndoArrayEntries];
	cff_undo_t	*redo[kUndoArrayEntries];

	const char	*pict_type;	/* 'I', 'P', 'B' .. */
	unsigned int	frame_width;	/* width of decoded frame */
	unsigned int	frame_height;	/* height of decoded frame */
	unsigned int	frame_aspect[2];/* aspect ratio of decoded frame */
	unsigned int	frame_total;	/* number of frames */
	unsigned int	frame_current;	/* current decoded frame */
	unsigned int	time_total[4];	/* duration (h:m:s.us) */
	unsigned int	time_current[4];/* time of decoded frame (h:m:s.us) */

} cff_vframe_t;

typedef struct
{
	cff_dts_t	*list;	/* array of cuttings */
	unsigned int	max;	/* max entries of the list */
	unsigned int	num;	/* number of cuttings */

} cff_cut_t;

typedef struct
{
	unsigned int	*index;		/* index list of video programs */
	unsigned int	num;		/* number of programs */
	unsigned int	selected;	/* current program */

} cff_program_t;

typedef struct
{
	AVFormatContext	*inFormatCtx;
	char		*fname;		/* name of video file */
	int64_t		fsize;		/* video file size */
	int		nb_streams;	/* number of streams */
	int		dirty;		/* video or cuttings not saved */
	cff_program_t	program;	/* programs */
	cff_vframe_t	vframe;		/* the video frame */
	cff_cut_t	cut;		/* cuttings */
	cff_copy_t	copy;		/* copy stream(s) */

} cff_video_t;


typedef struct
{
	const char	*bsf_name;		/* filter to use */
	int		ts_from_duration;	/* compute timestamps from duration */

} cff_media_t;

typedef struct
{
	const char		*fname;		/* name of output file */
	const char		*format_name;	/* muxer to use */
	int			*stream_idx;	/* list of stream id's */
	int			nb_streams;	/* number of streams */
	int			shortest;	/* finish with shortest stream duration */
	int			ts_keep_start;	/* keep original start timestamps */
	cff_media_t		media[kCFFMediaTotal];
	char			**info;		/* storage for result messsage */
	func_cff_callback_t	*func;		/* callback */
	void			*user_data;	/* and the data for it */

} cff_copy_param_t;

typedef struct
{
	const char		*fname;		/* name of image file */
	const char		*format_name;	/* the image format */
	int			width;		/* new width of the image */
	int			height;		/* new height of the image */
	int			keep_dimension;	/* keep width and height */
	int			keep_ratio;	/* keep aspect ratio */
	int			gray;		/* gray */
	char			**info;		/* storage for result messsage */

} cff_image_param_t;

extern cff_video_t gVideo;

char *cff_ffmpeg_get_libs_info(void);
char *cff_ffmpeg_get_formats_info(void);
char *cff_ffmpeg_get_codecs_info(void);
char *cff_ffmpeg_get_bitstream_filters_info(void);

cff_fmtlist_t  *cff_ffmpeg_get_format_names(cff_video_t *video);
cff_bsflist_t  *cff_ffmpeg_get_bitstream_filter_names(cff_video_t *video, int media_type);
/* returns number of formats and codecs, '0' if none was found */
int            cff_ffmpeg_get_image_format_names(char ***long_name, char ***format_name);
char*          cff_ffmpeg_get_outformat_extension(const char *format_name, int *slen);

int         cff_ffmpeg_get_current_loglevel_index(void);
const char *cff_ffmpeg_get_loglevel_name(int logindex);
int         cff_ffmpeg_set_loglevel(const char *logname, int lock, func_log_callback_t func, void *user_data);

int cff_ffmpeg_init(const char *logname);

int         cff_ffmpeg_file_open_cb(cff_video_t *video, double *progress, const char **error);
const char *cff_ffmpeg_file_open(cff_video_t *video, const char *fname, func_cff_callback_t func, void *user_data);
void        cff_ffmpeg_file_close(cff_video_t *video);
char       *cff_ffmpeg_file_info(cff_video_t *video);

const char *cff_ffmpeg_program_open(cff_video_t *video, unsigned int index, func_cff_callback_t func, void *user_data);
const char *cff_ffmpeg_get_program_name(cff_video_t *video, unsigned int index, int *prgnum);

int cff_ffmpeg_frame2rgb24(cff_video_t *video, int width, int height);
int cff_ffmpeg_get_frame_next(cff_video_t *video, int next, int width, int height, enum cff_FrameFlags frame_flags);
int cff_ffmpeg_get_frame_slider(cff_video_t *video, double next, int width, int height, enum cff_FrameFlags frame_flags);
int cff_ffmpeg_get_frame_cutpoint(cff_video_t *video, unsigned int next, int width, int height, enum cff_FrameFlags frame_flags);

int cff_ffmpeg_video_undo(cff_video_t *video);
int cff_ffmpeg_video_redo(cff_video_t *video);

/* returns '(unsigned int) -1' on error */
unsigned int cff_ffmpeg_get_cut_frame_number(cff_video_t *video, unsigned int index);

int  cff_ffmpeg_video_cut_find_current(const cff_video_t *video);
int  cff_ffmpeg_video_cut_add(cff_video_t *video);
int  cff_ffmpeg_video_cut_remove(cff_video_t *video);
void cff_ffmpeg_video_cut_list_clear(cff_video_t *video);
int  cff_ffmpeg_video_cut_list_load(cff_video_t *video, const char *fname);
int  cff_ffmpeg_video_cut_list_save(cff_video_t *video, const char *fname);

int cff_ffmpeg_get_stream_info(cff_video_t *video, int index, int *stream_idx, char *buf, int buf_sz);

int         cff_ffmpeg_copy_cb(cff_video_t *video, double *progress, const char **error);
const char *cff_ffmpeg_copy_open(cff_video_t *video, const cff_copy_param_t *param);
const char *cff_ffmpeg_save_image(cff_video_t *video, const cff_image_param_t *param);

#endif /* CFF_FFMPEG__H */
