/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019 Harald Foerster		*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <string.h>
#include <glib/gstdio.h>

#include "cff_macros-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"

/*
#include <gtk/gtk.h>
#include "cff_tools-gtk.h"
*/


#if GTK_MAJOR_VERSION == 2
#define Gui_DIALOG_ICON_INFO		GTK_STOCK_DIALOG_INFO
#define Gui_DIALOG_ICON_WARNING		GTK_STOCK_DIALOG_WARNING
#define Gui_DIALOG_ICON_ERROR		GTK_STOCK_DIALOG_ERROR
#else
#define Gui_DIALOG_ICON_INFO		"dialog-information"
#define Gui_DIALOG_ICON_WARNING		"dialog-warning"
#define Gui_DIALOG_ICON_ERROR		"dialog-error"
#endif

/* global */
cff_windata_t gWindata;


void guitl_window_set_title_filename(GtkWidget *window, const gchar *title)
{
	gchar *str, buf[2048];

	if (title == NULL)
	{
		title = gettext_nop("No video");
		str = cff_charconv_utf8_gettext(title);
	}

	else
	{
		str = cff_charconv_utf8_from_filename(title);
	}

	snprintf(buf, sizeof(buf), "CutterFF: %s", str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	gtk_window_set_title(GTK_WINDOW(window), buf);

} /* void guitl_window_set_title_filename(GtkWidget*, const gchar*) */

static void guitl_window_set_title_gettext(GtkWidget *window, const gchar *title)
{
	gchar *str;

	if (title != NULL)
	{
		str = cff_charconv_utf8_gettext(title);
		gtk_window_set_title(GTK_WINDOW(window), str);
		CFF_CHARCONV_UTF8_FREE(str, title);
	}

} /* static void guitl_window_set_title_gettext(GtkWidget*, const gchar*) */

static gint guitl_window_focus_in(GtkWidget *window,
				GdkEventFocus *e, cff_windata_t *windat)
{
	windat->topwin = window;

	return FALSE;

} /* static gint guitl_window_focus_in(GtkWidget*, GdkEventFocus*, windata_t*) */

GtkWidget *guitl_window_new(GtkWidget *child, gint border)
{
	GtkWidget *window;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_add(GTK_CONTAINER(window), child);

	if (border > 0)
	{
		gtk_container_set_border_width(GTK_CONTAINER(child), border);
	}

	GuiSignalConnect(window, "focus-in-event", guitl_window_focus_in, &gWindata);

	return window;

} /* GtkWidget *guitl_window_new(GtkWidget*, gint) */

GtkWidget *guitl_window_new_gettext(const gchar *title, GtkWidget *child, gint border)
{
	GtkWidget *window;

	window = guitl_window_new(child, border);
	guitl_window_set_title_gettext(window, title);

	return window;

} /* GtkWidget *guitl_window_new_gettext(const gchar*, GtkWidget*, gint) */

static void guitl_window_resize(GtkWidget *w, GtkAllocation *sallocation, gpointer data)
{
	/*
		either main window, info dialogs or filechooser
		-----------------------------------------------
		kPropIntMainWinWidth         + kPropIntMainWinHeight
		kPropIntFileSaveWinWidth     + kPropIntFileSaveWinHeight
		kPropIntFileImageWinWidth    + kPropIntFileImageWinHeight
		kPropIntFileInfoWinWidth     + kPropIntFileInfoWinHeight
		kPropIntHelpFFmpegWinWidth   + kPropIntHelpFFmpegWinHeight
		kPropIntHelpCutterFFWinWidth + kPropIntHelpCutterFFWinHeight
		kPropIntLogWinWidth          + kPropIntLogWinHeight
	*/

	int prop = GPOINTER_TO_INT(data);

	cff_gprop_set_int(prop, sallocation->width);
	cff_gprop_set_int(prop + 1, sallocation->height);

} /* static void guitl_window_resize(GtkWidget *w, GtkAllocation*, gpointer) */

void guitl_window_setup(GtkWidget *widget, gint prop)
{
	gint		width, height;
	GtkWindow	*window;

	window = GTK_WINDOW(widget);

	if ( (cff_gprop_get_int(prop, &width) == FALSE || width <= 0) ||
		(cff_gprop_get_int(prop + 1, &height) == FALSE || height <= 0) ||
			guitl_window_test_size(window, width, height) == FALSE )
	{
		width  = -1;
		height = -1;
	}

	gtk_window_set_resizable(window, TRUE);
	gtk_window_set_default_size(window, width, height);

	GuiSignalConnect(window, "size-allocate", guitl_window_resize, GINT_TO_POINTER(prop));

} /* void guitl_window_setup(GtkWidget*, gint) */

gboolean guitl_window_test_size(GtkWindow *window, gint width, gint height)
{
	GdkScreen *screen = gtk_window_get_screen(window);

	if (gdk_screen_get_width(screen) < width ||
			gdk_screen_get_height(screen) < height)
	{
		return FALSE;
	}

	return TRUE;

} /* guitl_window_test_size(GtkWindow *window, gint, gint) */

GtkWidget *guitl_scrolled_window_new(GtkWidget *container, gint wmin, gint hmin, gint policy)
{
	GtkWidget		*widget;
	GtkScrolledWindow	*swin;

	widget = gtk_scrolled_window_new(NULL, NULL);
	gtk_widget_set_size_request(widget, wmin, hmin);

	swin = GTK_SCROLLED_WINDOW(widget);
	gtk_scrolled_window_set_policy(swin, policy, policy);

#if GTK_MAJOR_VERSION > 2
	gtk_scrolled_window_set_shadow_type(swin, GTK_SHADOW_IN);
	gtk_widget_set_hexpand(widget, TRUE);
	gtk_widget_set_vexpand(widget, TRUE);
#endif

	if (container != NULL)
	{
		gtk_container_add(GTK_CONTAINER(container), widget);
	}

	return widget;

} /* GtkWidget *guitl_scrolled_window_new(GtkWidget*, gint, gint, gint) */

static void guitl_dialog_modal_destroy(GtkWidget *w, GtkWidget **menu)
{
	if (menu != NULL && *menu != NULL)
	{
		GuiWidgetSetState(*menu, Gui_STATE_NORMAL);
	}

} /* static void guitl_dialog_modal_destroy(GtkWidget*, GtkWidget**) */

void guitl_dialog_set_modal(GtkWidget *dialog, GtkWidget *parent, GtkWidget **menu)
{
	if (menu != NULL && *menu != NULL)
	{
		GuiWidgetSetState(*menu, Gui_STATE_PRELIGHT);
	}

	GuiDialogSetModal(dialog, parent);
	GuiSignalConnect(dialog, "destroy", guitl_dialog_modal_destroy, menu);

} /* void guitl_dialog_set_modal(GtkWidget*, GtkWidget*, GtkWidget**) */

static gint guitl_dialog_run(GtkDialog *dialog, const gchar *icon, const gchar *text, GtkWidget **mbar)
{
	gint		retv;
	GtkWidget	*content, *frame, *hbox, *image, *label;

	if (mbar)
	{
		GuiWidgetSetState(*mbar, Gui_STATE_PRELIGHT);
	}

	hbox = GuiHBoxNew(FALSE, 5);
	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);

	label = guitl_label_new_gettext(text);
	GuiMiscSetPadding(label, 15);
	image = GuiImageGetDialogIcon(icon);
	gtk_container_add(GTK_CONTAINER(hbox), image);
	gtk_container_add(GTK_CONTAINER(hbox), label);
	gtk_widget_show_all(frame);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), frame);

	retv = gtk_dialog_run(dialog);

	gtk_widget_destroy(GTK_WIDGET(dialog));

	if (mbar)
	{
		GuiWidgetSetState(*mbar, Gui_STATE_NORMAL);
	}

	return retv;

} /* static gint guitl_dialog_run(GtkDialog*, const gchar *, const gchar*, GtkWidget**) */

static void guitl_dialog_icon(const gchar *icon, const gchar *str, GtkWidget **mbar)
{
	static const gchar okay_str[] = gettext_nop("OK");

	gchar		*accept, *text;
	GtkWidget	*widget;

	accept = cff_charconv_utf8_gettext(okay_str);

	widget = gtk_dialog_new_with_buttons("CutterFF",
		GTK_WINDOW(gWindata.topwin),
		GTK_DIALOG_MODAL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	CFF_CHARCONV_UTF8_FREE(accept, okay_str);

	text = cff_charconv_utf8_gettext(str);
	guitl_dialog_run(GTK_DIALOG(widget), icon, text, mbar);
	CFF_CHARCONV_UTF8_FREE(text, str);

} /* static void guitl_dialog_icon(const gchar*, const gchar*, GtkWidget**) */

void guitl_dialog_info(const gchar *text, GtkWidget **mbar)
{
	guitl_dialog_icon(Gui_DIALOG_ICON_INFO, text, mbar);

} /* void guitl_dialog_info(const gchar*, GtkWidget**) */

void guitl_dialog_error(const gchar *text, GtkWidget **mbar)
{
	guitl_dialog_icon(Gui_DIALOG_ICON_ERROR, text, mbar);

} /* void guitl_dialog_error(const gchar*, GtkWidget**) */

void guitl_dialog_warning(const gchar *text, GtkWidget **mbar)
{
	guitl_dialog_icon(Gui_DIALOG_ICON_WARNING, text, mbar);

} /* void guitl_dialog_warning(const gchar*, GtkWidget**) */

gint guitl_dialog_warning_2btn(const gchar *text, const gchar *btn_no, const gchar *btn_yes, GtkWidget **mbar)
{
	gint		retv;
	GtkWidget	*widget;

	widget = guitl_dialog_new_gettext_2btn("CutterFF", btn_no, btn_yes, gWindata.topwin);
	retv   = guitl_dialog_run(GTK_DIALOG(widget), Gui_DIALOG_ICON_WARNING, text, mbar);

	if (retv != GTK_RESPONSE_ACCEPT)
	{
		return GTK_RESPONSE_CANCEL;
	}

	return retv;

} /* gint guitl_dialog_warning_2btn(const gchar*, const gchar*, const gchar*, GtkWidget**) */

GtkWidget *guitl_dialog_new_gettext_2btn(const gchar *title, const gchar *btn_no, const gchar *btn_yes, GtkWidget *parent)
{
	gchar		*accept, *cancel, *title_utf8;
	GtkWidget	*widget;

	cancel     = cff_charconv_utf8_gettext(btn_no);
	accept     = cff_charconv_utf8_gettext(btn_yes);
	title_utf8 = cff_charconv_utf8_gettext(title);

	widget = gtk_dialog_new_with_buttons(title_utf8,
		GTK_WINDOW(parent),
		GTK_DIALOG_MODAL,
		cancel, GTK_RESPONSE_CANCEL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	CFF_CHARCONV_UTF8_FREE(title_utf8, title);
	CFF_CHARCONV_UTF8_FREE(accept, btn_yes);
	CFF_CHARCONV_UTF8_FREE(cancel, btn_no);

	return widget;

} /* GtkWidget *guitl_dialog_new_gettext_2btn(const gchar*, const gchar*, const gchar*, GtkWidget*) */

GtkWidget *guitl_label_new_gettext(const gchar *str)
{
	gchar		*s;
	GtkWidget	*label;

	s = cff_charconv_utf8_gettext(str);
	label = gtk_label_new(s);
	CFF_CHARCONV_UTF8_FREE(s, str);

	return label;

} /* GtkWidget *guitl_label_new_gettext(const gchar*) */

GtkWidget *guitl_menu_item_new_gettext(const gchar *label)
{
	gchar		*str;
	GtkWidget	*mi;

	str = cff_charconv_utf8_gettext(label);
	mi = gtk_menu_item_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, label);

	return mi;

} /* GtkWidget *guitl_menu_item_new_gettext(const gchar*) */

GtkWidget *guitl_check_menu_item_new_gettext(const gchar *label)
{
	gchar		*str;
	GtkWidget	*mi;

	str = cff_charconv_utf8_gettext(label);
	mi = gtk_check_menu_item_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, label);

	return mi;

} /* GtkWidget *guitl_check_menu_item_new_gettext(const gchar*) */

GtkWidget *guitl_frame_new_add_child_gettext(const gchar *title, GtkWidget *child)
{
	gchar		*str;
	GtkWidget	*frame;

	str = cff_charconv_utf8_gettext(title);
	frame = gtk_frame_new(str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	gtk_container_add(GTK_CONTAINER(frame), child);

	return frame;

} /* GtkWidget *guitl_frame_new_add_child_gettext(const gchar*, GtkWidget*) */

GtkWidget *guitl_check_button_new_gettext(const gchar *label)
{
	gchar		*str;
	GtkWidget	*button;

	str = cff_charconv_utf8_gettext(label);
	button = gtk_check_button_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, label);

	return button;

} /* GtkWidget *guitl_check_button_new_gettext(const gchar*) */

GtkWidget *guitl_vbox_add_check_button_gettext(GtkWidget *vbox, const gchar *label)
{
	GtkWidget *button, *hbox;

	hbox = GuiHBoxNew(TRUE, 2);

	gtk_container_set_border_width(GTK_CONTAINER(hbox), 2);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 2);

	button = guitl_check_button_new_gettext(label);

	GuiHBoxPackStart(hbox, button, TRUE, FALSE, 2);

	return button;

} /* GtkWidget *guitl_vbox_add_check_button_gettext(GtkWidget*, const gchar*) */

GtkWidget *guitl_hbox_add_button(GtkWidget *hbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = gtk_button_new_with_label(text);

	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);
	GuiSignalConnect(button, "clicked", func, data);

	return button;

} /* GtkWidget *guitl_hbox_add_button(GtkWidget*, const gchar*) */

GtkWidget *guitl_vbox_add_button(GtkWidget *vbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = gtk_button_new_with_label(text);

	GuiVBoxPackStart(vbox, button, FALSE, FALSE, 1);
	GuiSignalConnect(button, "clicked", func, data);

	return button;

} /* GtkWidget *guitl_hbox_add_button(GtkWidget*, const gchar*) */

static GtkWidget *guitl_button_gettext(const gchar *text, GCallback func, gpointer data)
{
	gchar		*str;
	GtkWidget	*button;

	str = cff_charconv_utf8_gettext(text);
	button = gtk_button_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, text);

	if (func != NULL)
	{
		GuiSignalConnect(button, "clicked", func, data);
	}

	return button;

} /* static GtkWidget *guitl_button_gettext(const gchar*, GCallback, gpointer) */

GtkWidget *guitl_hbox_add_button_gettext(GtkWidget *hbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = guitl_button_gettext(text, func, data);

	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);

	return button;

} /* GtkWidget *guitl_hbox_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */

GtkWidget *guitl_vbox_add_button_gettext(GtkWidget *vbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = guitl_button_gettext(text, func, data);

	GuiVBoxPackStart(vbox, button, FALSE, FALSE, 1);

	return button;

} /* GtkWidget *guitl_hbox_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */

GtkWidget *guitl_hbutton_box_new(GtkWidget *parent, gint spacing, GtkButtonBoxStyle layout)
{
	GtkWidget *hbutton_box;

#if GTK_MAJOR_VERSION == 2
	if (GTK_IS_HBUTTON_BOX(parent))
#else
	if (GTK_IS_BUTTON_BOX(parent))
#endif
	{
		hbutton_box = parent;
	}

	else
	{
		hbutton_box = GuiHButtonBoxNew();

#if GTK_MAJOR_VERSION == 2
		if (GTK_IS_VBOX(parent))
		{
			gtk_box_pack_start(GTK_BOX(parent), hbutton_box, FALSE, FALSE, 0);
		}
		else
#endif
		{
			gtk_container_add(GTK_CONTAINER(parent), hbutton_box);
		}
	}

	GuiButtonBoxSetSpacing(hbutton_box, spacing);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbutton_box), layout);

	return hbutton_box;

} /* GtkWidget *guitl_hbutton_box_new(GtkWidget*, gint, GtkButtonBoxStyle) */

GtkWidget *guitl_button_box_add_button(GtkWidget *button_box, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = guitl_button_gettext(text, func, data);

	gtk_container_add(GTK_CONTAINER(button_box), button);
	gtk_container_set_border_width(GTK_CONTAINER(button), 2);

	GuiWidgetCanDefault(button);

	return button;

} /* GtkWidget *guitl_button_box_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */


#if GTK_MAJOR_VERSION > 2

GtkWidget *guitl_hgrid_new(gboolean homogeneous, gint spacing)
{
	GtkWidget *hgrid = gtk_grid_new();

	gtk_grid_set_column_homogeneous(GTK_GRID(hgrid), homogeneous);
	gtk_grid_set_column_spacing(GTK_GRID(hgrid), spacing);

	return hgrid;

} /* GtkWidget *guitl_hgrid_new(gboolean, gint) */

GtkWidget *guitl_vgrid_new(gboolean homogeneous, gint spacing)
{
	GtkWidget *vgrid = gtk_grid_new();

	gtk_grid_set_row_homogeneous(GTK_GRID(vgrid), homogeneous);
	gtk_grid_set_row_spacing(GTK_GRID(vgrid), spacing);
	gtk_orientable_set_orientation(GTK_ORIENTABLE(vgrid),
						GTK_ORIENTATION_VERTICAL);
	return vgrid;

} /* GtkWidget *guitl_vgrid_new(gboolean, gint) */

#endif /* GTK_MAJOR_VERSION > 2 */

/* Misc */

int tl_is_ascii_str(const char *str)
{
	if (str != NULL)
	{
		unsigned int		c;
		const unsigned char	*p;

		p = (const unsigned char *) str;

		while ((c = *(p++)) != '\0')
		{
			if (c < 0x20 || c > 0x7e)
			{
				return FALSE;
			}
		}
	}

	return TRUE;

} /* int tl_is_ascii_str(const char*) */

char *tl_replace_non_ascii_chars_in_str(char *str)
{
	if (str != NULL)
	{
		unsigned int	chr;
		unsigned char	*ptr;

		ptr = (unsigned char *) str;

		while ((chr = *(ptr++)) != '\0')
		{
			if (chr < 0x08 ||
				(chr > 0x0D && chr < 0x20) ||
							chr > 0x7e)
			{
				ptr[-1] = '?';
			}
		}
	}

	return str;

} /* char *tl_replace_non_ascii_chars_in_str(char*) */

int tl_file_exists(const char *path)
{
	if (path != NULL)
	{
		/* Open file for reading and writing */

		FILE *fd = g_fopen(path, "r+");

		if (fd)
		{
			/* File do exist */
			fclose(fd);
			return 1;
		}
	}

	/* File do not exist */
	return 0;

} /* int tl_file_exists(const char*) */

const char *tl_get_filename(const char *path)
{
	if (path != NULL)
	{
		const char *sep = strrchr(path, '/');

#ifdef BUILD_MINGW
		const char *s1 = strrchr(path, '\\');

		if (s1 > sep)
		{
			sep = s1;
		}
#endif
		if (sep != NULL)
		{
			return &sep[1];
		}
	}

	return path;

} /* const char *tl_get_filename(const char*) */


#ifdef BUILD_MINGW

void file_path_dos_to_unix(char *path)
{
	if (path != NULL)
	{
		while ((path = strchr(path, '\\')) != NULL)
		{
			*(path++) = '/';
		}
	}

} /* void file_path_dos_to_unix(char*) */

void file_path_unix_to_dos(char *path)
{
	if (path != NULL)
	{
		while ((path = strchr(path, '/')) != NULL)
		{
			*(path++) = '\\';
		}
	}

} /* void file_path_unix_to_dos(char*) */

#endif /* BUILD_MINGW */

/* EOF */
