/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019 Harald Foerster		*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include <gdk/gdkkeysyms.h>
#include <glib/gstdio.h>

#include "cff_macros-gtk.h"
#include "cff_accel-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"
#include "cff_ffmpeg.h"

#ifdef HAVE_SYS_NPROCESSORS

#ifdef BUILD_MINGW
#include <windows.h>
#elif _MACOS
#include <sys/param.h>
#include <sys/sysctl.h>
#else
#include <unistd.h>
#endif

#endif /* HAVE_SYS_NPROCESSORS */

/*
#include <stdio.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include "cff_tools-gtk.h"
*/

#ifndef BUILD_MINGW
/* application icon */
#include "../cutterff.xpm"
#endif

#if GTK_MAJOR_VERSION > 2
#ifndef HAVE_GTKOVERWRITECONFIRMATION
#define HAVE_GTKOVERWRITECONFIRMATION	1
#endif
#endif /* GTK_MAJOR_VERSION > 2 */

#ifdef BUILD_MINGW
#define CFF_IS_DIR_SEPARATOR(chr)	((chr) == '/' || (chr) == '\\')
#else
#define CFF_IS_DIR_SEPARATOR(chr)	((chr) == '/')
#endif


typedef struct
{
	GtkWidget	*some;
	GtkWidget	*all;
	cff_ListStore_t	list;

} cff_streamlist_t;

typedef struct
{
	GtkWidget	*ts_from_duration;
	GtkWidget	*bsf_option;
	cff_bsflist_t	*bsf_list;

} cff_mediaopts_t;

typedef struct
{
	GtkWidget		*window;
	GtkEntry		*entry;

	struct
	{
		GtkWidget	*option;
		gchar		*extension_orig;

		union
		{
			cff_fmtlist_t	*list;   /* Video */
			gchar		**names; /* Image */
		} u;

	} format;

	struct
	{
		GtkWidget	*widget;
		guint		bits;

	} okay;

	GtkWidget		**mbar;
	cff_video_t		*video;
	gchar			*filechooser_utf8str;
	gint			path_prop;

} cff_save_common_t;

typedef struct
{
	cff_save_common_t	common;
	GtkWidget		*up;
	GtkWidget		*down;
	GtkWidget		*shortest;
	GtkWidget		*ts_keep_start;
	cff_streamlist_t	avail;
	cff_streamlist_t	out;
	cff_mediaopts_t		media[kCFFMediaTotal];

} cff_save_video_t;

typedef struct
{
	GtkSpinButton		*widget;
	gint			prev_value;

} cff_spinbutton_t;

typedef struct
{
	cff_save_common_t	common;
	cff_spinbutton_t	width;
	cff_spinbutton_t	height;
	GtkWidget		*keep_dimension;
	GtkWidget		*keep_ratio;
	GtkWidget		*gray;

} cff_save_image_t;

typedef struct
{
	int (*func)(cff_video_t*, double*, const char**);

	cff_video_t		*video;
	const gchar		*title;
	GtkWidget		**mbar;
	GuiMutex		*mutex;
	GtkProgressBar		*progress_bar;
	const gchar		*error;
	gint			retv;
	gboolean		is_thread;
	volatile gboolean	is_locked;

} cff_progressbar_t;

typedef struct
{
	GtkWidget	*window;
	GtkWidget	*textview;
	GtkTextTag	*tag;

	struct
	{
		gchar	*str;
		guint	len;
		guint	max;

	} utf8;

	guint		nThreads;
	gboolean	is_mainthread;
	gboolean	is_progressbar;
	guint		sigid;
	GuiMutex	mutex;

#if GTK_MAJOR_VERSION == 2
	GdkColor	bg;
#endif

} cff_logwindow_t;

static cff_logwindow_t gLogWindow = {NULL};

static const gchar gLogWinTextColor[] = "#00EE00";
static const gchar gLogWinBackgroundColor[] = "#000000";


/* Open, Program or Save */
static gboolean gIsBusy = FALSE;

/* Ignore these events if busy */
#define kButtonPressEvent	0
#define kKeyPressEvent		1

/* Filter */

#define kFileFilterEntriesMax	10
#define kFileFilterCharsMax	15

typedef struct
{
	GtkFileChooser	*chooser;
	GtkWidget	*btn_add;
	GtkWidget	*btn_remove;
	GtkFileFilter	*buf[kFileFilterEntriesMax];
	guint		num;
	const gchar	*ext;
	gint		prop;

} cff_filefilter_t;

static const gchar cff_video_ext[] = "* avi mov mp4 mpg mpeg mkv ts webm";
static const gchar cff_image_ext[] = "* jpg";



/*
	------------
	 PROTOTYPES
	------------
*/

/* File Menu */
static void cff_gui_file_open(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_save(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_image(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_close(GtkWidget *w, gpointer data);
static void cff_gui_file_info(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_save_rc(GtkWidget *w, gpointer data);
static void cff_gui_file_quit(GtkWidget *w, GtkWidget **mbar);

/* Edit Menu */
static void cff_gui_edit_undo(GtkWidget *w, cff_video_t *video);
static void cff_gui_edit_redo(GtkWidget *w, cff_video_t *video);
static void cff_gui_edit_load(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_edit_save(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_edit_clear(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_edit_add(GtkWidget *w, cff_video_t *video);
static void cff_gui_edit_remove(GtkWidget *w, cff_video_t *video);

/* Submenu for frame numbers */
static void cff_gui_build_frame_submenu(GtkWidget *submenu, gpointer data);

/* Submenu for programs */
static void cff_gui_build_program_submenu(GtkWidget *submenu, gpointer data);

/* Submenu for loglevels */
static void cff_gui_build_loglevel_submenu(GtkWidget *submenu, gpointer data);

/* Options Menu */
static void cff_gui_opt_toggle_bool(GtkWidget *w, gpointer data);

/* Help Menu */
static void cff_gui_help_ffmpeg(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_help_cutterff(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_help_about(GtkWidget *w, GtkWidget **mbar);

/* called from cff_gaccel.c and cff_gui-gtk.c */
void cff_gui_adjust_menu(GtkWidget *w, gpointer data);

/* cff_main.c */
extern char *cff_main_save_rc(void);


/*
	--------
	 CURSOR
	--------
*/

typedef struct
{
	GdkWindow	*window;
	GdkCursor	*busy;

} cff_cursor_t;

static cff_cursor_t gCursor = { NULL };


/*
	-------------
	 VIDEO FRAME
	-------------
*/

static GtkWidget *gDrawingAreaFrame;
static GtkWidget *gDrawingAreaBar;
static GtkWidget *gDrawingAreaSlider;
static GtkLabel  *gLabelFrame;
static GtkLabel  *gLabelTime;

enum
{
	kColorBlack,
	kColorDarkGrey,
	kColorGrey,
	kColorGreen,
	kColorRed,
	kColorTotal
};

#if GTK_MAJOR_VERSION == 2

static const char cff_configure_event[] = "configure-event";
static const char cff_expose_event[] = "expose-event";

static GdkGC *gContext[kColorTotal];

#else

static const char cff_configure_event[] = "size-allocate";
static const char cff_expose_event[] = "draw";

static struct
{
	cairo_surface_t	*surfaceRGB24;
	cairo_surface_t	*surfaceFrame;
	cairo_surface_t	*surfaceBar;
	cairo_surface_t	*surfaceSlider;
	gint		rgb24_width;
	gint		rgb24_height;
	GdkRGBA		RGBA[kColorTotal];

} gCairo;

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


/*
	----------------
	 BUTTON WIDGETS
	----------------
*/

enum
{
	kButtonVideoFramePrev4,
	kButtonVideoFramePrev3,
	kButtonVideoFramePrev2,
	kButtonVideoFramePrev1,
	kButtonVideoFrameNext1,
	kButtonVideoFrameNext2,
	kButtonVideoFrameNext3,
	kButtonVideoFrameNext4,
	kButtonVideoFrameTotal
};

static gchar *gButtonVideoFrameLabel[] =
{
	"|<",
	"<<<",
	"<<",
	"<",
	">",
	">>",
	">>>",
	">|"
};

static GtkWidget *gui_button_video_frame[kButtonVideoFrameTotal];

enum
{
	kButtonVideoCutAdd,
	kButtonVideoCutRemove,
	kButtonVideoCutTotal
};

static GtkWidget *gui_button_video_cut[kButtonVideoCutTotal];


/*
	--------------
	 MENU WIDGETS
	--------------
*/

enum
{
	kMenuBarFile,
	kMenuBarEdit,
	kMenuBarOption,
	kMenuBarHelp,
	kMenuBarTotal
};

/* enbled or disabled */

enum
{
	kMenuFileSave,
	kMenuFileImage,
	kMenuFileClose,
	kMenuFileInfo,
	kMenuEditUndo,
	kMenuEditRedo,
	kMenuEditLoad,
	kMenuEditSave,
	kMenuEditClear,
	kMenuEditAdd,
	kMenuEditRemove,
	kMenuEditFrame,
	kMenuEditProgram,
	kMenuItemsTotal
};

enum
{
	kPopupFileSave,
	kPopupEditUndo,
	kPopupEditRedo,
	kPopupEditLoad,
	kPopupEditSave,
	kPopupEditClear,
	kPopupEditFrame,
	kPopupItemsTotal
};

static GtkWidget *cff_gui_menu_bar[kMenuBarTotal];
static GtkWidget *cff_gui_menu_widget[kMenuItemsTotal];
static GtkWidget *cff_gui_popup_widget[kPopupItemsTotal];

/* kPropBoolOptKeyFrame, kPropBoolOptIFrame and kPropBoolOptPFrame */
static GtkWidget *cff_gui_checkmenu_frame_widget[3];
static GtkWidget *cff_gui_popup_checkmenu_frame_widget[3];

static GtkWidget *cff_gui_popup_menu;


/*
	------------
	 MENU TYPES
	------------
*/

enum
{
	kMTypeEndOfList,
	kMTypeItem,
	kMTypeCheckItem,
	kMTypeSeparator,
	kMTypeSubmenuCheckItem
};


/*
	------------
	 MENU ITEMS
	------------
*/

static const gchar mn_file_open_str[]	= gettext_nop("Open ...");
static const gchar mn_file_save_str[]	= gettext_nop("Save ...");
static const gchar mn_edit_undo_str[]	= gettext_nop("Undo");
static const gchar mn_edit_redo_str[]	= gettext_nop("Redo");
static const gchar mn_edit_load_str[]	= gettext_nop("Load list ...");
static const gchar mn_edit_save_str[]	= gettext_nop("Save list ...");
static const gchar mn_edit_clear_str[]	= gettext_nop("Clear list ...");
static const gchar mn_edit_frame_str[]	= gettext_nop("Goto");
static const gchar mn_opt_keyframe_str[]= gettext_nop("Key-Frame");
static const gchar mn_opt_iframe_str[]	= gettext_nop("I-Frame");
static const gchar mn_opt_pframe_str[]	= gettext_nop("P-Frame");

typedef struct
{
	gint		type;		/* see above */
	const gchar	*label;		/* name to display */
	gint		accel_id;	/* accel key */
	GtkWidget	**storage;	/* cff_gui_menu_widget */
	GuiSignalFunc	callback;	/* function pointer */
	gpointer	data;		/* data for it */

} cff_guiMenuItem_t;

static const cff_guiMenuItem_t cff_gui_menuFile[] =
{
	{kMTypeItem, mn_file_open_str,
		kGAccelFileOpen, NULL,
		(GuiSignalFunc) cff_gui_file_open, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, mn_file_save_str,
		kGAccelFileSave, &cff_gui_menu_widget[kMenuFileSave],
		(GuiSignalFunc) cff_gui_file_save, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, gettext_nop("Save image ..."),
		kGAccelFileImage, &cff_gui_menu_widget[kMenuFileImage],
		(GuiSignalFunc) cff_gui_file_image, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Close"),
		kGAccelFileClose, &cff_gui_menu_widget[kMenuFileClose],
		(GuiSignalFunc) cff_gui_file_close, NULL},
	{kMTypeItem, gettext_nop("Info ..."),
		kGAccelFileInfo, &cff_gui_menu_widget[kMenuFileInfo],
		(GuiSignalFunc) cff_gui_file_info, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Save rc file"),
		kGAccelFileSaveRC, NULL,
		(GuiSignalFunc) cff_gui_file_save_rc, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Quit"),
		kGAccelFileQuit, NULL,
		(GuiSignalFunc) cff_gui_file_quit, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuFile[] */

static const cff_guiMenuItem_t cff_gui_menuEdit[] =
{
	{kMTypeItem, mn_edit_undo_str,
		kGAccelEditUndo, &cff_gui_menu_widget[kMenuEditUndo],
		(GuiSignalFunc) cff_gui_edit_undo, &gVideo},
	{kMTypeItem, mn_edit_redo_str,
		kGAccelEditRedo, &cff_gui_menu_widget[kMenuEditRedo],
		(GuiSignalFunc) cff_gui_edit_redo, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_edit_load_str,
		kGAccelEditLoad, &cff_gui_menu_widget[kMenuEditLoad],
		(GuiSignalFunc) cff_gui_edit_load, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_save_str,
		kGAccelEditSave, &cff_gui_menu_widget[kMenuEditSave],
		(GuiSignalFunc) cff_gui_edit_save, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_clear_str,
		kGAccelEditClear, &cff_gui_menu_widget[kMenuEditClear],
		(GuiSignalFunc) cff_gui_edit_clear, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Add"),
		kGAccelEditAdd, &cff_gui_menu_widget[kMenuEditAdd],
		(GuiSignalFunc) cff_gui_edit_add, &gVideo},
	{kMTypeItem, gettext_nop("Remove"),
		kGAccelEditRemove, &cff_gui_menu_widget[kMenuEditRemove],
		(GuiSignalFunc) cff_gui_edit_remove, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, mn_edit_frame_str,
		kGAccelKeyNone, &cff_gui_menu_widget[kMenuEditFrame],
		(GuiSignalFunc) cff_gui_build_frame_submenu, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, gettext_nop("Program"),
		kGAccelKeyNone, &cff_gui_menu_widget[kMenuEditProgram],
		(GuiSignalFunc) cff_gui_build_program_submenu, &gVideo},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuEdit[] */

static const cff_guiMenuItem_t cff_gui_menuOptions[] =
{
	{kMTypeItem, gettext_nop("Keyboard ..."),
		kGAccelOptAccel, NULL,
		(GuiSignalFunc) cff_gaccel_dialog_open, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, mn_opt_keyframe_str,
		kGAccelOptKeyFrame, &cff_gui_checkmenu_frame_widget[0],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptKeyFrame)},
	{kMTypeCheckItem, mn_opt_iframe_str,
		kGAccelOptIFrame, &cff_gui_checkmenu_frame_widget[1],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptIFrame)},
	{kMTypeCheckItem, mn_opt_pframe_str,
		kGAccelOptPFrame, &cff_gui_checkmenu_frame_widget[2],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptPFrame)},
	{kMTypeCheckItem, gettext_nop("Cursor"),
		kGAccelOptCursor, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptCursor)},
	{kMTypeCheckItem, gettext_nop("Overwrite"),
		kGAccelOptOverwrite, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptOverwrite)},
	{kMTypeCheckItem, gettext_nop("Delete copy on error"),
		kGAccelOptDelete, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptDelete)},
	{kMTypeCheckItem, gettext_nop("Beep"),
		kGAccelOptBeep, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptBeep)},
	{kMTypeCheckItem, gettext_nop("Save rc on exit"),
		kGAccelOptSave, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptSaveConfig)},
	{kMTypeSubmenuCheckItem, gettext_nop("Log ..."),
		kGAccelKeyNone, NULL,
		(GuiSignalFunc) cff_gui_build_loglevel_submenu, NULL},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuOptions[] */

static const cff_guiMenuItem_t cff_gui_menuHelp[] =
{
	{kMTypeItem, gettext_nop("FFmpeg ..."),
		kGAccelHelpFFmpeg, NULL,
		(GuiSignalFunc) cff_gui_help_ffmpeg, &cff_gui_menu_bar[kMenuBarHelp]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("CutterFF ..."),
		kGAccelHelpCutterff, NULL,
		(GuiSignalFunc) cff_gui_help_cutterff, &cff_gui_menu_bar[kMenuBarHelp]},
	{kMTypeItem, gettext_nop("About ..."),
		kGAccelHelpAbout, NULL,
		(GuiSignalFunc) cff_gui_help_about, &cff_gui_menu_bar[kMenuBarHelp]},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuHelp[] */

static const cff_guiMenuItem_t cff_gui_menuPopup[] =
{
	{kMTypeItem, mn_edit_undo_str,
		kGAccelKeyNone,  &cff_gui_popup_widget[kPopupEditUndo],
		(GuiSignalFunc) cff_gui_edit_undo, &gVideo},
	{kMTypeItem, mn_edit_redo_str,
		kGAccelKeyNone,  &cff_gui_popup_widget[kPopupEditRedo],
		(GuiSignalFunc) cff_gui_edit_redo, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_file_open_str,
		kGAccelKeyNone, NULL,
		(GuiSignalFunc) cff_gui_file_open, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, mn_file_save_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupFileSave],
		(GuiSignalFunc) cff_gui_file_save, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_edit_load_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditLoad],
		(GuiSignalFunc) cff_gui_edit_load, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_save_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditSave],
		(GuiSignalFunc) cff_gui_edit_save, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_clear_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditClear],
		(GuiSignalFunc) cff_gui_edit_clear, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, mn_edit_frame_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditFrame],
		(GuiSignalFunc) cff_gui_build_frame_submenu, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, mn_opt_keyframe_str,
		kGAccelKeyNone, &cff_gui_popup_checkmenu_frame_widget[0],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptKeyFrame)},
	{kMTypeCheckItem, mn_opt_iframe_str,
		kGAccelKeyNone, &cff_gui_popup_checkmenu_frame_widget[1],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptIFrame)},
	{kMTypeCheckItem, mn_opt_pframe_str,
		kGAccelKeyNone, &cff_gui_popup_checkmenu_frame_widget[2],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptPFrame)},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuPopup[] */


/* ------------------- */


/*
	------
	 MISC
	------
*/

static gboolean cff_gui_key_or_mouse_button_pressed(GtkWidget *widget,
					GdkEvent *event, gboolean *isBusy)
{
	/* Supress mouse or button event if not the main level */

	if (*isBusy == TRUE || gtk_main_level() > 1)
	{
		return TRUE;
	}

	return FALSE;

} /*  static gboolean cff_gui_key_or_mouse_button_pressed(GtkWidget*,
							GdkEvent*, gboolean) */

static void cff_gui_signal_connect_key_or_mouse_button(GtkWidget *widget, gint type)
{
	const gchar *str;

	if (type == 0)
	{
		str = "button-press-event";
	}

	else
	{
		str = "key-press-event";
	}

	GuiSignalConnect(widget, str,
		cff_gui_key_or_mouse_button_pressed, &gIsBusy);

} /* static void cff_gui_signal_connect_key_or_mouse_button(GtkWidget*, gint) */

void cff_gui_beep(void)
{
	gint active;

	if (cff_gprop_get_bool(kPropBoolOptBeep, &active) && active)
	{
		GdkDisplay *display;
		GtkWidget  *widget;

		widget = gWindata.topwin;

		if (widget == NULL)
		{
			display = gdk_display_get_default();
		}

		else
		{
			display = gtk_widget_get_display(widget);
		}

		gdk_display_beep(display);
	}

} /* void cff_gui_beep(void) */

static int cff_gui_overwrite(const gchar *path, GtkWidget **mbar)
{
	gint overwrite;

	if (cff_gprop_get_bool(kPropBoolOptOverwrite, &overwrite) && overwrite)
	{
		return TRUE;
	}

	if (tl_file_exists(path) == 0)
	{
		return TRUE;
	}

	if (guitl_dialog_warning_2btn(gettext_nop("File already exists! Overwrite?"),
			gettext_nop("Cancel"),
			gettext_nop("Yes"),
			mbar) == GTK_RESPONSE_ACCEPT)
	{
		return TRUE;
	}

	return FALSE;

} /* static int cff_gui_overwrite(const gchar*, GtkWidget **mbar) */

static gpointer cff_gui_thread_run(gpointer data)
{
	gpointer result;

	struct
	{
		GThreadFunc	func;
		gpointer	data;

	} *s = data;

	gLogWindow.nThreads++;
	result = (*s->func)(s->data);
	gLogWindow.nThreads--;

	gtk_main_quit();

	return result;

} /* static gpointer cff_gui_thread_run(gpointer) */

static gboolean cff_gui_thread_new(GThreadFunc func, gpointer data, gpointer *retv)
{
	int nThreads;

#ifdef HAVE_GNUMPROCESSORS
	nThreads = g_get_num_processors();
#else

#ifdef HAVE_SYS_NPROCESSORS

#ifdef BUILD_MINGW

	SYSTEM_INFO sysinfo;

	GetSystemInfo(&sysinfo);
	nThreads = sysinfo.dwNumberOfProcessors;

#elif _MACOS

	size_t		len;
	uint32_t	count;
	int		nm[2];

	len   = sizeof(uint32_t);
	nm[0] = CTL_HW;
	nm[1] = HW_AVAILCPU;

	sysctl(nm, 2, &count, &len, NULL, 0);

	if(count < 1)
	{
		nm[1] = HW_NCPU;
		sysctl(nm, 2, &count, &len, NULL, 0);
	}

	nThreads = count;

#else /* UNIX */

	nThreads = sysconf(_SC_NPROCESSORS_ONLN);
#endif

#else
	nThreads = 0;

#endif /* #ifdef HAVE_SYS_NPROCESSORS .. #else */
#endif /* #ifdef HAVE_GNUMPROCESSORS .. #else */

	if (nThreads > 1)
	{
		struct
		{
			GThreadFunc	func;
			gpointer	data;
		} s;

		s.func = func;
		s.data = data;

		GThread *thread = GuiThreadNew(cff_gui_thread_run, &s);

		if (thread != NULL)
		{
			gpointer result;

			gtk_main();
			result = g_thread_join(thread);

			if (retv != NULL)
			{
				*retv = result;
			}

			return TRUE;
		}

	} /* if (nThreads > 1) */

	return FALSE;

} /* static gboolean cff_gui_thread_new(GThreadFunc, gpointer*) */


/*
	--------------
	 PROGRESS-BAR
	--------------
*/

static gboolean cff_gui_dialog_progress_halt(GtkWidget *w, GdkEvent *e, cff_progressbar_t *data)
{
	/* 'button-press-event' or 'delete-event' */

	/* Stop */
	data->progress_bar = NULL;

	return TRUE;

} /* static gboolean cff_gui_dialog_progress_halt(GtkWidget*, GdkEvent*, cff_progressbar_t*) */

static gboolean cff_gui_dialog_progress_set_fraction(gpointer data)
{
	struct
	{
		GtkProgressBar	*progress_bar;
		GuiMutex	*mutex;
		gdouble		progress;

	} *s = data;

	if (gtk_main_level() > 1)
	{
		GtkProgressBar *pbar;

		if (s->mutex != NULL)
		{
			GuiMutexLock(s->mutex);
		}

		pbar = s->progress_bar;

		if (pbar != NULL)
		{
			gtk_progress_bar_set_fraction(pbar, s->progress);
			s->progress_bar = NULL;
		}

		if (s->mutex != NULL)
		{
			GuiMutexUnlock(s->mutex);
		}

	} /* if (gtk_main_level() > 1) */

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_set_fraction(gpointer) */

static gboolean cff_gui_dialog_progress_callback(cff_progressbar_t *data)
{
	double		progress;
	GuiMutex	mutex;

	struct
	{
		volatile GtkProgressBar	*progress_bar;
		GuiMutex		*mutex;
		gdouble			progress;
	} s;

	if (data->is_thread == FALSE)
	{
		s.mutex = NULL;
	}

	else
	{
		s.mutex = &mutex;
		GuiMutexInit(&mutex);
	}

	s.progress_bar = NULL;

	while ((data->retv = (*data->func)(data->video,
					&progress, &(data->error))) == 0)
	{
		if (data->error != NULL || data->progress_bar == NULL)
		{
			break;
		}

		if (s.mutex != NULL)
		{
			GuiMutexLock(s.mutex);
		}

		s.progress = (gdouble) progress;

		if (s.progress_bar == NULL)
		{
			s.progress_bar = data->progress_bar;

			g_main_context_invoke_full(NULL, G_PRIORITY_HIGH,
				cff_gui_dialog_progress_set_fraction, &s, NULL);
		}

		if (s.mutex == NULL)
		{
			return Gui_SOURCE_CONTINUE;
		}

		GuiMutexUnlock(s.mutex);
	}

	/* Wait */

	if (data->is_thread == FALSE)
	{
		while (s.progress_bar != NULL)
		{
			gtk_main_iteration();
		}
	}

	else
	{
		while (s.progress_bar != NULL)
		{
			g_main_context_wakeup(NULL);
		}

		GuiMutexFree(&mutex);
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_callback(cff_progressbar_t*) */

static gboolean cff_gui_dialog_progress_bar(cff_progressbar_t *data)
{
	GtkWidget *widget, *window, *frame, *button, *hbox, *vbox;

	if (data->mutex && data->is_locked == FALSE)
	{
		GuiMutexLock(data->mutex);
		data->is_locked = TRUE;

		/*
			If called from the log-mainthread with
			'gtk_main_iteration', tell mainthread
			that the progressbar is now running.
			Return now to finish that function, and
			try later again.
		*/

		if (gLogWindow.is_mainthread)
		{
			gLogWindow.is_progressbar = TRUE;
			return Gui_SOURCE_CONTINUE;
		}
	}

	gLogWindow.is_progressbar = FALSE;

	vbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);

	window = guitl_window_new_gettext(data->title, vbox, 8);
	guitl_dialog_set_modal(window, gWindata.topwin, data->mbar);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	widget = gtk_progress_bar_new();
	data->progress_bar = GTK_PROGRESS_BAR(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 4);

	button = gtk_button_new_with_label(gettext_nop("Halt"));
	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);

	GuiWidgetCanDefault(button);

	GuiSignalConnect(window, "delete-event",
			cff_gui_dialog_progress_halt, data);
	GuiSignalConnect(button, "button-press-event",
			cff_gui_dialog_progress_halt, data);

	gtk_widget_show_all(window);

	/* Halt-button */
	data->retv = 0;

	data->is_thread = TRUE;

	if (cff_gui_thread_new((GThreadFunc)
		cff_gui_dialog_progress_callback, data, NULL) == FALSE)
	{
		data->is_thread = FALSE;

		g_idle_add_full(G_PRIORITY_LOW,
			(GSourceFunc) cff_gui_dialog_progress_callback,
					data, (GDestroyNotify) gtk_main_quit);
		gtk_main();
	}

	gtk_widget_destroy(window);

	if (data->retv == 0 && data->error == NULL)
	{
		data->error = gettext_nop("User canceled");
	}

	if (data->mutex)
	{
		GuiMutexUnlock(data->mutex);
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_bar(cff_progressbar_t*) */


/*
	----------------
	 FILE-SELECTION
	----------------
*/

static gchar *cff_gui_convert_filename(gchar *fname, gboolean delete)
{
	gchar *str = fname;

	if (str != NULL)
	{
		str = cff_charconv_utf8_to_filename(fname);

		if (str != fname && delete == TRUE)
		{
			g_free(fname);
		}
	}

	return str;

} /* static gchar *cff_gui_convert_filename(gchar*, gboolean) */

static gchar *cff_gui_get_gprop_path(gint prop)
{
	gchar *path, *sep;

	path = cff_gprop_get_str(prop);

	if (path == NULL)
	{
		return NULL;
	}

#ifdef BUILD_MINGW
	file_path_unix_to_dos(path);
#endif
	path = cff_gui_convert_filename(path, TRUE);

	if (path == NULL)
	{
		return NULL;
	}

	sep = &path[strlen(path)];

	do
	{
		GStatBuf dirstat;

		sep[0] = '\0';

		if ( g_stat(path, &dirstat) == 0 &&
				S_ISDIR(dirstat.st_mode) )
		{
			return path;
		}
	}
	while ((sep = strrchr(path, G_DIR_SEPARATOR)) != NULL);

	g_free(path);

	return NULL;

} /* static gchar *cff_gui_get_gprop_path(gint) */

static gchar *cff_gui_get_gprop_fname(const gchar *fname, gint prop)
{
	gchar *path, *str;

	path = cff_gui_get_gprop_path(prop);

	if (path == NULL)
	{
		str = NULL;

		if (fname != NULL)
		{
			str = g_strdup(fname);
#ifdef BUILD_MINGW
			file_path_unix_to_dos(str);
#endif
			str = cff_gui_convert_filename(str, TRUE);
		}
	}

	else
	{
		gchar *fn;

		fname	= tl_get_filename(fname);
		fn	= cff_gui_convert_filename((gchar *) fname, FALSE);

		str = g_strconcat(path, G_DIR_SEPARATOR_S, fn, NULL);
		CFF_CHARCONV_UTF8_FREE(fn, fname);

		g_free(path);

	} /* if (path == NULL) .. elese */

	return str;

} /* static gchar *cff_gui_get_gprop_fname(const gchar*, gint) */

static void cff_gui_set_gprop_path(const gchar *fname, gint prop)
{
	gchar *path, *sep;

	if (fname == NULL || fname[0] == '\0')
	{
		return;
	}

	path = g_strdup(fname);

#ifdef BUILD_MINGW
	file_path_dos_to_unix(path);
#endif

	if ((sep = strrchr(path, '/')) != NULL)
	{
		GStatBuf dirstat;

		*sep = '\0';

		if ( g_stat(path, &dirstat) == 0 &&
				S_ISDIR(dirstat.st_mode) )
		{
			gchar *str;

			str = cff_charconv_utf8_from_filename(path);

			if (str != NULL)
			{
				cff_gprop_set_str(prop, str);
				CFF_CHARCONV_UTF8_FREE(str, path);
			}
		}

	} /* if ((sep = strrchr(path, '/')) != NULL) */

	g_free(path);

} /* static void cff_gui_set_gprop_path(const gchar*, gint) */

static gchar *cff_gui_get_dirname(const gchar *fname)
{
	gchar		*path;
	const gchar	*fn;
	GStatBuf	dirstat;

	fn = tl_get_filename(fname);

	if (fn == fname)
	{
		return NULL;
	}

	path = g_strndup(fname, fn - fname);

#ifdef BUILD_MINGW
	file_path_unix_to_dos(path);
#endif
	path = cff_gui_convert_filename(path, TRUE);

	if ( g_stat(path, &dirstat) != 0 ||
			! S_ISDIR(dirstat.st_mode) )
	{
		g_free(path);
		path = NULL;
	}

	return path;

} /* static gchar *cff_gui_get_dirname(const gchar*) */

static gchar *cff_gui_file_filter_validate_string(const gchar *text, gint length, guint skip)
{
	gboolean	changed;
	gint		i, n;
	gchar		*str, *s;
	const gchar	*txt;

	n = length;

	if (n < 0)
	{
		n = strlen(text);
	}

	changed	= FALSE;

	while (n > 0 && *(text) == ' ')
	{
		n--;
		text++;
		changed = TRUE;
	}

	if (n == 0)
	{
		return NULL;
	}

	txt = text;
	i   = n;

	while (--i >= 0)
	{
		/* Skip space if string contains multiple extensions */

		guint chr = *(txt++);

		if (chr != skip)
		{
			if (g_ascii_isalnum(chr) == FALSE)
			{
				/* Not alphanumeric */
				return NULL;
			}

			if (g_ascii_isupper(chr))
			{
				/* Convert to lower */
				changed = TRUE;
			}
		}

	} /* while (--i >= 0) */

	if (changed == FALSE)
	{
		return (gchar *) text;
	}

	str = g_malloc(n + 1);
	s   = str;

	while (--n >= 0)
	{
		*(s++) = g_ascii_tolower(*(text++));
	}

	*s = '\0';

	return str;

} /* static gchar *cff_gui_file_filter_validate_string(const gchar*, gint, guint) */

static gchar *cff_gui_file_filter_extension_next(gchar *str, gchar **nextptr)
{
	guint chr;
	gchar *end, *next;

	end = strchr(str, ' ');

	if (end == NULL)
	{
		end = strchr(str, '\0');
	}

	chr  = *end;
	*end = '\0';
	next = end;

	while (chr == ' ')
	{
		chr = *(++next);
	}

	*nextptr = next;

	return end;

} /* static gchar *cff_gui_file_filter_extension_next(gchar *beg, gchar**) */

static gboolean cff_gui_file_filter_extension_compare(const gchar *str, const gchar *pattern)
{
	if (str != NULL)
	{
		const gchar *match;

		while ((match = strstr(str, pattern)) != NULL)
		{
			/* (match == pattern) => found itself */

			if (match != pattern)
			{
				guint  chr;
				size_t len;

				len = strlen(pattern);
				chr = match[len];

				if (chr == ' ' || chr == '\0')
				{
					/* Last characters are equal */

					chr = ' ';

					if (match > str)
					{
						chr = *(--match);
						match++;
					}

					if (chr == ' ')
					{
						/* First characters are equal */
						return TRUE;
					}
				}

			} /* if (match != pattern) */

			str = &match[1];

		} /* while ((match = strstr(str, pattern)) != NULL) */

	} /* if (str != NULL) */

	return FALSE;

} /* static gboolean cff_gui_file_filter_extension_compare(const gchar*, const gchar*) */

static gchar *cff_gui_file_filter_extension_remove(gchar *beg, gchar *end, gchar *start)
{
	if (*end == '\0')
	{
		/* Last extension in the string */

		while (beg > start)
		{
			/* Not the first entry */

			if (*(--beg) != ' ')
			{
				beg++;
				break;
			}
		}

		/* Remove last extension */
		*beg = '\0';
	}

	else
	{
		/* Move following extension(s) down */
		beg = memmove(beg, end, strlen(end) + 1);
	}

	return beg;

} /* static gchar *cff_gui_file_filter_extension_remove(gchar*, gchar*, gchar*) */

static void cff_gui_file_filter_validate_prop(gint prop)
{
	gchar *ext2 = cff_gprop_get_str(prop);

	if (ext2 != NULL)
	{
		gboolean changed;
		gchar    *str;

		changed	= FALSE;
		str	= cff_gui_file_filter_validate_string(ext2, -1, ' ');

		if (str != ext2)
		{
			/* 'ext2' was changed or is invalid */
			g_free(ext2);
			ext2	= str;
			changed	= TRUE;
		}

		if (ext2 != NULL)
		{
			gint		n;
			const gchar	*ext;

			if (prop == kPropStrFileVideoFilter)
			{
				ext = cff_video_ext;
			}

			else /* kPropStrFileImageFilter */
			{
				ext = cff_image_ext;
			}

			n = 0;

			do
			{
				gchar *end, *next;

				end = cff_gui_file_filter_extension_next(str, &next);

				if ((next - end) > 1)
				{
					/* More than one space */

					if (*next != '\0')
					{
						/* No following extension(s) */
						next--;
					}

					/* Move following spaces down */
					cff_gui_file_filter_extension_remove(end, next, ext2);
					changed = TRUE;
				}

				else if ((end - str) >= kFileFilterCharsMax ||
					cff_gui_file_filter_extension_compare(ext, str) ||
						cff_gui_file_filter_extension_compare(ext2, str))
				{
					/* Move following extensions down */
					str = cff_gui_file_filter_extension_remove(str, next, ext2);
					changed = TRUE;
				}

				else
				{
					if (++n >= kFileFilterEntriesMax)
					{
						/* String was terminated (end) */
						changed = TRUE;
						break;
					}

					if (*next != '\0')
					{
						/* Not the last extension */
						*end = ' ';
					}

					str = next;
				}
			}
			while (*str != '\0');

		} /* if (ext2 != NULL) */

		if (changed)
		{
			cff_gprop_set_str(prop, ext2);
		}

		g_free(ext2);

	} /* if (ext2 != NULL) */

} /* static void cff_gui_file_filter_validate_prop(gint) */

static const gchar *cff_gui_file_filter_add(GtkFileChooser *chooser, const gchar *ext, GtkFileFilter **filter)
{
	static const gchar all_files[] = gettext_nop("All files");

	guint chr;
	gchar *str, *name;
	gchar filter_glob[256], filter_name[256];

	do
	{
		chr = *(ext++);
	}
	while (chr == ' ');

	if (chr == '\0')
	{
		ext--;
		return ext;
	}

	str = filter_glob;

	if (chr == '*')
	{
		*(str++) = chr;
		chr = *(ext++);

		name = cff_charconv_utf8_gettext(all_files);
		strcpy(filter_name, name);
		CFF_CHARCONV_UTF8_FREE(name, all_files);
	}

	else
	{
		name = filter_name;

		*(str++) = '*';
		*(str++) = '.';

		do
		{
			/* Extension in lower and uppercase */
			*(str++) = '[';
			*(str++) = chr;

			/* Name in uppercase */
			chr = g_ascii_toupper(chr);
			*(str++)  = chr;
			*(name++) = chr;

			*(str++) = ']';

			chr = *(ext++);

			if (chr == '\0')
			{
				/* Done */
				ext--;
				break;
			}
		}
		while (chr != ' ');

		*name = '\0';

	} /* if (chr == '*') .. else */

	*str = '\0';

	*filter = gtk_file_filter_new();
	gtk_file_filter_set_name(*filter, filter_name);
	gtk_file_filter_add_pattern(*filter, filter_glob);
	gtk_file_chooser_add_filter(chooser, *filter);

	return ext;

} /* static const gchar *cff_gui_file_filter_add(GtkFileChooser*, const gchar*, GtkFileFilter**) */

static gint cff_gui_file_filter_init(GtkFileChooser *chooser, gint prop, GtkFileFilter *filterbuf[512])
{
	static const gchar cutterff_ext[] = "cff";

	gint		n, select;
	guint		num;
	gchar		*ext2;
	const gchar	*ext;

	if (prop == kPropStrEditPath)
	{
		ext = cutterff_ext;
		select = -1;
	}

	else if (prop == kPropStrFileImagePath)
	{
		ext	= cff_image_ext;
		select	= 1;
		prop	= kPropStrFileImageFilter;
	}

	else
	{
		if (cff_gprop_get_int(kPropIntFileFilterIndex, &select) == FALSE)
		{
			select = -1;
		}

		ext  = cff_video_ext;
		prop = kPropStrFileVideoFilter;
	}

	n   = -1;

	do
	{
		ext = cff_gui_file_filter_add(chooser, ext, &filterbuf[0]);

		if (++n == select)
		{
			gtk_file_chooser_set_filter(chooser, filterbuf[0]);
		}
	}
	while (*ext != '\0');


	num = 0;

	if (prop != kPropStrEditPath &&
			(ext2 = cff_gprop_get_str(prop)) != NULL)
	{
		ext = ext2;

		do
		{
			ext = cff_gui_file_filter_add(chooser, ext, &filterbuf[num]);

			if (++n == select)
			{
				gtk_file_chooser_set_filter(chooser, filterbuf[num]);
			}
		}
		while (++num < kFileFilterEntriesMax && *ext != '\0');

		g_free(ext2);

	} /* if (prop != kPropStrEditPath &&
				(ext2 = cff_gprop_get_str(prop)) != NULL) */

	return num;

} /* static gint cff_gui_file_filter_init(GtkFileChooser*, gint, GtkFileFilter*[512]) */

static void cff_gui_file_selection_add_filter_entry_cb(GtkEditable *editable, const gchar *text, gint length, gint *pos, gint *sigid)
{
	gchar *str;

	GuiSignalHandlerBlock(editable, *sigid);

	str = cff_gui_file_filter_validate_string(text, length, '\0');

	if (str != NULL)
	{
		gtk_editable_insert_text(editable, str, length, pos);

		if (str != text)
		{
			g_free(str);
		}
	}

	GuiSignalEmitStopByName(editable, "insert-text");
	GuiSignalHandlerUnblock(editable, *sigid);

} /* static void cff_gui_file_selection_add_filter_entry_cb(GtkEditable *editable, const gchar*, gint, gint*, gint*) */

static void cff_gui_file_selection_add_filter_cb(GtkButton *button, cff_filefilter_t *filter)
{
	guint		sigid;
	gchar		*ext, *ext2;
	const gchar	*space;
	GtkWidget	*border, *content, *entry, *frame, *label, *vbox, *widget;
	GtkDialog	*dialog;

	if (filter->num >= kFileFilterEntriesMax)
	{
		return;
	}

	widget = guitl_dialog_new_gettext_2btn(gettext_nop("Add filter"),
						gettext_nop("Cancel"),
						gettext_nop("OK"),
						GTK_WIDGET(filter->chooser));
	dialog = GTK_DIALOG(widget);
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	vbox = GuiVBoxNew(FALSE, 2);
	border = guitl_frame_new_add_child_gettext(NULL, vbox);
	gtk_container_set_border_width(GTK_CONTAINER(border), 8);

	entry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(entry), kFileFilterCharsMax);

#if GTK_MAJOR_VERSION == 2
	widget = GuiHBoxNew(FALSE, 2);
	GuiHBoxPackStart(widget, entry, TRUE,  TRUE, 8);
	GuiVBoxPackStart(vbox, widget, FALSE, FALSE, 4);
#else
	gtk_widget_set_hexpand(entry, TRUE);
	GuiMiscSetPadding(entry, 8);
	GuiVBoxPackStart(vbox, entry, FALSE, FALSE, 4);
#endif

	sigid = GuiSignalConnect(GTK_EDITABLE(entry), "insert-text",
			cff_gui_file_selection_add_filter_entry_cb, &sigid);

	ext2  = cff_gprop_get_str(filter->prop);
	space = NULL;

	if (ext2 != NULL)
	{
		space = " ";
	}

	/* Skip "* " */
	ext = g_strconcat(&filter->ext[2], space, ext2, NULL);

	label = gtk_label_new(ext);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	GuiMiscSetPadding(label, 8);

	g_free(ext);

	frame = guitl_frame_new_add_child_gettext(gettext_nop("Filters"), label);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 8);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 2);

	gtk_widget_show_all(border);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), border);

	while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT)
	{
		const gchar *str = gtk_entry_get_text(GTK_ENTRY(entry));

		if (str != NULL && str[0] != '\0')
		{
			gint		slen;
			gchar		*s;
			GtkFileFilter	**f;

			if (cff_gui_file_filter_extension_compare(filter->ext, str) ||
					cff_gui_file_filter_extension_compare(ext2, str))
			{
				/* Filter already exists */
				cff_gui_beep();
				continue;
			}

			f = &filter->buf[filter->num];

			cff_gui_file_filter_add(filter->chooser, str, f);
			gtk_file_chooser_set_filter(filter->chooser, *f);
			filter->num++;

			slen = (ext2 == NULL) ? 0 : strlen(ext2);
			/* Plus '\0' or ' ' */
			ext2 = g_realloc(ext2, strlen(str) + slen + 1);
			s    = &ext2[slen];

			if (slen > 0)
			{
				*(s++) = ' ';
			}

			strcpy(s, str);
			cff_gprop_set_str(filter->prop, ext2);

			gtk_widget_set_sensitive(filter->btn_remove, TRUE);

			if (filter->num >= kFileFilterEntriesMax)
			{
				gtk_widget_set_sensitive(filter->btn_add, FALSE);
			}

			break;

		} /* if (str != NULL && str[0] != '\0') */

	} /* while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT) */

	g_free(ext2);
	gtk_widget_destroy(GTK_WIDGET(dialog));

} /* static void cff_gui_file_selection_add_filter_cb(GtkButton*, cff_filefilter_t*) */

static void cff_gui_file_selection_remove_filter_cb(GtkButton *button, cff_filefilter_t *filter)
{
	guint			num;
	gchar			*ext2, *str;
	GtkWidget		*content, *frame, *swin, *widget;
	GtkDialog		*dialog;
	GtkListStore		*store;
	GtkTreeView		*view;
	GtkTreeViewColumn	*column;
	GtkTreeSelection	*selection;
	GtkCellRenderer		*renderer;
	GtkTreeIter		iter;

	if (filter->num == 0)
	{
		return;
	}

	ext2 = cff_gprop_get_str(filter->prop);

	if (ext2 == NULL)
	{
		return;
	}

	store = gtk_list_store_new(4, G_TYPE_STRING, G_TYPE_POINTER,
						G_TYPE_POINTER, G_TYPE_INT);
	view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
	gtk_tree_view_set_headers_visible(view, FALSE);

	g_object_unref(G_OBJECT(store));

	selection = gtk_tree_view_get_selection(view);
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

	column   = gtk_tree_view_column_new();
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", 0, NULL);

	gtk_tree_view_append_column(view, column);


	num = 0;
	str = ext2;

	do
	{
		gchar *end, *next;

		end  = cff_gui_file_filter_extension_next(str, &next);

		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
				0, str, 1, str, 2, next, 3, num, -1);

		if (end != next)
		{
			*end = ' ';
		}

		str = next;
	}
	while (*str != '\0' && ++num < filter->num);


	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 8);

	swin = guitl_scrolled_window_new(frame, -1, 150, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(swin), GTK_WIDGET(view));

	widget = guitl_dialog_new_gettext_2btn(gettext_nop("Remove filter"),
						gettext_nop("Cancel"),
						gettext_nop("OK"),
						GTK_WIDGET(filter->chooser));
	dialog = GTK_DIALOG(widget);
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	gtk_widget_show_all(frame);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), frame);

	while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT)
	{
		guint		n;
		gchar		*s, *e;
		GtkTreeModel	*model;

		if (gtk_tree_selection_get_selected(selection,
						&model, &iter) == FALSE)
		{
			/* Nothing selected */
			cff_gui_beep();
			continue;
		}

		gtk_tree_model_get(model, &iter, 1, &s, 2, &e, 3, &n, -1);

		cff_gui_file_filter_extension_remove(s, e, ext2);
		gtk_file_chooser_remove_filter(filter->chooser, filter->buf[n]);

		num = filter->num - (n + 1);

		if (num != 0)
		{
			/* Move following entries one position down */
			memmove(&filter->buf[n], &filter->buf[n + 1],
						num * sizeof(GtkFileFilter*));
		}

		filter->num--;

		if (filter->num == 0)
		{
			gtk_widget_set_sensitive(filter->btn_remove, FALSE);
		}

		if (filter->num < kFileFilterEntriesMax)
		{
			gtk_widget_set_sensitive(filter->btn_add, TRUE);
		}

		cff_gprop_set_str(filter->prop, ext2);

		break;

	} /* while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT) */

	g_free(ext2);
	gtk_widget_destroy(GTK_WIDGET(dialog));

} /* static void cff_gui_file_selection_remove_filter_cb(GtkButton*, cff_filefilter_t*) */

static gchar *cff_gui_file_selection_new_gettext(const gchar *title, const gchar *fname, GtkWidget **mbar, gint prop)
{
	gboolean		path_prop;
	gchar			*str, *btn_ok, *btn_cancel, *accept, *cancel;
	GtkWidget		*widget;
	GtkFileChooserAction	action;
#ifdef HAVE_GTKOVERWRITECONFIRMATION
	int			overwrite;
#endif
#if GTK_MAJOR_VERSION > 2
	gint			width, height;
	GtkWindow		*window;
#endif
	cff_filefilter_t	filter_s;

	if (mbar)
	{
		GuiWidgetSetState(*mbar, Gui_STATE_PRELIGHT);
	}

	if (fname == NULL)
	{
		if (prop == kPropStrFileOpenPath)
		{
			btn_ok = gettext_nop("Open");
		}

		else
		{
			btn_ok = gettext_nop("Load");
		}

		action = GTK_FILE_CHOOSER_ACTION_OPEN;
	}

	else
	{
		if (prop == kPropStrFileSavePath ||
				prop == kPropStrFileImagePath)
		{
			btn_ok = gettext_nop("OK");
		}

		else
		{
			btn_ok = gettext_nop("Save");
		}

		action = GTK_FILE_CHOOSER_ACTION_SAVE;

	} /* if (fname == NULL) .. else */

	btn_cancel = gettext_nop("Cancel");

	accept	= cff_charconv_utf8_gettext(btn_ok);
	cancel	= cff_charconv_utf8_gettext(btn_cancel);
	str	= cff_charconv_utf8_gettext(title);

	widget = gtk_file_chooser_dialog_new(str,
			GTK_WINDOW(gWindata.topwin),
			action,
			cancel, GTK_RESPONSE_CANCEL,
			accept, GTK_RESPONSE_ACCEPT,
			NULL);

	CFF_CHARCONV_UTF8_FREE(str, title);
	CFF_CHARCONV_UTF8_FREE(cancel, btn_cancel);
	CFF_CHARCONV_UTF8_FREE(accept, btn_ok);

	filter_s.chooser = GTK_FILE_CHOOSER(widget);

	if (prop != kPropStrEditPath)
	{
		/* Extra widget */

		GtkWidget *frame, *hbox;
#if GTK_MAJOR_VERSION == 2
		GtkWidget *bbox;
#endif

		if (prop == kPropStrFileImagePath)
		{
			filter_s.ext  = cff_image_ext;
			filter_s.prop = kPropStrFileImageFilter;
		}

		else
		{
			filter_s.ext  = cff_video_ext;
			filter_s.prop = kPropStrFileVideoFilter;
		}

		hbox = GuiHButtonBoxNew();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_START);
		frame = guitl_frame_new_add_child_gettext(gettext_nop("Filter"), hbox);
		gtk_container_set_border_width(GTK_CONTAINER(frame), 4);

#if GTK_MAJOR_VERSION == 2
		bbox = GuiHBoxNew(FALSE, 1);
		GuiHBoxPackStart(bbox, frame, FALSE, FALSE, 1);
		frame = bbox;
#endif
		filter_s.btn_add = guitl_button_box_add_button(hbox,
				gettext_nop("Add ..."),
				(GCallback) cff_gui_file_selection_add_filter_cb,
				&filter_s);
		filter_s.btn_remove = guitl_button_box_add_button(hbox,
				gettext_nop("Remove ..."),
				(GCallback) cff_gui_file_selection_remove_filter_cb,
				&filter_s);

		gtk_widget_show_all(frame);
		gtk_file_chooser_set_extra_widget(filter_s.chooser, frame);

	} /* if (prop != kPropStrEditPath) */

	if ((prop == kPropStrFileSavePath ||
		prop == kPropStrFileImagePath) &&
			(fname == NULL || fname[0] == '\0'))
	{
		fname = "noname";
	}

	if (fname != NULL)
	{
		const gchar *fn;

		fn  = tl_get_filename(fname);
		str = cff_gui_convert_filename((gchar *) fn, FALSE);

		if (str != NULL)
		{
			gtk_file_chooser_set_current_name(filter_s.chooser, str);
			CFF_CHARCONV_UTF8_FREE(str, fn);
		}

	} /* if (fname != NULL) */

	if (prop == kPropStrFileSavePath || prop == kPropStrFileImagePath)
	{
		/* Get path from filename (may changed in the entry) */
		str = cff_gui_get_dirname(fname);
		path_prop = FALSE;
	}

	else
	{
		path_prop = TRUE;
	}

	if (path_prop || str == NULL)
	{
		/* Get path stored in gprop */
		str = cff_gui_get_gprop_path(prop);

		if (str == NULL && path_prop)
		{
			/* Get path from filename */
			str = cff_gui_get_dirname(fname);
		}
	}

	if (str != NULL)
	{
		gtk_file_chooser_set_current_folder(filter_s.chooser, str);
		g_free(str);
	}

#ifdef HAVE_GTKOVERWRITECONFIRMATION
	if (action == GTK_FILE_CHOOSER_ACTION_SAVE &&
			(cff_gprop_get_bool(kPropBoolOptOverwrite,
				&overwrite) == FALSE || overwrite == FALSE))
	{
		gtk_file_chooser_set_do_overwrite_confirmation(filter_s.chooser, TRUE);
	}
#endif

	filter_s.num = cff_gui_file_filter_init(filter_s.chooser,
						prop, filter_s.buf);
	if (prop != kPropStrEditPath)
	{
		if (filter_s.num == 0)
		{
			gtk_widget_set_sensitive(filter_s.btn_remove, FALSE);
		}

		else if (filter_s.num >= kFileFilterEntriesMax)
		{
			gtk_widget_set_sensitive(filter_s.btn_add, FALSE);
		}

	} /* if (prop != kPropStrEditPath) */

#if GTK_MAJOR_VERSION > 2

	/*
		Running a 32-bit version on a 64-bit platform
		might display a huge filechooser, with the
		titlebar and butttons out of the screen.
	*/

	gtk_widget_show_all(widget);

	window = GTK_WINDOW(widget);
	gtk_window_get_size(window, &width, &height);

	if (guitl_window_test_size(window, width, height) == FALSE)
	{
		gtk_window_unmaximize(window);
		gtk_window_move(window, 16, 16);
	}

#endif /* #if GTK_MAJOR_VERSION > 2 */

	str = NULL;

	if (gtk_dialog_run(GTK_DIALOG(widget)) == GTK_RESPONSE_ACCEPT)
	{
		if (prop == kPropStrFileOpenPath || prop == kPropStrFileSavePath)
		{
			gint		index;
			GtkFileFilter	*filter;
			GSList		*list;

			filter = gtk_file_chooser_get_filter(filter_s.chooser);
			list   = gtk_file_chooser_list_filters(filter_s.chooser);
			index  = g_slist_index(list, filter);
			g_slist_free(list);

			cff_gprop_set_int(kPropIntFileFilterIndex, index);

		} /* if (prop == kPropStrFileOpenPath || prop == kPropStrFileSavePath) */

		str = gtk_file_chooser_get_filename(filter_s.chooser);

		/* The save dialogs store the path itself */

		if (prop == kPropStrFileOpenPath || prop == kPropStrEditPath)
		{
			cff_gui_set_gprop_path(str, prop);
		}

	} /* if (gtk_dialog_run(GTK_DIALOG(widget)) == GTK_RESPONSE_ACCEPT) */

	gtk_widget_destroy(widget);

	if (mbar)
	{
		GuiWidgetSetState(*mbar, Gui_STATE_NORMAL);
	}

	return str;

} /* static gchar *cff_gui_file_selection_new_gettext(const gchar*, const gchar*, GtkWidget**, gint) */


/*
	-------------
	 SAVE COMMON
	-------------
*/

static void cff_gui_save_okay_button_set_sensitive(cff_save_common_t *vscommon, guint bits)
{
	if (bits != vscommon->okay.bits)
	{
		gboolean active;

		vscommon->okay.bits = bits;

		active = TRUE;

		if (vscommon->okay.bits != 0)
		{
			/* Invalid entry, or no stream in the output list */
			active = FALSE;
		}

		gtk_widget_set_sensitive(vscommon->okay.widget, active);

	} /* if (bits != vscommon->okay.bits) */

} /* static void cff_gui_save_okay_button_set_sensitive(cff_save_common_t*, guint) */

static void cff_gui_save_entry_changed(cff_save_common_t *vscommon)
{
	/* Disable the okay-button if the entry is empty or invalid */

	const gchar	*str;
	guint		bits;

	str  = gtk_entry_get_text(vscommon->entry);
	bits = vscommon->okay.bits;

	if (str == NULL)
	{
		/* Empty */
		bits |= 0x01;
	}

	else
	{
		guint    chr, last_chr;
		gboolean first, ext;

		first    = TRUE;
		ext      = FALSE;
		last_chr = '\0';

		while ((chr = *(str++)) != '\0')
		{
			if (isspace(chr))
			{
				if (first)
				{
					break;
				}

				if (last_chr == '.')
				{
					ext = TRUE;
				}
			}

			last_chr = chr;
			first    = CFF_IS_DIR_SEPARATOR(chr);

			if (first)
			{
				ext = FALSE;
			}

		} /* while ((chr = *(str++)) != '\0') */

		if (chr != '\0' || first == TRUE ||
					ext == TRUE || isspace(last_chr))
		{
			/* Empty, or contains leading spaces */
			bits |= 0x01;
		}

		else
		{
			/* Entry is okay */
			bits &= ~0x01;
		}

	} /* if (str == NULL .. else */

	cff_gui_save_okay_button_set_sensitive(vscommon, bits);

} /* static void cff_gui_save_entry_changed(cff_save_common_t*) */

static void cff_gui_save_entry_delete_callback(GtkEntryBuffer *buffer,
		guint position, guint n_chars, cff_save_common_t *vscommon)
{
	cff_gui_save_entry_changed(vscommon);

} /* static void cff_gui_save_entry_delete_callback(GtkEntryBuffer*,
					guint, guint, cff_save_common_t*) */

static void cff_gui_save_entry_insert_callback(GtkEntryBuffer *buffer,
		guint position, gchar *chars, guint n_chars, cff_save_common_t *vscommon)
{
	cff_gui_save_entry_changed(vscommon);

} /* static void cff_gui_save_entry_insert_callback(GtkEntryBuffer *buffer,
					guint, gchar*, guint, cff_save_common_t*) */

static void cff_gui_save_selection_callback(GtkWidget *w, cff_save_common_t *vscommon)
{
	gchar *ename, *fname, *str;

	ename	= (gchar *) gtk_entry_get_text(vscommon->entry);
	str	= ename;

#ifndef BUILD_MINGW

	if (ename && ename[0] != '\0')
	{
		str = cff_charconv_utf8_to_filename(ename);

		if (str == NULL)
		{
			str = ename;
		}
	}
#endif

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Select ..."),
						str, NULL, vscommon->path_prop);

#ifndef BUILD_MINGW
	CFF_CHARCONV_UTF8_FREE(str, ename);
#endif

	if (fname != NULL)
	{
		g_free(vscommon->filechooser_utf8str);

#ifdef BUILD_MINGW
		/* already utf8 */
		vscommon->filechooser_utf8str = fname; /* Do not free 'fname' */
		gtk_entry_set_text(vscommon->entry, fname);
#else
		str = cff_charconv_utf8_from_filename(fname);
		vscommon->filechooser_utf8str = str; /* Do not free 'str' */
		gtk_entry_set_text(vscommon->entry, str);

		if (fname != str)
		{
			g_free(fname);
		}
#endif

	} /* if (fname != NULL) */

} /* static void cff_gui_save_selection_callback(GtkWidget*, cff_save_common_t*) */

static void cff_gui_save_optionmenu_callback(GuiOptionMenu *option, cff_save_common_t *vscommon)
{
	gint		i, pos, slen;
	gchar		*extension;
	const gchar	*str, *end;
	GtkEditable	*editable;

	i = GuiOptionMenuGetActive(option);

	extension = NULL;

	if (i > 0)
	{
		const gchar *format_name = NULL;

		if (vscommon->path_prop == kPropStrFileSavePath)
		{
			if (vscommon->format.u.list != NULL)
			{
				/* Array without "Auto" */
				format_name =
					vscommon->format.u.list[i - 1].name;
			}
		}

		else /* kPropStrFileImagePath */
		{
			if (vscommon->format.u.names != NULL)
			{
				/* Array without "Auto" */
				format_name = vscommon->format.u.names[i - 1];
			}
		}

		if (format_name != NULL)
		{
			extension = cff_ffmpeg_get_outformat_extension(format_name, &slen);
		}

	} /* if (i > 0) */

	if (extension == NULL)
	{
		if ((extension = vscommon->format.extension_orig) == NULL)
		{
			return;
		}

		slen = -1;
	}

	str = gtk_entry_get_text(vscommon->entry);
	end = strrchr(str, '.');

	editable = GTK_EDITABLE(vscommon->entry);

	if (end != NULL)
	{
		if (strcmp(end, extension) == 0)
		{
			if (extension != vscommon->format.extension_orig)
			{
				/* malloc'ed */
				free(extension);
			}

			return;
		}

		pos = end - str;
		gtk_editable_delete_text(editable, pos, -1);
	}

	/* Append new extension */
	pos = -1;
	gtk_editable_insert_text(editable, extension, slen, &pos);
	gtk_editable_set_position(editable, -1);

	g_free(vscommon->filechooser_utf8str);
	vscommon->filechooser_utf8str = NULL;

	if (extension != vscommon->format.extension_orig)
	{
		/* malloc'ed */
		free(extension);
	}

} /* static void cff_gui_save_optionmenu_callback(GuiOptionMenu*, cff_save_common_t*) */

static GtkWidget *cff_gui_save_build_common(cff_save_common_t *vscommon, const gchar *title, gint prop)
{
	gchar		*fname, *end, *str;
	const gchar	*dot, *ext;
	GtkWidget	*entry, *hbox, *vbox, *frame;
	GtkEntryBuffer	*buffer;
	gchar		num[128];

	vscommon->path_prop = prop;

	vbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);

	vscommon->window = guitl_window_new_gettext(title, vbox, 8);
	guitl_dialog_set_modal(vscommon->window, gWindata.mainwin, vscommon->mbar);

	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	vbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 4);

	entry = gtk_entry_new();
	vscommon->entry = GTK_ENTRY(entry);
	GuiHBoxPackStart(hbox, entry, TRUE, TRUE, 1);

#if GTK_MAJOR_VERSION > 2
	gtk_widget_set_hexpand(entry, TRUE);
#endif

	vscommon->filechooser_utf8str = NULL;
	guitl_hbox_add_button_gettext(hbox, "..",
		(GCallback) cff_gui_save_selection_callback, vscommon);

	fname = cff_gui_get_gprop_fname((vscommon->video)->fname,
							vscommon->path_prop);

	end = strrchr(fname, '.');
	dot = end;
	ext = NULL;

	if (end != NULL)
	{
		*end = '\0';
		dot  = ".";
	}

	if (vscommon->path_prop == kPropStrFileSavePath)
	{
		if (end != NULL)
		{
			ext  = &end[1];
		}

		num[0] = '\0';
	}

	else
	{
		/* kPropStrFileImagePath */
		dot  = ".";
		ext  = "jpg";
		snprintf(num, sizeof(num), "_%u",
			(vscommon->video)->vframe.frame_current);
	}

	str = g_strconcat(fname, num, "_cff", dot, ext, NULL);
	g_free(fname);

	vscommon->format.extension_orig = NULL;

#ifdef BUILD_MINGW

	/* Already utf8 */
	gtk_entry_set_text(vscommon->entry, str);

	ext = strrchr(str, '.');
	vscommon->format.extension_orig = g_strdup(ext);
#else
	fname = cff_charconv_utf8_from_filename(str);

	if (fname != NULL)
	{
		gtk_entry_set_text(vscommon->entry, fname);

		ext = strrchr(fname, '.');
		vscommon->format.extension_orig = g_strdup(ext);

		CFF_CHARCONV_UTF8_FREE(fname, str);
	}

#endif /* #ifdef BUILD_MINGW .. #else */

	g_free(str);

	buffer = gtk_entry_get_buffer(vscommon->entry);

	GuiSignalConnect(buffer, "deleted-text",
				cff_gui_save_entry_delete_callback, vscommon);
	GuiSignalConnect(buffer, "inserted-text",
				cff_gui_save_entry_insert_callback, vscommon);
	return vbox;

} /* static GtkWidget *cff_gui_save_build_common(cff_save_common_t *vscommon, const gchar*, gint) */

static gchar *cff_gui_save_get_fname(cff_save_common_t *vscommon, gboolean *fname_free)
{
	gboolean	ask_overwrite;
	gchar		*fname;
	const gchar	*utf8;

	if ((vscommon->video)->fname == NULL)
	{
		return NULL;
	}

	utf8 = gtk_entry_get_text(vscommon->entry);

	if (utf8 == NULL || utf8[0] == '\0')
	{
		cff_gui_beep();
		return NULL;
	}

#ifdef BUILD_MINGW
	fname = (gchar *) utf8;
#else
	fname = cff_charconv_utf8_to_filename(utf8);

	if (fname == NULL)
	{
		fname = (gchar *) utf8;
	}
#endif

	if (strcmp(fname, (vscommon->video)->fname) == 0)
	{
		static const gchar text[] =
			gettext_nop("The filename or the folder must differ!");

#ifndef BUILD_MINGW
		CFF_CHARCONV_UTF8_FREE(fname, utf8);
#endif
		cff_gui_beep();
		guitl_dialog_error(text, NULL);
		return NULL;
	}

	ask_overwrite = FALSE;

#ifdef HAVE_GTKOVERWRITECONFIRMATION
	if (vscommon->filechooser_utf8str == NULL ||
			strcmp(utf8, vscommon->filechooser_utf8str) != 0)
#endif
	{
		ask_overwrite = TRUE;
	}

	if (ask_overwrite && cff_gui_overwrite(fname, NULL) == FALSE)
	{

#ifndef BUILD_MINGW
		CFF_CHARCONV_UTF8_FREE(fname, utf8);
#endif
		return NULL;
	}

	*fname_free = FALSE;

	if (fname != utf8)
	{
		*fname_free = TRUE;
	}

	cff_gui_set_gprop_path(utf8, vscommon->path_prop);

	return fname;

} /* static gchar *cff_gui_save_get_fname(cff_save_common_t*, gboolean*) */


/*
	------------
	 SAVE VIDEO
	------------
*/

static int cff_gui_save_copy_callback(gpointer data)
{
	const gchar		**str;
	cff_progressbar_t	user_data;

	user_data.func  = cff_ffmpeg_copy_cb;
	user_data.video = &gVideo;
	user_data.title = gettext_nop("Copy Video File");
	user_data.mbar  = NULL;
	user_data.mutex = NULL;

	cff_gui_dialog_progress_bar(&user_data);

	str = (const gchar **) data;

	if (str != NULL)
	{
		*str = user_data.error;
	}

	return user_data.retv;

} /* static int cff_save_copy_callback(gpointer) */

static void cff_gui_save_start(GtkWidget *w, cff_save_video_t *vsave)
{
	gboolean	retv, fname_free;
	gint		i, n, rows, nb_streams, *stream_idx;
	gchar		*fname;
	gchar		*info;
	const gchar	*error, *result;
	GtkWidget	**mbar;
	GtkTreeModel	*model;
	GtkTreeIter	iter;
	cff_copy_param_t param;

	model = GTK_TREE_MODEL(vsave->out.list.store);
	rows  = gtk_tree_model_iter_n_children(model, NULL);

	if (rows <= 0)
	{
		cff_gui_beep();
		return;
	}

	fname = cff_gui_save_get_fname(&(vsave->common), &fname_free);

	if (fname == NULL)
	{
		return;
	}

	g_free(vsave->common.format.extension_orig);
	g_free(vsave->common.filechooser_utf8str);


	/* Save video ... */

	stream_idx = g_malloc(rows * sizeof(gint));

	retv = gtk_tree_model_get_iter_first(model, &iter);

	nb_streams = 0;

	while (retv == TRUE)
	{
		gtk_tree_model_get(model, &iter,
				1, &stream_idx[nb_streams], -1);
		nb_streams++;
		retv = gtk_tree_model_iter_next(model, &iter);
	}

	memset(&param, '\0', sizeof(cff_copy_param_t));

	if (vsave->common.format.u.list != NULL)
	{
		i = GuiOptionMenuGetActive(vsave->common.format.option);

		if (i > 0)
		{
			/* Array without "Auto" */
			param.format_name =
				vsave->common.format.u.list[i - 1].name;
		}

		/* malloc'ed */
		free(vsave->common.format.u.list);
	}

	n = kCFFMediaVideo;

	do
	{
		cff_mediaopts_t *mediaopts = &(vsave->media[n]);

		if (mediaopts->bsf_list != NULL)
		{
			i = GuiOptionMenuGetActive(mediaopts->bsf_option);

			if (i > 0)
			{
				/* Array without "None" */
				param.media[n].bsf_name = (mediaopts->bsf_list[i - 1]).name;
			}

			/* malloc'ed */
			free(mediaopts->bsf_list);
		}

		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mediaopts->ts_from_duration)) == TRUE)
		{
			param.media[n].ts_from_duration = 1;
		}
	}
	while (++n < kCFFMediaTotal);

	param.fname	= fname;
	param.info	= &info;
	param.func	= cff_gui_save_copy_callback;
	param.user_data	= &error;

	param.stream_idx = stream_idx;
	param.nb_streams = nb_streams;

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vsave->shortest)) == TRUE)
	{
		param.shortest = 1;
	}

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vsave->ts_keep_start)) == TRUE)
	{
		param.ts_keep_start = 1;
	}

	info	= NULL;
	gIsBusy	= TRUE;
	result	= cff_ffmpeg_copy_open(vsave->common.video, &param);
	gIsBusy	= FALSE;

	g_free(stream_idx);

	if (result == NULL)
	{
		result = error;
	}

	if (result != NULL)
	{
		gint delete;

		if (cff_gprop_get_bool(kPropBoolOptDelete, &delete) == FALSE || delete)
		{
			/* Error: remove the output file */
			g_unlink(fname);
		}
	}

#ifndef BUILD_MINGW
	if (fname_free == TRUE)
	{
		g_free(fname);
	}
#endif
	gtk_widget_destroy(vsave->common.window);
	mbar = vsave->common.mbar;
	g_free(vsave);

	if (result != NULL)
	{
		guitl_dialog_error(result, mbar);
	}

	else
	{
		guitl_dialog_info((info == NULL) ?
			gettext_nop("No info available!") : info, mbar);

		if (info != NULL)
		{
			/* malloc'ed */
			free(info);
		}

	} /* if (result != NULL) .. else */

} /* static void cff_gui_save_start(GtkWidget*, cff_save_video_t*) */

static gint cff_gui_save_delevnt(GtkWidget *w, GdkEvent *e, cff_save_video_t *vsave)
{
	gint n;

	if (vsave->common.format.u.list != NULL)
	{
		/* malloc'ed */
		free(vsave->common.format.u.list);
	}

	n = kCFFMediaVideo;

	do
	{
		cff_mediaopts_t *mediaopts = &(vsave->media[n]);

		if (mediaopts->bsf_list != NULL)
		{
			/* malloc'ed */
			free(mediaopts->bsf_list);
		}
	}
	while (++n < kCFFMediaTotal);

	g_free(vsave->common.format.extension_orig);
	g_free(vsave->common.filechooser_utf8str);
	g_free(vsave);

	return FALSE;

} /* static gint cff_gui_save_delevnt(GtkWidget*, GdkEvent*, cff_save_video_t*) */

static void cff_gui_save_destroy(GtkWidget *w, cff_save_video_t *vsave)
{
	GtkWidget *window = vsave->common.window;

	cff_gui_save_delevnt(w, NULL, vsave);
	gtk_widget_destroy(window);

} /* static void cff_gui_save_destroy(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_changed(GtkTreeSelection *selection, cff_save_video_t *vsave)
{

	gint			rows, num;
	guint			bits;
	gboolean		button, up, down, first_row, last_row;
	cff_streamlist_t	*stream;

	if (selection == vsave->avail.list.selection)
	{
		stream = &(vsave->avail);
	}

	else
	{
		stream = &(vsave->out);
	}

	num = gtk_tree_selection_count_selected_rows(selection);

	button = FALSE;

	if (num > 0)
	{
		button = TRUE;
	}

	gtk_widget_set_sensitive(stream->some, button);

	rows = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(stream->list.store), NULL);

	button = FALSE;

	if (rows > 0)
	{
		button = TRUE;
	}

	gtk_widget_set_sensitive(stream->all, button);

	if (selection == vsave->avail.list.selection)
	{
		return;
	}


	bits = vsave->common.okay.bits;

	if (button == FALSE)
	{
		bits |= 0x02;
	}

	else
	{
		bits &= ~0x02;
	}

	cff_gui_save_okay_button_set_sensitive(&(vsave->common), bits);


	first_row = FALSE;
	last_row  = FALSE;

	if (num > 0 && rows > 1)
	{
		GtkTreeModel	*model;
		GList		*list, *ptr;

		up   = TRUE;
		down = TRUE;

		list = gtk_tree_selection_get_selected_rows(selection, &model);
		ptr  = list;

		do
		{
			/*
				This is an array of integers, each representing
				a node in a tree. This value should not be freed.
			*/

			gint *i = gtk_tree_path_get_indices(ptr->data);

			if (*i == 0)
			{
				first_row = TRUE;
			}

			else if (*i == rows - 1)
			{
				last_row = TRUE;
			}

			gtk_tree_path_free(ptr->data);
		}
		while ((ptr = ptr->next) != NULL);

		g_list_free(list);
	}

	else
	{
		up   = FALSE;
		down = FALSE;

	} /* if (num > 0 && rows > 1) .. else */

	gtk_widget_set_sensitive(vsave->up,
			(first_row == FALSE) ? up : FALSE);
	gtk_widget_set_sensitive(vsave->down,
			(last_row == FALSE) ? down : FALSE);

} /* static void cff_gui_save_list_changed(GtkTreeSelection*, cff_save_video_t*) */

static GList *cff_gui_save_get_all(GtkTreeModel *model)
{
	gboolean	retv;
	GList		*list;
	GtkTreeIter	iter;

	list = NULL;
	retv = gtk_tree_model_get_iter_first(model, &iter);

	while (retv == TRUE)
	{
		list = g_list_append(list, g_memdup(&iter, sizeof(GtkTreeIter)));
		retv = gtk_tree_model_iter_next(model, &iter);
	}

	return list;

} /* static GList *cff_gui_save_get_all(GtkTreeModel*) */

static GList *cff_gui_save_get_selected(GtkTreeSelection *selection)
{
	GtkTreeModel	*model;
	GList		*list, *ptr;

	list = gtk_tree_selection_get_selected_rows(selection, &model);

	if ((ptr = list) == NULL)
	{
		return list;
	}

	do
	{
		GtkTreeIter *iter;

		iter = g_malloc(sizeof(GtkTreeIter));
		gtk_tree_model_get_iter(model, iter, ptr->data);
		gtk_tree_path_free(ptr->data);

		ptr->data = iter;
	}
	while ((ptr = ptr->next) != NULL);

	return list;

} /* static GList *cff_gui_save_get_selected(GtkTreeSelection*) */

static void cff_gui_save_list_updown(GtkWidget *w, cff_save_video_t *vsave)
{
	gint		add, i, rows, stop, *order;
	GtkTreeModel	*model;
	GtkTreePath	*path;
	GList		*list, *selected;

	add = 1;

	if (w == vsave->up)
	{
		add = -1;
	}

	list = gtk_tree_selection_get_selected_rows(vsave->out.list.selection, &model);

	if (list == NULL)
	{
		return;
	}

	rows = gtk_tree_model_iter_n_children(model, NULL);

	if (rows <= 0)
	{
		return;
	}

	order = malloc(rows * sizeof(gint));

	i = 0;

	do
	{
		order[i] = i;
	}
	while (++i < rows);

	if (add == -1)
	{
		/* move selected rows up */

		selected = list;
		i = 0;
		stop = rows;
	}

	else
	{
		/* move selected rows down */

		selected = g_list_last(list);
		i = rows - 1;
		stop = -1;
	}

	path = gtk_tree_path_new_from_indices(i, -1);

	do
	{
		if (gtk_tree_path_compare(path, selected->data) == 0)
		{
			/* exchange selected row with next (previous) row */

			gint prev = order[i + add];

			order[i + add] = i;
			order[i] = prev;

			gtk_tree_path_free(selected->data);

			if (add == -1)
			{
				selected = selected->next;
			}

			else
			{
				selected = selected->prev;
			}

			if (selected == NULL)
			{
				break;
			}

		} /* if (gtk_tree_path_compare(path, selected->data) == 0) */

		if (add == -1)
		{
			gtk_tree_path_next(path);
		}

		else
		{
			gtk_tree_path_prev(path);
		}
	}
	while ((i -= add) != stop);

	gtk_tree_path_free(path);
	g_list_free(list);

	gtk_list_store_reorder(vsave->out.list.store, order);
	g_free(order);

	/* Set buttons (up, down) sensitive or insensitive */
	cff_gui_save_list_changed(vsave->out.list.selection, vsave);

} /* static void cff_gui_save_list_updown(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_leftright(GtkWidget *w, cff_save_video_t *vsave)
{
	GList		*list, *ptr;
	cff_ListStore_t	*src, *dst;

	if (w == vsave->avail.some || w == vsave->avail.all)
	{
		src = &(vsave->avail.list);
		dst = &(vsave->out.list);
	}

	else
	{
		src = &(vsave->out.list);
		dst = &(vsave->avail.list);
	}

	if (w == vsave->avail.all || w == vsave->out.all)
	{
		list = cff_gui_save_get_all(GTK_TREE_MODEL(src->store));
	}

	else
	{
		list = cff_gui_save_get_selected(src->selection);
	}

	if ((ptr = list) == NULL)
	{
		return;
	}

	do
	{
		gint		stream_idx;
		gchar		*str;
		GtkTreeIter	*iter;

		iter = ptr->data;

		gtk_tree_model_get(GTK_TREE_MODEL(src->store), iter,
						0, &str, 1, &stream_idx, -1);

		gtk_list_store_remove(src->store, iter);

		gtk_list_store_append(dst->store, iter);
		gtk_list_store_set(dst->store, iter, 0, str, 1, stream_idx, -1);

		g_free(str);
		g_free(iter);
	}
	while ((ptr = ptr->next) != NULL);

	g_list_free(list);

	/*
		Set buttons (up, down, right, right-all,
		left, left-all and okay) sensitive or insensitive
	*/

	cff_gui_save_list_changed(src->selection, vsave);
	cff_gui_save_list_changed(dst->selection, vsave);

} /* static cff_gui_save_list_leftright(GtkWidget*, cff_save_video_t*) */

static gint cff_gui_video_list_focus(GtkWidget *list, GdkEvent *e, gpointer data)
{
	if (gtk_widget_is_focus(list) == FALSE)
	{
		gtk_widget_grab_focus(list);
	}

	return FALSE;

} /* static gint cff_gui_video_list_focus(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_build_list(const gchar *title, cff_ListStore_t *list, GtkWidget *hbox)
{
	gchar			*str;
	GtkWidget		*frame, *swin;
	GtkTreeViewColumn	*column;
	GtkCellRenderer		*renderer;

	list->store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_INT);

	list->view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(list->store));

	GuiSignalConnect(GTK_WIDGET(list->view), "button-press-event",
						cff_gui_video_list_focus, NULL);

	gtk_widget_set_can_focus(GTK_WIDGET(list->view), TRUE);

	list->selection = gtk_tree_view_get_selection(list->view);
	gtk_tree_selection_set_mode(list->selection, GTK_SELECTION_MULTIPLE);

	column = gtk_tree_view_column_new();

	str = cff_charconv_utf8_gettext(title);
	gtk_tree_view_column_set_title(column, str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", 0, NULL);

	gtk_tree_view_append_column(list->view, column);
	g_object_set(list->view,
			"headers-visible", TRUE,
			"headers-clickable", FALSE,
			NULL);

	g_object_unref(G_OBJECT(list->store));

	frame = gtk_frame_new(NULL);
	GuiHBoxPackStart(hbox, frame, TRUE, TRUE, 4);
	
	swin = guitl_scrolled_window_new(frame, -1, 150, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(swin), GTK_WIDGET(list->view));

} /* static void cff_gui_build_list(const gchar*, cff_ListStore_t*, GtkWidget*) */

static GtkWidget *cff_gui_build_option_menu(GtkWidget *vbox,
		const gchar *title, const gchar *entry, const gchar **name_list)
{
	gchar		*str;
	const gchar	*name;
	GtkWidget	*option, *frame, *hbox;

#if GTK_MAJOR_VERSION == 2
	gint		i;
	GtkWidget	*menu;
#endif

	option = GuiOptionMenuTextNew();
	gtk_container_set_border_width(GTK_CONTAINER(option), 4);

	hbox = GuiHBoxNew(FALSE, 2);
	GuiVBoxPackStart(vbox, hbox, TRUE, FALSE, 2);

	frame = guitl_frame_new_add_child_gettext(title, option);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 2);

#if GTK_MAJOR_VERSION == 2
	menu = gtk_menu_new();
	gtk_widget_realize(menu);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(option), menu);
	i = 0;
#endif

	if (name_list == NULL || name_list[0] == NULL)
	{
		name_list = NULL;
		entry = gettext_nop("Empty");
	}

	if (entry != NULL)
	{
		str = cff_charconv_utf8_gettext(entry);
		GuiOptionMenuTextAppend(option, menu, 0, str);
		CFF_CHARCONV_UTF8_FREE(str, entry);
	}

	if (name_list == NULL)
	{
		gtk_widget_set_sensitive(option, FALSE);
	}

	else
	{
		while ((name = *(name_list++)) != NULL)
		{
#if GTK_MAJOR_VERSION == 2
			i++;
#endif
			GuiOptionMenuTextAppend(option, menu, i, name);
		}

	} /* if (name_list == NULL) .. else */

	GuiOptionMenuSetActive(option, 0);

	return option;

} /* static GtkWidget *cff_gui_build_option_menu(GtkWidget*, const gchar*, const gchar*, const gchar**) */

static GtkWidget *cff_gui_notebook_append_page(GtkWidget *notebook, const gchar *title)
{
	GtkWidget *label, *vbox;

	vbox = GuiVBoxNew(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 2);

	label = guitl_label_new_gettext(title);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);

	return vbox;

} /* static GtkWidget *cff_gui_notebook_append_page(GtkWidget*, const gchar*) */

static void cff_gui_build_media_option(GtkWidget *notebook, cff_save_video_t *vsave, gint media)
{
	gchar		*str;
	gchar		**ptr;
	const gchar	*title;
	GtkWidget	*vbox;
	cff_mediaopts_t *mediaopts;
	cff_bsflist_t	*bsf_list, *bsf_info;
	gchar		*buf[512];

	if (media == kCFFMediaVideo)
	{
		title = gettext_nop("Video");
	}

	else if (media == kCFFMediaAudio)
	{
		title = gettext_nop("Audio");
	}

	else
	{
		title = gettext_nop("Other");
	}

	vbox = cff_gui_notebook_append_page(notebook, title);

	mediaopts = &(vsave->media[media]);

	/* Check button */

	mediaopts->ts_from_duration =
		guitl_vbox_add_check_button_gettext(vbox,
			gettext_nop("Create timestamps from packet duration"));

	/* Option menu */

	bsf_list = cff_ffmpeg_get_bitstream_filter_names(vsave->common.video, media);
	mediaopts->bsf_list = bsf_list;

	ptr = buf;

	if ((bsf_info = bsf_list) != NULL)
	{
		while (bsf_info->name != NULL)
		{
			if (ptr - buf < (sizeof(buf) / sizeof(gchar*)) - 1)
			{
				*(ptr++) = g_strconcat(bsf_info->name, " ",
							bsf_info->codecs, NULL);
			}

			if (bsf_info->codecs != NULL)
			{
				/* malloc'ed */
				free(bsf_info->codecs);
				bsf_info->codecs = NULL;
			}

			bsf_info++;
		}

	} /* if ((bsf_info = bsf_list) != NULL) */

	*ptr = NULL;

	mediaopts->bsf_option = cff_gui_build_option_menu(vbox,
					gettext_nop("Bitstream Filter"),
					gettext_nop("None"),
					(const gchar **) buf);
	/* Free strings in array */

	ptr = buf;

	while ((str = *(ptr++)) != NULL)
	{
		g_free(str);
	}

} /* static void cff_gui_build_media_option(GtkWidget*, cff_save_video_t*, gint) */


/*
	------------
	 SAVE IMAGE
	------------
*/

static void cff_gui_image_keep_dimension_toggled(GtkToggleButton *button, cff_save_image_t *vimage)
{
	gint			n;
	gboolean		active;
	cff_spinbutton_t	*spinbutton;

	active = ! gtk_toggle_button_get_active(button);

	spinbutton = &(vimage->width);

	n = 2;

	do
	{
		gint value;

		value = gtk_spin_button_get_value_as_int(spinbutton->widget);
		gtk_spin_button_set_value(spinbutton->widget,
						(gdouble) spinbutton->prev_value);
		spinbutton->prev_value = value;

		gtk_widget_set_sensitive(GTK_WIDGET(spinbutton->widget), active);
		spinbutton = &(vimage->height);
	}
	while (--n > 0);

} /* static void ccff_gui_image_keep_dimension_toggled(GtkToggleButton*, cff_save_image_t*) */

static GtkSpinButton *cff_gui_image_spinbutton(GtkWidget *hbox, gint value, const gchar *title)
{
	GtkAdjustment	*adjustment;
	GtkWidget	*widget, *bbox, *frame;
	GtkSpinButton	*spinbutton;

	bbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);

	frame = guitl_frame_new_add_child_gettext(title, bbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 4);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 4);

	adjustment = (GtkAdjustment *)
			gtk_adjustment_new((gdouble) value,
						1.0,
						99999.0,
						1.0,
						10.0,
						0.0);

	widget = gtk_spin_button_new(adjustment, 1.0, 0);
	GuiHBoxPackStart(bbox, widget, FALSE, FALSE, 4);

	spinbutton = GTK_SPIN_BUTTON(widget);
	gtk_spin_button_set_numeric(spinbutton, TRUE);
	gtk_spin_button_set_update_policy(spinbutton, GTK_UPDATE_IF_VALID);

	return spinbutton;

} /* static GtkSpinButton *cff_gui_image_spinbutton(GtkWidget*, gint, const gchar*) */

static void cff_gui_save_image(GtkWidget *w, cff_save_image_t *vimage)
{
	gboolean	fname_free;
	gchar		*fname, *info;
	const gchar	*error;
	GtkWidget	**mbar;
	cff_image_param_t param;

	fname = cff_gui_save_get_fname(&(vimage->common), &fname_free);

	if (fname == NULL)
	{
		return;
	}

	g_free(vimage->common.format.extension_orig);
	g_free(vimage->common.filechooser_utf8str);


	/* Save image ... */

	memset(&param, '\0', sizeof(cff_image_param_t));

	param.fname = fname;
	param.info  = &info;

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vimage->keep_dimension)) == TRUE)
	{
		param.keep_dimension = 1;
	}

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vimage->keep_ratio)) == TRUE)
	{
		param.keep_ratio = 1;
	}

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vimage->gray)) == TRUE)
	{
		param.gray = 1;
	}

	if (param.keep_dimension == FALSE)
	{
		param.width  = gtk_spin_button_get_value_as_int(vimage->width.widget);
		param.height = gtk_spin_button_get_value_as_int(vimage->height.widget);
	}

	if (vimage->common.format.u.names != NULL)
	{
		gint i = GuiOptionMenuGetActive(vimage->common.format.option);

		if (i > 0)
		{
			/* Array without "Auto" */
			param.format_name = vimage->common.format.u.names[i - 1];
		}

		/* malloc'ed */
		free(vimage->common.format.u.names);
	}

	info	= NULL;
	gIsBusy	= TRUE;
	error	= cff_ffmpeg_save_image(vimage->common.video, &param);
	gIsBusy	= FALSE;

	if (error != NULL)
	{
		gint delete;

		if (cff_gprop_get_bool(kPropBoolOptDelete, &delete) == FALSE || delete)
		{
			/* Error: remove the image */
			g_unlink(fname);
		}
	}

#ifndef BUILD_MINGW
	if (fname_free == TRUE)
	{
		g_free(fname);
	}
#endif
	gtk_widget_destroy(vimage->common.window);
	mbar = vimage->common.mbar;
	g_free(vimage);

	if (error != NULL)
	{
		guitl_dialog_error(error, mbar);
	}

	else
	{
		guitl_dialog_info((info == NULL) ?
			gettext_nop("No info available!") : info, mbar);

		if (info != NULL)
		{
			/* malloc'ed */
			free(info);
		}

	} /* if (error != NULL) .. else */

} /* static void cff_gui_save_image(GtkWidget*, cff_save_image_t*) */

static gint cff_gui_image_delevnt(GtkWidget *w, GdkEvent *e, cff_save_image_t *vimage)
{
	if (vimage->common.format.u.names != NULL)
	{
		/* malloc'ed */
		free(vimage->common.format.u.names);
	}

	g_free(vimage->common.format.extension_orig);
	g_free(vimage->common.filechooser_utf8str);
	g_free(vimage);

	return FALSE;

} /* static gint cff_gui_image_delevnt(GtkWidget*, GdkEvent*, cff_save_image_t*) */

static void cff_gui_image_destroy(GtkWidget *w, cff_save_image_t *vimage)
{
	GtkWidget *window = vimage->common.window;

	cff_gui_image_delevnt(w, NULL, vimage);
	gtk_widget_destroy(window);

} /* static void cff_gui_image_destroy(GtkWidget*, cff_save_image_t*) */


/*
	--------------
	 INFO WINDOWS
	--------------
*/

static void cff_gui_scrolled_info_window(const char *title, GtkWidget *ibox, GtkWidget **mbar, int prop)
{
	GtkWidget *button, *hbox, *vbox, *frame, *swin, *window;

	vbox = GuiVBoxNew(FALSE, 4);
	window = guitl_window_new_gettext(title, vbox, 8);
	guitl_dialog_set_modal(window, gWindata.mainwin, mbar);

	if (prop >= 0)
	{
		guitl_window_setup(window, prop);
	}

	frame = gtk_frame_new(NULL);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 4);

	swin = guitl_scrolled_window_new(frame, 320, 180, GTK_POLICY_AUTOMATIC);
	gtk_container_set_border_width(GTK_CONTAINER(ibox), 8);

#if GTK_MAJOR_VERSION == 2
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(swin), ibox);
#else
	/* Above is deprecated since version 3.8 */
	gtk_container_add(GTK_CONTAINER(swin), ibox);
#endif

	hbox = guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	button = guitl_button_box_add_button(hbox,
				gettext_nop("OK"), NULL, NULL);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, window);
	gtk_widget_grab_default(button);

	gtk_widget_show_all(window);

} /* static void cff_gui_scrolled_info_window(const char*, GtkWidget*, GtkWidget**, int) */

static void cff_gui_ffmpeg_info_label(const char *title, char *str, GtkWidget *parent)
{
	GtkWidget *frame, *label;

	if (str == NULL || str[0] == '\0')
	{
		label = guitl_label_new_gettext(gettext_nop("No info available!"));
	}

	else
	{
		gchar			*text;
		PangoFontDescription	*desc;
#if GTK_MAJOR_VERSION > 2
		PangoAttribute		*attr;
		PangoAttrList		*attrlist;
#endif
		text = cff_charconv_utf8_from_locale(str, NULL);

		label = gtk_label_new(text);
		CFF_CHARCONV_UTF8_FREE(text, str);

		desc = pango_font_description_new();
		pango_font_description_set_family(desc, "monospace");

#if GTK_MAJOR_VERSION == 2
		gtk_widget_modify_font(label, desc);
#else
		attr = pango_attr_font_desc_new(desc);
		attrlist = pango_attr_list_new();
		pango_attr_list_insert(attrlist, attr);
		gtk_label_set_attributes(GTK_LABEL(label), attrlist);
		pango_attr_list_unref(attrlist);
#endif
		pango_font_description_free(desc);

	} /* if (str == NULL || str[0] == '\0') .. else */

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	GuiMiscSetAlignment(label, Gui_ALIGN_START, Gui_ALIGN_CENTER);
	GuiMiscSetPadding(label, 15);

	frame = guitl_frame_new_add_child_gettext(title, label);
	GuiVBoxPackStart(parent, frame, FALSE, FALSE, 5);

} /* static void cff_gui_ffmpeg_info_label(const char*, char*, GtkWidget*) */


/*
	-----
	 CSS
	-----
*/

#if GTK_MAJOR_VERSION > 2

static gboolean cff_gui_css_load_from_data(const gchar *str)
{
	GtkCssProvider	*provider;
	GdkDisplay	*display;
	GdkScreen	*screen;
	GError		*error;
 
	display  = gdk_display_get_default();
	screen   = gdk_display_get_default_screen(display);
	provider = gtk_css_provider_new();

	gtk_style_context_add_provider_for_screen(screen,
				GTK_STYLE_PROVIDER(provider),
				GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
	error = NULL;
	gtk_css_provider_load_from_data(GTK_CSS_PROVIDER(provider),
				str,
				-1,
				&error);

	g_object_unref(provider);

	if (error != NULL)
	{
		g_error_free(error);
		return FALSE;
	}

	return TRUE;

} /* static gboolean cff_gui_css_load_from_data(const gchar*) */

#endif /* GTK_MAJOR_VERSION > 2 */


/*
	------------
	 LOG-WINDOW
	------------
*/

static gint cff_gui_logwindow_color_parse(gint prop, gpointer color)
{
	/* Return 'FALSE' if color is set to the default */

	gchar *colorstr, *defcolor;

	if (prop == kPropStrLogWinText)
	{
		defcolor = (gchar *) gLogWinTextColor;
	}

	else
	{
		defcolor = (gchar *) gLogWinBackgroundColor;
	}

	colorstr = cff_gprop_get_str(prop);

	if (colorstr == NULL)
	{
		colorstr = defcolor;
	}

	while (GuiColorParse(colorstr, color) == FALSE)
	{
		if (colorstr == defcolor)
		{
			return FALSE;
		}

		g_free(colorstr);
		colorstr = defcolor;
	}

	if (colorstr != defcolor)
	{
		g_free(colorstr);
		return TRUE;
	}

	return FALSE;

} /* static gint cff_gui_logwindow_color_parse(gint, gpointer) */

static void cff_gui_logwindow_color_set(cff_logwindow_t *log)
{
	gchar		*text;

#if GTK_MAJOR_VERSION == 2
	GdkWindow	*gdkwindow;
#else
	guint		i;
	gchar		*background, buf[256];
#endif

	text = cff_gprop_get_str(kPropStrLogWinText);

	if (text == NULL)
	{
		text = (gchar *) gLogWinTextColor;
	}

#if GTK_MAJOR_VERSION == 2

	gdkwindow = gtk_text_view_get_window(GTK_TEXT_VIEW(log->textview),
							GTK_TEXT_WINDOW_TEXT);
	if (gdkwindow != NULL)
	{
		GdkColormap *colormap;

		if (cff_gui_logwindow_color_parse(kPropStrLogWinBackground,
							&(log->bg)) == FALSE)
		{
			if (text != gLogWinTextColor)
			{
				g_free(text);
				text = (gchar *) gLogWinTextColor;
			}
		}

		colormap = gtk_widget_get_colormap(log->textview);
		gdk_rgb_find_color(colormap, &(log->bg));
		gdk_window_set_background(gdkwindow, &(log->bg));

		g_object_set(G_OBJECT(log->tag), "foreground", text, NULL);

	} /* if (gdkwindow != NULL) */

	if (text != gLogWinTextColor)
	{
		g_free(text);
	}

#else /* GTK_MAJOR_VERSION != 2 */

	background = cff_gprop_get_str(kPropStrLogWinBackground);

	if (background == NULL)
	{
		background = (gchar *) gLogWinBackgroundColor;
	}

	i = 2;

	do
	{
		g_snprintf(buf, sizeof(buf),
				"GtkTextView#CutterFF-log *,"
				"textview.view#CutterFF-log > text {"
				"  color: %s;"
				"  background-color: %s;"
				"  border: Black;"
				"  border-width: 1px;"
				"  border-style: Solid;"
				"}",
				text,
				background);

		if (text != gLogWinTextColor)
		{
			g_free(text);
			text = (gchar *) gLogWinTextColor;
		}

		if (background != gLogWinBackgroundColor)
		{
			g_free(background);
			background = (gchar *) gLogWinBackgroundColor;
		}

		if (cff_gui_css_load_from_data(buf) == TRUE)
		{
			break;
		}
	}
	while (--i > 0);

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */

} /* static void cff_gui_logwindow_color_set(cff_logwindow_t*) */

static GtkWidget *cff_gui_logwindow_colorbutton_new(gint prop, const gchar *title, gpointer color)
{
	gchar		*str;
	GtkWidget	*button, *label, *hbox, *vbox;
	GuiSignalFunc	func;

	vbox = GuiVBoxNew(FALSE, 5);
	hbox = GuiHBoxNew(FALSE, 5);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 3);

	cff_gui_logwindow_color_parse(prop, color);
	button = GuiColorButtonNewWithColor(color);
	GuiHBoxPackStart(hbox, button, TRUE, FALSE, 5);

	str = cff_charconv_utf8_gettext(title);
	gtk_color_button_set_title(GTK_COLOR_BUTTON(button), str);

#if GTK_MAJOR_VERSION == 2
	func = (GuiSignalFunc) gtk_color_button_get_color;
#else
	func = (GuiSignalFunc) gtk_color_chooser_get_rgba;
#endif

	GuiSignalConnect(button, "color-set", func, color);

	label = gtk_label_new(str);
	GuiVBoxPackStart(vbox, label, FALSE, FALSE, 3);

	CFF_CHARCONV_UTF8_FREE(str, title);

	return vbox;

} /* static GtkWidget *cff_gui_logwindow_colorbutton_new(gint, const gchar*, gpointer) */

static void cff_gui_logwindow_color(GtkWidget *w, cff_logwindow_t *log)
{
	static const gchar btn_cancel[] = gettext_nop("Cancel");
	static const gchar btn_okay[]   = gettext_nop("OK");

	gint		retv;
	gchar		*accept, *cancel, *text, *background;
	GtkWidget	*content, *frame, *hbox, *widget;
	GtkDialog	*dialog;

#if GTK_MAJOR_VERSION == 2
	GdkColor	fg, bg;
#else
	GdkRGBA		fg, bg;
#endif

	cancel = cff_charconv_utf8_gettext(btn_cancel);
	accept = cff_charconv_utf8_gettext(btn_okay);

	widget = gtk_dialog_new_with_buttons("CutterFF: Color",
		GTK_WINDOW(gWindata.topwin),
		GTK_DIALOG_MODAL,
		cancel, GTK_RESPONSE_CANCEL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	CFF_CHARCONV_UTF8_FREE(accept, btn_okay);
	CFF_CHARCONV_UTF8_FREE(cancel, btn_cancel);

	gtk_window_set_resizable(GTK_WINDOW(widget), FALSE);

	dialog = GTK_DIALOG(widget);

	hbox = GuiHBoxNew(TRUE, 5);
	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);

	widget = cff_gui_logwindow_colorbutton_new(kPropStrLogWinText,
						gettext_nop("Text"), &fg);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	widget = cff_gui_logwindow_colorbutton_new(kPropStrLogWinBackground,
						gettext_nop("Background"), &bg);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	gtk_widget_show_all(frame);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), frame);

	retv = gtk_dialog_run(dialog);

	gtk_widget_destroy(GTK_WIDGET(dialog));

	if (retv != GTK_RESPONSE_ACCEPT)
	{
		return;
	}

	text = GuiColorToString(&fg);
	cff_gprop_set_str(kPropStrLogWinText, text);
	g_free(text);

	background = GuiColorToString(&bg);
	cff_gprop_set_str(kPropStrLogWinBackground, background);
	g_free(background);

	cff_gui_logwindow_color_set(log);

} /* static void cff_gui_logwindow_color(GtkWidget*, cff_logwindow_t*) */

static void cff_gui_logwindow_popup(GtkTextView *textview, GtkMenu *menu, cff_logwindow_t *log)
{
	/* Append to the GtkTextView popup */

	GtkWidget *mi;

	mi = GuiMenuSeparatorNew();
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

	mi = guitl_menu_item_new_gettext(gettext_nop("Color ..."));
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

	GuiSignalConnect(mi, "activate", cff_gui_logwindow_color, log);

} /* static void cff_gui_logwindow_popup(GtkTextView*, GtkMenu*, cff_logwindow_t*) */


#if defined(BUILD_MINGW) && GTK_MAJOR_VERSION == 2

static gboolean cff_gui_logwindow_copy_clipboard_cb(gpointer data)
{
	struct
	{
		gchar	*str;
		gint	len;
		gint	lines;

	} *s = data;

	if (s == NULL)
	{
		return FALSE;
	}

	if (s->str == NULL)
	{
		g_free(s);
		return FALSE;
	}

	if (s->lines >= 0 && s->len > 0 && s->str[0] != '\0')
	{
		if (OpenClipboard(NULL))
		{
			HGLOBAL	clipbuffer;

			EmptyClipboard();
			clipbuffer = GlobalAlloc(GMEM_MOVEABLE, s->len + s->lines + 1);

			if (clipbuffer != NULL)
			{
				char *dst = (char *) GlobalLock(clipbuffer);

				if (dst != NULL)
				{
					/* UNIX to DOS : LF to CR-LF */

					unsigned int	chr;
					char		*src;

					src = s->str;

					do
					{
						chr = *(src++);

						if (chr == 0x0A)
						{
							*(dst++) = 0x0D;
						}

						*(dst++) = chr;
					}
					while (chr != '\0');

					GlobalUnlock(clipbuffer);
					SetClipboardData(CF_TEXT, clipbuffer);
				}

			} /* if (clipbuffer != NULL) */

			CloseClipboard();

		} /* if (OpenClipboard(NULL)) */

	} /* if (s->lines >= 0 && s->len > 0 && s->str[0] != '\0') */

	g_free(s->str);
	g_free(s);

	return FALSE;

} /* static gboolean cff_gui_logwindow_copy_clipboard_cb(gpointer) */

static void cff_gui_logwindow_copy_clipboard(GtkTextView *textview, gpointer data)
{
	/* Copying to the global clipwork do not work with GTK-2 and Windows. */

	struct
	{
		gchar	*str;
		gint	len;
		gint	lines;

	} s;

	GtkTextBuffer	*buffer;
	GtkTextIter	start, end;

	buffer = gtk_text_view_get_buffer(textview);

	if (gtk_text_buffer_get_selection_bounds(buffer, &start, &end) == FALSE)
	{
		return;
	}

	s.str = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

	if (s.str == NULL)
	{
		return;
	}

	s.len = strlen(s.str);

	if (s.len == 0)
	{
		return;
	}

	s.lines  = gtk_text_iter_get_line(&end);
	s.lines -= gtk_text_iter_get_line(&start);

	if (s.lines < 0)
	{
		return;
	}

	/* Wait a litle bit, else copying does not work */

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_logwindow_copy_clipboard_cb,
					g_memdup(&s, sizeof(s)), NULL);

} /* static void cff_gui_logwindow_copy_clipboard(GtkTextView*, gpointer) */

#endif /* #if defined(BUILD_MINGW) && GTK_MAJOR_VERSION == 2 */


static gboolean cff_gui_logwindow_mainthread(gpointer data)
{
	gint		lines;
	guint		lock;
	GtkTextBuffer	*buffer;
	cff_logwindow_t	*log;
	GtkTextIter	end;

	log = data;

	if (log == NULL || log->window == NULL)
	{
		return Gui_SOURCE_REMOVE;
	}

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(log->textview));
	lines  = gtk_text_buffer_get_line_count(buffer);

	if (lines > 1000)
	{
		GtkTextIter start;

		gtk_text_buffer_get_start_iter(buffer, &start);
		gtk_text_buffer_get_iter_at_line(buffer, &end, lines - 1000);

		gtk_text_buffer_delete(buffer, &start, &end);
	}

	gtk_text_buffer_get_end_iter(buffer, &end);

	lock = log->nThreads;

	if (lock)
	{
		GuiMutexLock(&(log->mutex));
	}

	gtk_text_buffer_insert_with_tags(buffer, &end,
			log->utf8.str, log->utf8.len, log->tag, NULL);

	log->is_mainthread = TRUE;

	while (gtk_events_pending() == TRUE)
	{
		gtk_main_iteration_do(FALSE);

		if (log->window == NULL || log->is_progressbar)
		{
			break;
		}
	}

	log->is_mainthread = FALSE;

	/* Window still open ? */

	if (log->window != NULL)
	{
		if (log->utf8.max > 4096)
		{
			log->utf8.max = 4096;
			log->utf8.str = g_realloc(log->utf8.str, log->utf8.max);
		}

		log->utf8.len = 0;

		gtk_text_buffer_get_end_iter(buffer, &end);
		gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(log->textview),
						&end, 0.0, FALSE, 0.0, 0.0);

	}

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_logwindow_mainthread(cff_logwindow_t*) */

static int cff_gui_logwindow_cb(const char *str, int len, void *data)
{
	gint		slen;
	guint		lock, max;
	gchar		*utf8;
	cff_logwindow_t	*log;

	log = data;

	if (log == NULL || log->window == NULL)
	{
		return -1;
	}

	slen = len;
	utf8 = cff_charconv_utf8_from_locale(str, &slen);

	lock = log->nThreads;

	if (lock)
	{
		if (GuiMutexTryLock(&(log->mutex)) == FALSE)
		{
			if (log->window == NULL)
			{
				CFF_CHARCONV_UTF8_FREE(utf8, str);
				return -1;
			}

			GuiMutexLock(&(log->mutex));
		}
	}

	slen += log->utf8.len;
	max   = log->utf8.max;

	if (max == 0)
	{
		max = 4096;
	}

	while (slen + 1 > max)
	{
		max <<= 1;
	}

	if (max > log->utf8.max)
	{
		log->utf8.max = max;
		log->utf8.str = g_realloc(log->utf8.str, max);
	}

	strcpy(&log->utf8.str[log->utf8.len], utf8);
	CFF_CHARCONV_UTF8_FREE(utf8, str);

	len = log->utf8.len;
	log->utf8.len = slen;

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	if (len == 0)
	{
		g_main_context_invoke_full(NULL, G_PRIORITY_HIGH,
			(GSourceFunc) cff_gui_logwindow_mainthread, log, NULL);
	}

	return 0;

} /* static int cff_gui_logwindow_cb(const char*, int, void*) */

static gboolean cff_gui_logwindow_free(GtkWidget *widget, GdkEvent *e, gpointer data)
{
	guint		lock;
	const gchar	*logname;
	cff_logwindow_t *log;

	if (widget == (gpointer) -1)
	{
		log = &gLogWindow;
	}

	else
	{
		log = data;
	}

	lock = log->nThreads;


	if (lock)
	{
		/*
			If the log-window is currently busy,
			and 'nThreads' is not zero, 'ffmpeg'
			runs in another thread.
		*/

#ifndef HAVE_SEMAPHORE

		/* Keep the window open */
		return TRUE;
#else
		/* Close the window */
		log->window = NULL;

		if (log->is_mainthread ||
			GuiMutexTryLock(&(log->mutex)) == FALSE)
		{

			if (log->sigid != 0)
			{
				/* Try again */
				return Gui_SOURCE_CONTINUE;
			}

			log->sigid  = g_idle_add_full(G_PRIORITY_LOW,
				(GSourceFunc) cff_gui_logwindow_free,
							(gpointer) -1, NULL);
			/* Destroy the window */
			return FALSE;
		}

#endif /* #ifndef HAVE_SEMAPHORE .. #else */

	} /* if (lock) */


	g_free(log->utf8.str);
	log->utf8.str = NULL;
	log->utf8.len = 0;
	log->utf8.max = 0;

	if (widget != NULL)
	{
		/* Destroyed */
		log->window = NULL;
	}

	/* Restore previous level */
	cff_ffmpeg_set_loglevel(NULL, lock, NULL, NULL);

	logname = cff_ffmpeg_get_loglevel_name(0);
	cff_gprop_set_str(kPropStrOptLoglevel, logname);

	log->is_mainthread = FALSE;
	log->sigid = 0;

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	return FALSE;

} /* static gboolean cff_gui_logwindow_free(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_logwindow_set_title(GtkWidget *window, const gchar *logname)
{
	static const gchar logwin_title[] = "Cutterff: FFmpeg-Log (%s)";

	if (logname != NULL)
	{
		gchar buf[256];

		snprintf(buf, sizeof(buf), logwin_title, logname);
		gtk_window_set_title(GTK_WINDOW(window), buf);
	}

} /* static void cff_gui_logwindow_set_title(GtkWidget*, const gchar*) */

static gboolean cff_gui_logwindow_open(const gchar *logname)
{
	GtkWidget	*swin;
	GtkTextBuffer	*buffer;
	cff_logwindow_t	*log;

	log = &gLogWindow;

	if (cff_ffmpeg_set_loglevel(logname, 0, cff_gui_logwindow_cb, log) < 0)
	{
		return FALSE;
	}

	if (log->window != NULL)
	{
		cff_gui_logwindow_set_title(log->window, logname);
		return TRUE;
	}

	log->textview = gtk_text_view_new();

	g_object_set(G_OBJECT(log->textview), "cursor-visible", FALSE,
						"editable", FALSE,
						"wrap-mode", GTK_WRAP_NONE,
						NULL);

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(log->textview));
	log->tag = gtk_text_buffer_create_tag(buffer, NULL,
					"family", "monospace",
				   	 NULL);

	swin = guitl_scrolled_window_new(NULL, 320, 180, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(swin), log->textview);

	log->window = guitl_window_new(swin, 8);
	cff_gui_logwindow_set_title(log->window, logname);
	guitl_window_setup(log->window, kPropIntLogWinWidth);

	GuiSignalConnect(log->window, "delete-event",
					cff_gui_logwindow_free, log);

	cff_gui_signal_connect_key_or_mouse_button(log->window, kButtonPressEvent);
	cff_gui_signal_connect_key_or_mouse_button(log->window, kKeyPressEvent);

	GuiSignalConnect(log->textview, "populate-popup",
					cff_gui_logwindow_popup, log);

#if defined(BUILD_MINGW) && GTK_MAJOR_VERSION == 2
	GuiSignalConnect(log->textview, "copy-clipboard",
					cff_gui_logwindow_copy_clipboard, NULL);
#endif

	cff_gui_signal_connect_key_or_mouse_button(log->textview, kButtonPressEvent);
	cff_gui_signal_connect_key_or_mouse_button(log->textview, kKeyPressEvent);

#if GTK_MAJOR_VERSION > 2
	gtk_widget_set_name(log->textview, "CutterFF-log");
#endif

	gtk_widget_show_all(log->window);

	cff_gui_logwindow_color_set(log);

	return TRUE;

} /* static gboolean cff_gui_logwindow_open(const gchar*) */


/*
	-------
	 CAIRO
	-------
*/

#if GTK_MAJOR_VERSION > 2

static gboolean _cff_gui_cairo_surface_create(GtkWidget *widget, cairo_surface_t **surface, const cff_video_t *video)
{
	GdkWindow	*wnd;
	cairo_surface_t	*ptr;

	ptr = *surface;

	if (ptr != NULL)
	{
		cairo_surface_destroy(ptr);
	}

	wnd = gtk_widget_get_window(widget);

	if (video == NULL)
	{
		ptr = gdk_window_create_similar_surface(
					wnd,
					CAIRO_CONTENT_COLOR,
					gtk_widget_get_allocated_width(widget),
					gtk_widget_get_allocated_height(widget));
	}

	else
	{
		gCairo.rgb24_width  = video->vframe.rgb24.width;
		gCairo.rgb24_height = video->vframe.rgb24.height;

		ptr = gdk_window_create_similar_image_surface(
					wnd,
					CAIRO_FORMAT_RGB24,
					video->vframe.rgb24.width,
					video->vframe.rgb24.height,
					0); /* scale */

	} /* if (video == NULL) .. else */

	if (ptr != NULL &&
		cairo_surface_status(ptr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy(ptr);
		ptr = NULL;
	}

	*surface = ptr;

	return (ptr == NULL) ? FALSE : TRUE;

} /* static gboolean _cff_gui_cairo_surface_create(GtkWidget*, cairo_surface_t**, const cff_video_t*) */

static gboolean cff_gui_cairo_surface_create(GtkWidget *widget, cairo_surface_t **surface)
{
	return _cff_gui_cairo_surface_create(widget, surface, NULL);

} /* static gboolean cff_gui_cairo_surface_create(GtkWidget*, cairo_surface_*t*) */

static gboolean cff_gui_cairo_surface_image_create(GtkWidget *widget, cairo_surface_t **surface, const cff_video_t *video)
{
	return _cff_gui_cairo_surface_create(widget, surface, video);

} /* static gboolean cff_gui_cairo_surface_image_create(GtkWidget*, cairo_surface_t**, const cff_video_t*) */

static void cff_gui_cairo_surface_image_fill(cairo_surface_t *surface, const cff_video_t *video)
{
	int i;
	unsigned char z, *src, *dst;

	if (surface == NULL)
	{
		return;
	}

#if G_BYTE_ORDER == G_BIG_ENDIAN
	src = video->vframe.rgb24.data;
#else
	src = &(video->vframe.rgb24.data[3]);
#endif

	dst = cairo_image_surface_get_data(surface);

	i = video->vframe.rgb24.width * video->vframe.rgb24.height;
	z = 0;

	do
	{

#if G_BYTE_ORDER == G_BIG_ENDIAN
		*(dst++) = z;
		*(dst++) = *(src++);
		*(dst++) = *(src++);
		*(dst++) = *(src++);
#else
		*(dst++) = *(--src);
		*(dst++) = *(--src);
		*(dst++) = *(--src);
		*(dst++) = z;
		src += 6;
#endif

	}
	while (--i > 0);

} /* static void cff_gui_cairo_surface_image_fill(cairo_surface_t*, const cff_video_t*) */

static void cff_gui_cairo_draw_rect(cairo_t *cr, gint color,
			gint fill, gint x, gint y, gint width, gint height)
{
	GdkRectangle rect;

	rect.x = (gdouble) x;
	rect.y = (gdouble) y;
	rect.width  = (gdouble) width;
	rect.height = (gdouble) height;

	gdk_cairo_set_source_rgba(cr, &gCairo.RGBA[color]);
	gdk_cairo_rectangle(cr, &rect);

	if (fill)
	{
		cairo_fill(cr);
	}

	else
	{
		cairo_set_line_width(cr, 1.0);
		cairo_stroke(cr);
	}

} /* static void cff_gui_cairo_draw_rect(cairo_t*, gint, gint, gint, gint, gint, gint) */

static void cff_gui_cairo_draw_line(cairo_t *cr, gint color,
					gint x1, gint y1, gint x2, gint y2)
{
	gdk_cairo_set_source_rgba(cr, &gCairo.RGBA[color]);

	cairo_move_to(cr, (gdouble) x1, (gdouble) y1);
	cairo_line_to(cr, (gdouble) x2, (gdouble) y2);
	cairo_stroke(cr);

	cairo_close_path(cr);

} /* static void cff_gui_cairo_draw_line(cairo_t*, gint, gint, gint, gint, gint) */

#endif /* GTK_MAJOR_VERSION > 2 */


/*
	-------
	 VIDEO
	-------
*/

static void cff_gui_video_button_set_sensitive(gint sensitive)
{
	/*
		Called from
		  'cff_gui_init'
		  'cff_gui_video_open'
		  'cff_gui_file_close'
	*/

	gint		i;
	GtkWidget	**button;

	button = gui_button_video_frame;
	i = 0;

	do
	{
		gtk_widget_set_sensitive(*(button++), sensitive);
	}
	while (++i < kButtonVideoFrameTotal);

	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutAdd], sensitive);
	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutRemove], FALSE);

} /* static void cff_gui_video_button_set_sensitive(gint) */

static void cff_gui_video_button_cut_set_sensitive(gint sensitive)
{
	/* either add or remove a cut point */

	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutAdd], sensitive);

	sensitive = (sensitive == TRUE) ? FALSE : TRUE;
	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutRemove], sensitive);

} /* static void cff_gui_video_button_cut_set_sensitive(gint) */

static void cff_gui_widget_queue_draw_area(GtkWidget *widget)
{
	gint width, height;

	width	= GuiWidgetGetAllocatedWidth(widget);
	height	= GuiWidgetGetAllocatedHeight(widget);

	gtk_widget_queue_draw_area(widget,
				0, 0, /* x + y */
				width, height);

} /* static void cff_gui_widget_queue_draw_area(GtkWidget*) */

static void cff_gui_frame_info(const cff_video_t *video)
{
	gchar buf[128];

	snprintf(buf, sizeof(buf), "[%ux%u] [%u:%u] %s-Frame : %u / %u",
			video->vframe.frame_width,
			video->vframe.frame_height,
			video->vframe.frame_aspect[0],
			video->vframe.frame_aspect[1],
			video->vframe.pict_type,
			video->vframe.frame_current,
			video->vframe.frame_total);

	gtk_label_set_text(gLabelFrame, buf);

	snprintf(buf, sizeof(buf), "Time: %02u:%02u:%02u.%03u / %02u:%02u:%02u.%03u",
			video->vframe.time_current[0],
			video->vframe.time_current[1],
			video->vframe.time_current[2],
			video->vframe.time_current[3],
			video->vframe.time_total[0],
			video->vframe.time_total[1],
			video->vframe.time_total[2],
			video->vframe.time_total[3]);

	gtk_label_set_text(gLabelTime, buf);

} /* static void cff_gui_frame_info(const cff_video_t*) */

#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_frame(GtkWidget *widget, GdkEvent *e, gpointer data)
#else
static gint cff_gui_draw_frame(GtkWidget *widget, cairo_t *cr, gpointer data)
#endif
{
	gint		color, width, height;
	cff_video_t	*video;

#if GTK_MAJOR_VERSION > 2

	cairo_surface_t	**surface;

	surface = &(gCairo.surfaceFrame);

	if (*surface == NULL)
	{
		if (cff_gui_cairo_surface_create(widget, surface) == FALSE)
		{
			return TRUE;
		}
	}

	cairo_set_source_surface(cr, *surface, 0, 0);
#endif
	video  = data;
	width  = GuiWidgetGetAllocatedWidth(widget);
	height = GuiWidgetGetAllocatedHeight(widget);
	color  = kColorGrey;

	if (video != NULL && video->vframe.rgb24.data != NULL)
	{
		color = kColorDarkGrey;
	}


#if GTK_MAJOR_VERSION == 2

	gdk_draw_rectangle(widget->window,
			gContext[color],
			TRUE,
			0, 0, /* x + y */
			width, height);
#else
	cff_gui_cairo_draw_rect(cr,
			color,
			TRUE,
			0, 0, /* x + y */
			width, height);
#endif

	if (color == kColorDarkGrey)
	{
		gint x, y;

		x = (width - video->vframe.rgb24.width) / 2;
		y = (height - video->vframe.rgb24.height) / 2;

#if GTK_MAJOR_VERSION == 2
		gdk_draw_rgb_image(widget->window,
			gContext[kColorBlack],
			x, y,
			video->vframe.rgb24.width,
			video->vframe.rgb24.height,
			GDK_RGB_DITHER_NONE,
			video->vframe.rgb24.data,
			video->vframe.rgb24.width * 3);
#else
			if (gCairo.surfaceRGB24 != NULL)
			{
				cairo_set_source_surface(cr, gCairo.surfaceRGB24, x, y);
				cairo_paint(cr);
			}
#endif

	} /* if (color == kColorDarkGrey) */	

	return TRUE;

} /* static gint cff_gui_draw_frame(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_mainwindow_cursor_new(GtkWidget *widget, GdkScreen *prev_screen, gpointer data)
{
	GdkDisplay	*display;
	cff_cursor_t	*cursor;

	cursor	= data;
	display = gtk_widget_get_display(widget);

	if (cursor->busy != NULL)
	{
		if (gdk_cursor_get_display(cursor->busy) == display)
		{
			return;
		}

		GuiCursorUnref(cursor->busy);
	}

	cursor->busy   = gdk_cursor_new_for_display(display, GDK_WATCH);
	cursor->window = gtk_widget_get_window(widget);

} /* static void cff_gui_mainwindow_cursor_new(GtkWidget*, GdkScreen*, gpointer) */


#if GTK_MAJOR_VERSION > 2

static gint cff_gui_drawing_area_configure_cb(GtkWidget *widget, GdkRectangle *rect, gpointer data)
{
	cff_gui_cairo_surface_create(widget, data);

	return TRUE;

} /* static gint cff_gui_drawing_area_configure_cb(GtkWidget*, GdkRectangle*, gpointer) */

#endif /* GTK_MAJOR_VERSION > 2 */


#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_frame_configure_cb(GtkWidget *widget, GdkEvent *e, gpointer data)
#else
static gint cff_gui_draw_frame_configure_cb(GtkWidget *widget, GdkRectangle *rect, gpointer data)
#endif
{
	/* 'gDrawingAreaFrame' changes size */

	cff_video_t *video = data;

	if (video->vframe.rgb24.data != NULL && video->vframe.avframe != NULL)
	{
		gint width, height;

		width  = GuiWidgetGetAllocatedWidth(widget);
		height = GuiWidgetGetAllocatedHeight(widget);

		cff_ffmpeg_frame2rgb24(video, width, height);

#if GTK_MAJOR_VERSION > 2
		cff_gui_cairo_surface_image_create(widget, &gCairo.surfaceRGB24, video);
		cff_gui_cairo_surface_image_fill(gCairo.surfaceRGB24, video);
#endif
	}

	return TRUE;

} /* static gint cff_gui_draw_frame_configure_cb(GtkWidget*, GdkEvent*, gpointer) */

static gpointer cff_gui_video_frame_thread(gpointer data)
{
	gint retv;

	struct
	{
		cff_video_t		*video;
		gdouble			*next_d;
		gint			next_i;
		gint			width;
		gint			height;
		enum cff_FrameFlags	frame_flags;

	} *s = data;

	if (s->next_d == NULL)
	{
		retv = cff_ffmpeg_get_frame_next(s->video, s->next_i,
					s->width, s->height, s->frame_flags);
	}

	else if (s->next_d == (gdouble *) -1)
	{
		retv = cff_ffmpeg_get_frame_cutpoint(s->video, s->next_i,
					s->width, s->height, s->frame_flags);
	}

	else
	{
		retv = cff_ffmpeg_get_frame_slider(s->video, *(s->next_d),
					s->width, s->height, s->frame_flags);
	}

	return GINT_TO_POINTER(retv);

} /* static gpointer cff_gui_video_frame_thread(gpointer) */

static gint cff_gui_video_frame(GtkWidget *widget, cff_video_t *video, gdouble *next_d, gint next_i)
{
	gint		cursor, retv;
	gpointer	result;

	struct
	{
		cff_video_t		*video;
		gdouble			*next_d;
		gint			next_i;
		gint			width;
		gint			height;
		enum cff_FrameFlags	frame_flags;
	} s;

	if (gtk_main_level() > 1)
	{
		return TRUE;
	}

	if (video->vframe.ic == NULL)
	{
		return TRUE;
	}

	s.video  = video;
	s.next_d = next_d;
	s.next_i = next_i;
	s.width  = GuiWidgetGetAllocatedWidth(widget);
	s.height = GuiWidgetGetAllocatedHeight(widget);

	s.frame_flags = 0;

	if (cff_gprop_get_bool(kPropBoolOptKeyFrame, &retv) && retv)
	{
		s.frame_flags |= kKeyFrameFlag;
	}

	if (cff_gprop_get_bool(kPropBoolOptIFrame, &retv) && retv)
	{
		s.frame_flags |= kIFrameFlag;
	}

	if (cff_gprop_get_bool(kPropBoolOptPFrame, &retv) && retv)
	{
		s.frame_flags |= kPFrameFlag;
	}

	if (cff_gprop_get_bool(kPropBoolOptCursor, &cursor) == FALSE)
	{
		cursor = TRUE;
	}

	if (cursor == TRUE && gCursor.window != NULL)
	{
		gdk_window_set_cursor(gCursor.window, gCursor.busy);
	}

	if (gLogWindow.window == NULL ||
		cff_gui_thread_new((GThreadFunc)
			cff_gui_video_frame_thread, &s, &result) == FALSE)
	{
		gtk_main_iteration();
		result = cff_gui_video_frame_thread(&s);
	}

	retv = GPOINTER_TO_INT(result);

	if (cursor == TRUE && gCursor.window != NULL)
	{
		gdk_window_set_cursor(gCursor.window, NULL);
	}

	if (retv == -1)
	{
		cff_gui_beep();
	}

	else
	{
		gboolean cut_add;

#if GTK_MAJOR_VERSION > 2

		if (gCairo.surfaceRGB24 == NULL ||
			gCairo.rgb24_width != video->vframe.rgb24.width ||
				gCairo.rgb24_height != video->vframe.rgb24.height)
		{
			if (cff_gui_cairo_surface_image_create(widget,
					&gCairo.surfaceRGB24, video) == FALSE)
			{
				return TRUE;
			}
		}

		cff_gui_cairo_surface_image_fill(gCairo.surfaceRGB24, video);
#endif
		cff_gui_widget_queue_draw_area(widget);
		cff_gui_widget_queue_draw_area(gDrawingAreaSlider);
		cff_gui_frame_info(video);

		cut_add = TRUE;

		if (cff_ffmpeg_video_cut_find_current(video) >= 0)
		{
			cut_add = FALSE;
		}

		cff_gui_video_button_cut_set_sensitive(cut_add);
	}

	return retv;

} /* static gint cff_gui_video_frame(GtkWidget*, cff_video_t*, gdouble*, gint) */

static gint cff_gui_video_frame_button(GtkWidget *button, GdkEvent *e, gpointer data)
{
	gint		next;
	GdkEventButton	*bevent;

	bevent = (GdkEventButton *) e;

	switch (GPOINTER_TO_INT(data))
	{
		case kButtonVideoFramePrev4:
			next = kVideoFramePrevCut;
			break;
		case kButtonVideoFramePrev3:
			next = kVideoFramePrev05Sec;
			break;
		case kButtonVideoFramePrev2:
			next = kVideoFramePrevHalfSec;
			break;
		case kButtonVideoFramePrev1:
			next = kVideoFramePrevSingle;
			break;
		case kButtonVideoFrameNext1:
			next = kVideoFrameNextSingle;
			break;
		case kButtonVideoFrameNext2:
			next = kVideoFrameNextHalfSec;
			break;
		case kButtonVideoFrameNext3:
			next = kVideoFrameNext05Sec;
			break;
		case kButtonVideoFrameNext4:
			next = kVideoFrameNextCut;
			break;
		default:
			return TRUE;

	} /* switch (GPOINTER_TO_INT(data)) */

	if ((bevent->state & ~(GDK_SHIFT_MASK | GDK_LOCK_MASK)) != 0)
	{
		return FALSE;
	}

	if ((bevent->state & GDK_SHIFT_MASK) != 0)
	{
		next += kVideoFrameShiftAdd;
	}

	cff_gui_video_frame(gDrawingAreaFrame, &gVideo, NULL, next);

	return TRUE;

} /* static gint cff_gui_video_frame_button(GtkWidget*, GdkEvent*, gpointer) */

static gint cff_gui_video_frame_key(GtkWidget *window, GdkEvent *e, gpointer data)
{
	gint		next;
	GdkEventKey	*kevent;

	kevent = (GdkEventKey *) e;

	switch (kevent->keyval)
	{
		case GDK_KEY_Home:
		case GDK_KEY_KP_Home:
		case GDK_KEY_KP_7:
			next = kVideoFramePrevCut;
			break;
		case GDK_KEY_Page_Down:
		case GDK_KEY_KP_Page_Down:
		case GDK_KEY_KP_3:
			next = kVideoFramePrev05Sec;
			break;
		case GDK_KEY_Down:
		case GDK_KEY_KP_Down:
		case GDK_KEY_KP_2:
			next = kVideoFramePrevHalfSec;
			break;
		case GDK_KEY_Left:
		case GDK_KEY_KP_Left:
		case GDK_KEY_KP_4:
			next = kVideoFramePrevSingle;
			break;
		case GDK_KEY_Right:
		case GDK_KEY_KP_Right:
		case GDK_KEY_KP_6:
			next = kVideoFrameNextSingle;
			break;
		case GDK_KEY_Up:
		case GDK_KEY_KP_Up:
		case GDK_KEY_KP_8:
			next = kVideoFrameNextHalfSec;
			break;
		case GDK_KEY_Page_Up:
		case GDK_KEY_KP_Page_Up:
		case GDK_KEY_KP_9:
			next = kVideoFrameNext05Sec;
			break;
		case GDK_KEY_End:
		case GDK_KEY_KP_End:
		case GDK_KEY_KP_1:
			next = kVideoFrameNextCut;
			break;
		default:
			return FALSE;

	} /* switch (GPOINTER_TO_INT(data)) */

	if ((kevent->state & ~(GDK_SHIFT_MASK | GDK_LOCK_MASK)) != 0)
	{
		return FALSE;
	}

	if ((kevent->state & GDK_SHIFT_MASK) != 0)
	{
		next += kVideoFrameShiftAdd;
	}

	cff_gui_video_frame(gDrawingAreaFrame, &gVideo, NULL, next);

	return TRUE;

} /* static gint cff_gui_video_frame_key(GtkWidget*, GdkEvent*, gpointer) */

#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_bar(GtkWidget *widget, GdkEvent *e, gpointer data)
#else
static gint cff_gui_draw_bar(GtkWidget *widget, cairo_t *cr, gpointer data)
#endif
{
	gint		x1, color, width, height;
	cff_video_t	*video;

#if GTK_MAJOR_VERSION > 2

	cairo_surface_t	**surface;

	surface = &(gCairo.surfaceBar);

	if (*surface == NULL)
	{
		if (cff_gui_cairo_surface_create(widget, surface) == FALSE)
		{
			return TRUE;
		}
	}

	cairo_set_source_surface(cr, *surface, 0, 0);
#endif
	video	= data;
	width	= GuiWidgetGetAllocatedWidth(widget);
	height	= GuiWidgetGetAllocatedHeight(widget);

	x1	= 0;
	color	= kColorGrey;

	if (video->vframe.frame_total != 0 && video->vframe.stream_duration > 0)
	{
		gint		i;
		cff_dts_t	*dts;

		color	= kColorGreen;
		dts	= video->cut.list;
		i	= video->cut.num;

		if (i > 0 && dts != NULL)
		{
			double ratio =
				(double) width / (double) video->vframe.stream_duration;;

			color = kColorRed;

			do
			{
				gint x2 = (double) (dts->frame - video->vframe.dts_start) * ratio;
				dts++;

				if (x2 > 0)
				{
#if GTK_MAJOR_VERSION == 2
					gdk_draw_rectangle(widget->window,
						gContext[color],
						TRUE,
						x1, 0, /* y */
						x2 - x1, height);
#else
					cff_gui_cairo_draw_rect(cr,
						color,
						TRUE,
						x1, 0, /* y */
						x2 - x1, height);
#endif
					x1 = x2;
				}

				if (color == kColorGreen)
				{
					color = kColorRed;
				}

				else
				{
					color = kColorGreen;
				}

			}
			while (--i > 0);

		} /* if (i > 0 && ptr != NULL) */

	} /* if (video->vframe.frame_total != 0 && video->vframe.duration > 0) */


#if GTK_MAJOR_VERSION == 2
	gdk_draw_rectangle(widget->window,
			gContext[color],
			TRUE,
			x1, 0, /* y */
			width - x1, height);

	gdk_draw_rectangle(widget->window,
			gContext[kColorBlack],
			FALSE,
			0, 0, /* x + y */
			width - 1, height - 1);
#else
	cff_gui_cairo_draw_rect(cr,
			color,
			TRUE,
			x1, 0, /* y */
			width - x1, height);

	cff_gui_cairo_draw_rect(cr,
			kColorBlack,
			FALSE,
			0, 0, /* x + y */
			width - 1, height - 1);

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */

	return TRUE;

} /* static gint cff_gui_draw_bar(GtkWidget*, GdkEvent*, gpointer) */

static gint cff_gui_draw_bar_button(GtkWidget *widget, GdkEvent *e, gpointer data)
{
	GdkEventButton	*bevent;
	cff_video_t	*video;

	video	= data;
	bevent	= (GdkEventButton *) e;

	if (video->vframe.frame_total != 0 &&
			bevent->type == GDK_BUTTON_PRESS &&
						bevent->button == 1)
	{
		gint	width;
		double	next;

		width	= GuiWidgetGetAllocatedWidth(widget);
		next	= bevent->x / (double) width;

		cff_gui_video_frame(gDrawingAreaFrame, &gVideo, &next, 0);

	} /* if (video->vframe.frame_total != 0 &&
			bevent->type == GDK_BUTTON_PRESS &&
						bevent->button == 1) */
	return TRUE;

} /* static gint cff_gui_draw_bar_button(GtkWidget*, GdkEvent*, gpointer) */


#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_slider(GtkWidget *widget, GdkEvent *e, gpointer data)
#else
static gint cff_gui_draw_slider(GtkWidget *widget, cairo_t *cr, gpointer data)
#endif
{
	gint		x, width, height;
	uint64_t	max;
	cff_video_t	*video;

	video	= data;
	width	= GuiWidgetGetAllocatedWidth(widget);
	height	= GuiWidgetGetAllocatedHeight(widget);

#if GTK_MAJOR_VERSION == 2

	gdk_draw_line(widget->window, gContext[kColorBlack],
				1, height / 2, width - 1, height / 2);
#else

	cairo_surface_t	**surface;

	surface = &(gCairo.surfaceSlider);

	if (*surface == NULL)
	{
		if (cff_gui_cairo_surface_create(widget, surface) == FALSE)
		{
			return TRUE;
		}
	}

	cairo_set_source_surface(cr, *surface, 0, 0);

	cff_gui_cairo_draw_line(cr, kColorBlack,
				1, height / 2, width - 1, height / 2);
#endif

	max = video->vframe.stream_duration;

	if (max == 0)
	{
		x = 1;
	}

	else
	{
		x = width * (double) (video->vframe.current.frame - video->vframe.dts_start) / (double) max;

		/* 3 pixels width */

		if (x <= 0)
		{
			x = 1;
		}


		else if (x > 1)
		{
			if (x >= width)
			{
				x = width - 2;
			}

			else
			{
				x--;
			}
		}

	} /* if (max == 0) .. else */


#if GTK_MAJOR_VERSION == 2

	gdk_draw_rectangle(widget->window,
				gContext[kColorBlack],
				TRUE,
				x - 1, 0, /* x + y */
				3, height);
#else
	cff_gui_cairo_draw_rect(cr,
				kColorBlack,
				TRUE,
				x - 1, 0, /* x + y */
				3, height);
#endif

	return TRUE;

} /* static gint cff_gui_draw_slider(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_open_result(cff_video_t *video, const gchar *str, guint mbar_item)
{
	gint sensitive;

	if (str == NULL)
	{

#if GTK_MAJOR_VERSION > 2

		if (gCairo.surfaceRGB24 != NULL)
		{
			/* create a new surface */
			cairo_surface_destroy(gCairo.surfaceRGB24);
			gCairo.surfaceRGB24 = NULL;
		}
#endif
		cff_gui_video_frame(gDrawingAreaFrame,
					video, NULL, kVideoFrameStart);
		sensitive = TRUE;
	}

	else
	{
		cff_gui_widget_queue_draw_area(gDrawingAreaFrame);
		sensitive = FALSE;
	}

	cff_gui_video_button_set_sensitive(sensitive);

	cff_gui_frame_info(video);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

	if (str != NULL)
	{
		cff_gui_beep();
		guitl_dialog_error(str, &cff_gui_menu_bar[mbar_item]);
	}

} /* static void cff_gui_open_result(cff_video_t*, const gchar*, guint) */

static int cff_gui_open_callback(gpointer data)
{
	const gchar		**str;
	cff_progressbar_t	user_data;

	user_data.func  = cff_ffmpeg_file_open_cb;
	user_data.video = &gVideo;
	user_data.title = gettext_nop("Scanning ...");
	user_data.mbar  = &cff_gui_menu_bar[kMenuBarFile];

	if (gLogWindow.nThreads == 0)
	{
		user_data.mutex = NULL;
		cff_gui_dialog_progress_bar(&user_data);
	}

	else
	{
		GuiMutex mutex;

		user_data.mutex = &mutex;
		GuiMutexInit(&mutex);

		user_data.is_locked = FALSE;

		g_main_context_invoke_full(NULL, G_PRIORITY_HIGH,
			(GSourceFunc) cff_gui_dialog_progress_bar,
							&user_data, NULL);

		while (user_data.is_locked == FALSE)
		{
			g_usleep(250);
		}

		GuiMutexLock(&mutex);
		GuiMutexUnlock(&mutex);
		GuiMutexFree(&mutex);

	} /* if (gLogWindow.nThreads == 0) .. else */

	str = (const gchar **) data;

	if (str != NULL)
	{
		*str = user_data.error;
	}

	return user_data.retv;

} /* static int cff_gui_open_callback(gpointer) */

static gint cff_gui_program_open(gpointer data)
{
	const gchar	*error, *str;
	cff_video_t	*video;

	video = &gVideo;

	gIsBusy = TRUE;

	str = cff_ffmpeg_program_open(video, GPOINTER_TO_INT(data),
					cff_gui_open_callback, &error);

	cff_gui_open_result(video, (str == NULL) ? error : str, kMenuBarEdit);

	gIsBusy = FALSE;

	return Gui_SOURCE_REMOVE;

} /* static gint cff_gui_program_open(gpointer) */

static const gchar *cff_gui_video_open_cb(gpointer data)
{
	struct
	{
		cff_video_t	*video;
		gchar		*fname;
		const gchar	*error;

	} *s = data;

	return cff_ffmpeg_file_open(s->video, s->fname,
				cff_gui_open_callback, &(s->error));

} /* static const gchar *cff_gui_video_open_cb(gpointer) */

static gint cff_gui_video_open(gchar *fname)
{
	const gchar *str;

	struct
	{
		cff_video_t	*video;
		gchar		*fname;
		const gchar	*error;

	} s;

	s.fname = fname;
	s.video = &gVideo;
	gIsBusy = TRUE;

	if (gLogWindow.window == NULL ||
		cff_gui_thread_new((GThreadFunc)
			cff_gui_video_open_cb, &s, (gpointer) &str) == FALSE)
	{
		str = cff_ffmpeg_file_open(s.video, s.fname,
				cff_gui_open_callback, &(s.error));
	}

	cff_gui_open_result(s.video, (str == NULL) ? s.error : str, kMenuBarFile);

	guitl_window_set_title_filename(gWindata.mainwin,
				((s.video)->inFormatCtx == NULL) ? NULL : s.fname);
	g_free(s.fname);

	gIsBusy = FALSE;

	return Gui_SOURCE_REMOVE;

} /* static gint cff_gui_video_open(gchar*) */


/*
	-------------
	 Menu 'File'
	-------------
*/

static void cff_gui_file_open(GtkWidget *w, GtkWidget **mbar)
{
	gchar		*fname;
	cff_video_t	*video;

	video = &gVideo;

	if (video->dirty)
	{
		if (guitl_dialog_warning_2btn(gettext_nop("Discard changes?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
		{
			return;
		}
	}

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Open Video File"),
					NULL, mbar, kPropStrFileOpenPath);
	if (fname == NULL)
	{
		return;
	}

	cff_gui_file_close(NULL, NULL);

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_video_open, fname, NULL);

} /* static void cff_gui_file_open(GtkWidget*, GtkWidget**) */

static void cff_gui_file_save(GtkWidget *w, GtkWidget **mbar)
{
	gint			i, n_avail, n_video, n_audio;
	gchar			*str, **ptr;
	GtkWidget		*button, *notebook, *bbox, *hbox, *vbox, *frame;
	cff_fmtlist_t		*format_info;
	cff_save_video_t	*vsave;
	cff_video_t		*video;
	gchar			*buf[512];

	video = &gVideo;

	if (video->fname == NULL)
	{
		return;
	}

	vsave = g_malloc(sizeof(cff_save_video_t));

	vsave->common.mbar  = mbar;
	vsave->common.video = video;

	vbox = cff_gui_save_build_common(&(vsave->common),
		gettext_nop("Save Video File"), kPropStrFileSavePath);

	GuiSignalConnect(vsave->common.window, "delete-event", cff_gui_save_delevnt, vsave);

	guitl_window_setup(vsave->common.window, kPropIntFileSaveWinWidth);

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 4);

	/* List */

	cff_gui_build_list(gettext_nop("Available"), &(vsave->avail.list), hbox);

	/* Button */

	bbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);

	vsave->up =
		guitl_vbox_add_button_gettext(bbox, gettext_nop("Up"),
				(GCallback) cff_gui_save_list_updown, vsave);
	vsave->avail.all =
		guitl_vbox_add_button(bbox, ">>",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->avail.some =
		guitl_vbox_add_button(bbox, ">",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->out.some =
		guitl_vbox_add_button(bbox, "<",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->out.all =
		guitl_vbox_add_button(bbox, "<<",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->down =
		guitl_vbox_add_button_gettext(bbox, gettext_nop("Down"),
				(GCallback) cff_gui_save_list_updown, vsave);

	frame = guitl_frame_new_add_child_gettext(NULL, bbox);
	bbox = GuiVBoxNew(FALSE, 1);
	GuiVBoxPackStart(bbox, frame, FALSE, FALSE, 1);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);

	/* List */

	cff_gui_build_list(gettext_nop("Output"), &(vsave->out.list), hbox);

	GuiSignalConnect(vsave->avail.list.selection, "changed",
					cff_gui_save_list_changed, vsave);
	GuiSignalConnect(vsave->out.list.selection, "changed",
					cff_gui_save_list_changed, vsave);

	/* Check button */

	vsave->shortest = guitl_vbox_add_check_button_gettext(vbox,
					gettext_nop("Shortest duration"));

	/* Extra */

	notebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
	gtk_container_set_border_width(GTK_CONTAINER(notebook), 4);

	frame = guitl_frame_new_add_child_gettext(gettext_nop("Extra"), notebook);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 2);

	/* General-Tab */

	bbox = cff_gui_notebook_append_page(notebook, gettext_nop("General"));

	/* Check button */

	vsave->ts_keep_start = guitl_vbox_add_check_button_gettext(bbox,
					gettext_nop("Keep start timestamps"));
	/* Option menu */

	vsave->common.format.u.list =
			cff_ffmpeg_get_format_names(vsave->common.video);

	ptr = buf;

	if ((format_info = vsave->common.format.u.list) != NULL)
	{
		while (format_info->name != NULL)
		{
			if (ptr - buf < (sizeof(buf) / sizeof(gchar*)) - 1)
			{
				*(ptr++) = g_strconcat(format_info->name, " ",
							format_info->codecs, NULL);
			}

			if (format_info->codecs != NULL)
			{
				/* malloc'ed */
				free(format_info->codecs);
				format_info->codecs = NULL;
			}

			format_info++;
		}

	} /* if ((format_info = vsave->common.format.u.list) != NULL) */

	*ptr = NULL;

	vsave->common.format.option =
			cff_gui_build_option_menu(bbox,
				gettext_nop("Format"),
				gettext_nop("Auto"),
				(const gchar **) buf);

	if (vsave->common.format.u.list != NULL)
	{
		GuiSignalConnect(vsave->common.format.option, "changed",
			cff_gui_save_optionmenu_callback, &(vsave->common));
	}

	/* Free strings in array */

	ptr = buf;

	while ((str = *(ptr++)) != NULL)
	{
		g_free(str);
	}

	/* Media-Tabs */

	cff_gui_build_media_option(notebook, vsave, kCFFMediaVideo);
	cff_gui_build_media_option(notebook, vsave, kCFFMediaAudio);
	cff_gui_build_media_option(notebook, vsave, kCFFMediaOther);

	/* Button */

	hbox = guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	guitl_button_box_add_button(hbox, gettext_nop("Cancel"),
				(GCallback) cff_gui_save_destroy, vsave);

	button = guitl_button_box_add_button(hbox, gettext_nop("OK"),
				(GCallback) cff_gui_save_start, vsave);
	vsave->common.okay.widget = button;
	vsave->common.okay.bits   = 0;
	gtk_widget_grab_default(button);


	n_avail = 0;
	n_video = 0;
	n_audio = 0;

	i = 0;

	do
	{
		gint		stream_idx;
		gboolean	select;
		gchar		*str, buf[256];
		cff_ListStore_t *list;
		GtkTreeIter	iter;

		if (cff_ffmpeg_get_stream_info(video, i,
					&stream_idx, buf, sizeof(buf)) < 0)
		{
			continue;
		}

		select = FALSE;
		str = strchr(buf, ' ');

		if (str != NULL)
		{
			if (n_video == 0 &&
				g_ascii_strncasecmp(&str[1], "video", 5) == 0)
			{
				if (n_audio == 0)
				{
					select = TRUE;
				}

				n_video++;
				list = &(vsave->out.list);
			}

			else if (n_audio == 0 &&
				g_ascii_strncasecmp(&str[1], "audio", 5) == 0)
			{
				if (n_video == 0)
				{
					select = TRUE;
				}

				n_audio++;
				list = &(vsave->out.list);
			}

			else
			{
				str = NULL;
			}

		} /* if (str != NULL) */

		if (str == NULL)
		{
			if (n_avail == 0)
			{
				select = TRUE;
			}

			n_avail++;
			list = &(vsave->avail.list);
		}

		gtk_list_store_append(list->store, &iter);
		str = tl_replace_non_ascii_chars_in_str(buf);
		gtk_list_store_set(list->store, &iter,
						0, str, 1, stream_idx, -1);
		if (select == TRUE)
		{
			gtk_tree_selection_select_iter(list->selection, &iter);
		}
	}
	while(++ i < video->nb_streams);

	/* Set buttons (up, down, left, left-all and okay) insensitive */
	cff_gui_save_list_changed(vsave->avail.list.selection, vsave);
	cff_gui_save_list_changed(vsave->out.list.selection, vsave);

	gtk_widget_show_all(vsave->common.window);

} /* static void cff_gui_file_save(GtkWidget*, GtkWidget**) */

static void cff_gui_file_image(GtkWidget *w, GtkWidget **mbar)
{
	gint			fmt_num;
	gchar			**long_names;
	GtkWidget		*button, *hbox, *vbox, *vbox1, *bbox, *fbox, *frame;
	cff_save_image_t	*vimage;
	cff_video_t		*video;

	video = &gVideo;

	if (video->fname == NULL)
	{
		return;
	}

	vimage = g_malloc(sizeof(cff_save_image_t));

	vimage->common.mbar  = mbar;
	vimage->common.video = video;

	vbox = cff_gui_save_build_common(&(vimage->common),
		gettext_nop("Save Image"), kPropStrFileImagePath);

	GuiSignalConnect(vimage->common.window,
		"delete-event", cff_gui_image_delevnt, vimage);

	guitl_window_setup(vimage->common.window, kPropIntFileImageWinWidth);

	vbox1 = GuiVBoxNew(FALSE, 4);
	GuiVBoxPackStart(vbox, vbox1, FALSE, FALSE, 4);

	fbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(fbox), 4);
	GuiVBoxPackStart(vbox, fbox, FALSE, FALSE, 4);

	/* Spin buttons */

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(fbox, frame, FALSE, FALSE, 4);

	bbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 4);

	vimage->width.prev_value = video->vframe.frame_width;
	vimage->width.widget = cff_gui_image_spinbutton(bbox,
			video->vframe.frame_width, gettext_nop("Width"));

	vimage->height.prev_value = video->vframe.frame_height;
	vimage->height.widget = cff_gui_image_spinbutton(bbox,
			video->vframe.frame_height, gettext_nop("Height"));

	/* Check button */

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(fbox, frame, FALSE, FALSE, 4);

	bbox = GuiHBoxNew(FALSE, 4);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 4);

	hbox = bbox;
	bbox = GuiVBoxNew(FALSE, 4);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);

	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);

	vimage->keep_dimension =
		guitl_check_button_new_gettext(gettext_nop("Keep dimension"));
	GuiVBoxPackStart(bbox, vimage->keep_dimension, FALSE, FALSE, 4);
	GuiSignalConnect(vimage->keep_dimension,
		"toggled", cff_gui_image_keep_dimension_toggled, vimage);

	vimage->keep_ratio =
		guitl_check_button_new_gettext(gettext_nop("Keep aspect ratio"));
	GuiVBoxPackStart(bbox, vimage->keep_ratio, FALSE, FALSE, 4);

	vimage->gray = guitl_check_button_new_gettext(gettext_nop("Gray"));
	GuiVBoxPackStart(bbox, vimage->gray, FALSE, FALSE, 4);
				
	/* Option menu */

	fmt_num = cff_ffmpeg_get_image_format_names(&long_names,
					&(vimage->common.format.u.names));

	vimage->common.format.option =
			cff_gui_build_option_menu(vbox,
				gettext_nop("Format"),
				gettext_nop("Auto"),
				(const gchar **) long_names);

	if (fmt_num > 0)
	{
		if (long_names != NULL)
		{
			/* malloc'ed */
			free(long_names);
		}

		GuiSignalConnect(vimage->common.format.option, "changed",
			cff_gui_save_optionmenu_callback, &(vimage->common));
	}

	/* Padding */

	vbox1 = GuiVBoxNew(FALSE, 4);
	GuiVBoxPackStart(vbox, vbox1, TRUE, FALSE, 4);

	/* Button */

	hbox = guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);

	guitl_button_box_add_button(hbox, gettext_nop("Cancel"),
				(GCallback) cff_gui_image_destroy, vimage);

	button = guitl_button_box_add_button(hbox, gettext_nop("OK"),
				(GCallback) cff_gui_save_image, vimage);
	vimage->common.okay.widget = button;
	vimage->common.okay.bits   = 0;
	gtk_widget_grab_default(button);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(vimage->keep_ratio), TRUE);

	gtk_widget_show_all(vimage->common.window);

} /* static void cff_gui_file_image(GtkWidget*, GtkWidget**) */

static void cff_gui_file_close(GtkWidget *w, gpointer data)
{
	cff_ffmpeg_file_close(&gVideo);

	cff_gui_frame_info(&gVideo);
	cff_gui_video_button_set_sensitive(FALSE);

	cff_gui_widget_queue_draw_area(gDrawingAreaFrame);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	cff_gui_widget_queue_draw_area(gDrawingAreaSlider);

	guitl_window_set_title_filename(gWindata.mainwin, NULL);

} /* static void cff_gui_file_close(GtkWidget*, gpointer) */

static void cff_gui_file_info(GtkWidget *w, GtkWidget **mbar)
{
	char		*str;
	GtkWidget	*ibox;

	ibox = GuiVBoxNew(FALSE, 2);
	str  = cff_ffmpeg_file_info(&gVideo);

	str = tl_replace_non_ascii_chars_in_str(str);
	cff_gui_ffmpeg_info_label(NULL, str, ibox);

	if (str)
	{
		/* malloc'ed */
		free(str);
	}

	cff_gui_scrolled_info_window(gettext_nop("Video File Info"),
				ibox, mbar, kPropIntFileInfoWinWidth);

} /* static void cff_gui_file_info(GtkWidget*, GtkWidget**) */

static void cff_gui_file_save_rc(GtkWidget *w, gpointer data)
{
	cff_main_save_rc();

} /* static void cff_gui_file_save_rc(GtkWidget*, gpointer) */

static gboolean cff_gui_file_quit_delevnt(GtkWidget *w, GdkEvent *e, GtkWidget **mbar)
{
	cff_video_t	*video;
	cff_logwindow_t	*log;

	video = &gVideo;

	if (gIsBusy == TRUE || gtk_main_level() > 1)
	{
		return TRUE;
	}

	if (video->dirty &&
		guitl_dialog_warning_2btn(gettext_nop("Discard changes?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
	{
		return TRUE;
	}

	/* 'cff_main_save_rc' called from 'cff_main_atexit' */

	cff_ffmpeg_file_close(video);

	log = &gLogWindow;

	if (log->window != NULL)
	{
		gtk_widget_destroy(log->window);
	}

	return FALSE;

} /* static gboolean cff_gui_file_quit_delevnt(GtkWidget*, GdkEvent*, GtkWidget**) */

static void cff_gui_file_quit(GtkWidget *w, GtkWidget **mbar)
{
	if (cff_gui_file_quit_delevnt(w, NULL, mbar) == FALSE)
	{
		gtk_widget_destroy(gWindata.mainwin);
	}

} /* static void cff_gui_file_quit(GtkWidget*, GtkWidget**) */


/*
	-------------
	 Menu 'Edit'
	-------------
*/

static void cff_gui_edit_undo_redo(cff_video_t *video, int action)
{
	int retv;

	if (action == 0)
	{
		retv = cff_ffmpeg_video_undo(video);
	}

	else
	{
		retv = cff_ffmpeg_video_redo(video);
	}

	if (retv < 0)
	{
		cff_gui_beep();
	}

	else
	{
		gboolean cut_add = TRUE;

		if (cff_ffmpeg_video_cut_find_current(video) >= 0)
		{
			cut_add = FALSE;
		}

		cff_gui_video_button_cut_set_sensitive(cut_add);
		cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_undo_redo(cff_video_t*, int) */

static void cff_gui_edit_undo(GtkWidget *w, cff_video_t *video)
{
	cff_gui_edit_undo_redo(video, 0);

} /* static void cff_gui_edit_undo(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_redo(GtkWidget *w, cff_video_t *video)
{
	cff_gui_edit_undo_redo(video, 1);

} /* static void cff_gui_edit_redo(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_load(GtkWidget *w, GtkWidget **mbar)
{
	gint		n;
	gchar		*fname;
	cff_video_t	*video;

	video = &gVideo;

	if (video->dirty &&
		guitl_dialog_warning_2btn(gettext_nop("Add list to existing cutpoints?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
	{
		return;
	}

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Load Cut List"),
					NULL, mbar, kPropStrEditPath);
	if (fname == NULL)
	{
		return;
	}

	n = cff_ffmpeg_video_cut_list_load(&gVideo, fname);

	g_free(fname);

	if (n < 0)
	{
		cff_gui_beep();
		guitl_dialog_error(gettext_nop("Cannot load cut list!"), NULL);
	}

	else
	{
		int i;

		cff_gui_widget_queue_draw_area(gDrawingAreaBar);

		i = cff_ffmpeg_video_cut_find_current(video);

		cff_gui_video_button_cut_set_sensitive((i < 0) ? TRUE : FALSE);

		if (n > 0)
		{
			gchar *str = g_strdup_printf("%d %s",
				n, (n == 1) ?
				gettext("item was discarded!") :
					gettext("items were discarded!"));
			guitl_dialog_warning(str, mbar);
			g_free(str);
		}
	}

} /* static void cff_gui_edit_load(GtkWidget*, GtkWidget**) */

static void cff_gui_edit_save(GtkWidget *w, GtkWidget **mbar)
{
	gint		retv;
	gchar		*fname, *dot, *str;
	cff_video_t	*video;

	video = &gVideo;

	if (video->fname == NULL)
	{
		return;
	}

	dot = strrchr(video->fname, '.');

	if (dot != NULL)
	{
		*dot = '\0';
	}

	str = g_strconcat(video->fname, ".cff", NULL);

	if (dot != NULL)
	{
		*dot = '.';
	}

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Save Cut List"),
					str, mbar, kPropStrEditPath);
	g_free(str);

	if (fname == NULL)
	{
		return;
	}

#ifndef HAVE_GTKOVERWRITECONFIRMATION

	if (cff_gui_overwrite(fname, mbar) == FALSE)
	{
		g_free(fname);
		return;
	}
#endif

	/* save cut list ... */

	retv = cff_ffmpeg_video_cut_list_save(&gVideo, fname);

	g_free(fname);

	if (retv < 0)
	{
		cff_gui_beep();
		guitl_dialog_error(gettext_nop("Cannot save cut list!"), NULL);
	}

} /* static void cff_gui_edit_save(GtkWidget*, GtkWidget**) */

static void cff_gui_edit_clear(GtkWidget *w, GtkWidget **mbar)
{
	cff_video_t *video = &gVideo;

	if (video->cut.num == 0 ||
		guitl_dialog_warning_2btn(gettext_nop("Clear cut list?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
	{
		return;
	}

	cff_ffmpeg_video_cut_list_clear(video);

	cff_gui_video_button_cut_set_sensitive(TRUE);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

} /* static void cff_gui_edit_clear(GtkWidget*, GtkWidget**) */

static void cff_gui_edit_add(GtkWidget *w, cff_video_t *video)
{
	if (cff_ffmpeg_video_cut_add(video) < 0)
	{
		cff_gui_beep();
	}

	else
	{
		cff_gui_video_button_cut_set_sensitive(FALSE);
		cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_add(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_remove(GtkWidget *w, cff_video_t *video)
{
	if (cff_ffmpeg_video_cut_remove(video) < 0)
	{
		cff_gui_beep();
	}

	else
	{
		cff_gui_video_button_cut_set_sensitive(TRUE);
		cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_remove(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_frame_display(GtkWidget *w, gpointer data)
{
	cff_gui_video_frame(gDrawingAreaFrame, &gVideo,
				(gdouble *) -1, GPOINTER_TO_INT(data));

} /* static void cff_gui_edit_frame_display(GtkWidget*, gpointer) */

static void cff_gui_edit_program_open(GtkWidget *w, gpointer data)
{
	cff_video_t *video = &gVideo;

	if (video->dirty &&
		guitl_dialog_warning_2btn(gettext_nop("Clear cut list?"),
			gettext_nop("Cancel"),
			gettext_nop("Yes"),
			&cff_gui_menu_bar[kMenuBarEdit]) == GTK_RESPONSE_CANCEL)
	{
		return;
	}

	cff_ffmpeg_video_cut_list_clear(video);

	cff_gui_video_button_set_sensitive(FALSE);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_program_open, data, NULL);

} /* static void cff_gui_edit_program_open(GtkWidget*, gpointer) */


/*
	----------------
	 Menu 'Options'
	----------------
*/

static void cff_gui_opt_toggle_bool(GtkWidget *mi, gpointer data)
{
	static gboolean clicked = FALSE;

	int	active, prop;
	guint	n;

	if (clicked)
	{
		/* Already handled */
		clicked = FALSE;
		return;
	}

	active = GuiCheckMenuItemGetActive(GTK_CHECK_MENU_ITEM(mi));
	prop = GPOINTER_TO_INT(data);

	cff_gprop_set_bool(prop, active);

	n = prop - kPropBoolOptKeyFrame;

	if (n <= kPropBoolOptPFrame)
	{
		if (mi == cff_gui_checkmenu_frame_widget[n])
		{
			/* Menu clicked */
			mi = cff_gui_popup_checkmenu_frame_widget[n];
		}

		else
		{
			/* Popup-Menu clicked */
			mi = cff_gui_checkmenu_frame_widget[n];
		}

		/* Following function emits the activate-signal */
		clicked = TRUE;

		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), active);
	}

} /* static void cff_gui_opt_toggle_bool(GtkWidget*, gpointer) */

static void cff_gui_opt_logwindow_menu_selected(GtkWidget *mi, gpointer data)
{
	gint		logindex;
	const gchar	*logname;

	logindex = GPOINTER_TO_INT(data);
	logname  = gtk_menu_item_get_label(GTK_MENU_ITEM(mi));

	if (logindex == 0)
	{
		cff_gui_logwindow_free(NULL, NULL, &gLogWindow);
		cff_gui_logwindow_set_title(gLogWindow.window, logname);
	}

	else
	{
		if (cff_gui_logwindow_open(logname) == TRUE)
		{
			cff_gprop_set_str(kPropStrOptLoglevel, logname);
		}
	}

} /* static void cff_gui_opt_logwindow_menu_selected(GtkWidget*, gpointer) */


/*
	-------------
	 Menu 'Help'
	-------------
*/

static void cff_gui_help_ffmpeg(GtkWidget *w, GtkWidget **mbar)
{
	int		i;
	char		*str;
	const char	*title;
	GtkWidget	*vbox;

	vbox	= GuiVBoxNew(FALSE, 2);
	i	= 0;

	do
	{
		switch (i)
		{
			case 0:
				str	= cff_ffmpeg_get_libs_info();
				title	= gettext_nop("Libs");
				break;
			case 1:
				str	= cff_ffmpeg_get_formats_info();
				title	= gettext_nop("Formats");
				break;
			case 2:
				str	= cff_ffmpeg_get_codecs_info();
				title	= gettext_nop("Codecs");
				break;
			default: /*case 3: */
				str	= cff_ffmpeg_get_bitstream_filters_info();
				title	= gettext_nop("Bitstream Filters");
				break;

		} /* switch (i) */

		cff_gui_ffmpeg_info_label(title, str, vbox);

		if (str)
		{
			/* malloc'ed */
			free(str);
		}
	}
	while (++i < 4);

	cff_gui_scrolled_info_window(gettext_nop("CutterFF: FFmpeg"),
				vbox, mbar, kPropIntHelpFFmpegWinWidth);

} /* static void cff_gui_help_ffmpeg(GtkWidget*, GtkWidget**) */

static void cff_gui_help_cutterff(GtkWidget *w, GtkWidget **mbar)
{
	static const gchar *cff_help_str[] =
	{
		gettext_nop("Navigation"),
		gettext_nop("Button '<' or Key 'Left' | Button '>' or Key 'Right'"),
		gettext_nop("Previous or next frame"),
		gettext_nop("Button '<<' or Key 'Down' | Button '>>' or Key 'Up'"),
		gettext_nop("Half second back or forward"),
		gettext_nop("Button '<<<' or Key 'Page Down' | Button '>>>' or Key 'Page Up'"),
		gettext_nop("Five seconds back or forward"),
		gettext_nop("Button '|<' or Key 'Home' | Button '>|' or Key 'End'"),
		gettext_nop("Previous or next cutpoint or start or end of video"),
		gettext_nop("With shift key"),
		gettext_nop("Button '<' or Key 'Left' | Button '>' or Key 'Right'"),
		gettext_nop("One minute back or forward"),
		gettext_nop("Button '<<' or Key 'Down' | Button '>>' or Key 'Up'"),
		gettext_nop("Five minutes back or forward"),
		gettext_nop("Button '<<<' or Key 'Page Down' | Button '>>>' or Key 'Page Up'"),
		gettext_nop("Ten minutes back or forward"),
		gettext_nop("Button '|<' or Key 'Home' | Button '>|' or Key 'End'"),
		gettext_nop("Start or end of video"),
		gettext_nop("A mouseclick on the position bar will jump to this postion."),
		gettext_nop("Note: the times above depends on the video and may differ!")
	};

	gint		i;
	gchar		*markup, **dst, *str[sizeof(cff_help_str) / sizeof(gchar*)];
	const gchar	**src;
	GtkWidget	*vbox, *label;

	vbox = GuiVBoxNew(FALSE, 2);

	label = gtk_label_new(NULL);
	GuiVBoxPackStart(vbox, label, TRUE, TRUE, 4);

	i = sizeof(cff_help_str) / sizeof(gchar*);

	src = cff_help_str;
	dst = str;

	do
	{
		*(dst++) = cff_charconv_utf8_gettext(*(src++));
	}
	while (--i > 0);

	markup = g_markup_printf_escaped(
		"<span size=\"xx-large\" weight=\"ultrabold\" stretch=\"ultraexpanded\"><i><u>%s</u></i></span>\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<u><b>%s</b></u>\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n\n"
		"%s",
		str[0],
		str[1],
		str[2],
		str[3],
		str[4],
		str[5],
		str[6],
		str[7],
		str[8],
		str[9],
		str[10],
		str[11],
		str[12],
		str[13],
		str[14],
		str[15],
		str[16],
		str[17],
		str[18],
		str[19]);

	gtk_label_set_markup(GTK_LABEL(label), markup);

	i = sizeof(cff_help_str) / sizeof(gchar*);

	src = cff_help_str;
	dst = str;

	do
	{
		gchar		*d;
		const gchar	*s;

		s = *(src++);
		d = *(dst++);
		CFF_CHARCONV_UTF8_FREE(d, s);
	}
	while (--i > 0);

	cff_gui_scrolled_info_window(gettext_nop("CutterFF: Help"),
				vbox, mbar, kPropIntHelpCutterFFWidth);
	
} /* static void cff_gui_help_cutterff(GtkWidget*, GtkWidget**) */

static void cff_gui_help_about(GtkWidget *w, GtkWidget **mbar)
{
	static const gchar cff_about_str_0[] =
			gettext_nop("A small programm for cuttting videos");
	static const gchar cff_about_str_1[] =
			gettext_nop("Distributed under GNU-GPL 2 or later");
	static const gchar cff_about_str_2[] = gettext_nop("Version");
	static const gchar cff_about_str_3[] = gettext_nop("Author:");

	/* main.c */
	extern const gchar *cutterff_version_str;

	GtkWidget	*window, *vbox, *hbox, *button, *frame, *label;
	gchar		*markup, *str[4];

	vbox	= GuiVBoxNew(FALSE, 2);
	window	= guitl_window_new_gettext(gettext_nop("CutterFF: About"), vbox, 4);
	guitl_dialog_set_modal(window, gWindata.mainwin, mbar);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

	label = gtk_label_new(NULL);
	frame = guitl_frame_new_add_child_gettext(NULL, label);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 5);

	str[0] = cff_charconv_utf8_gettext(cff_about_str_0);
	str[1] = cff_charconv_utf8_gettext(cff_about_str_1);
	str[2] = cff_charconv_utf8_gettext(cff_about_str_2);
	str[3] = cff_charconv_utf8_gettext(cff_about_str_3);

	markup = g_markup_printf_escaped(
		"<span size=\"xx-large\" weight=\"ultrabold\" stretch=\"ultraexpanded\"><i><u>%s</u></i></span>\n\n"
		"<tt>%s\n%s <b>%s</b> %s <b>%s</b></tt>\n\n"
		"%s\n\n"
		"%s %s\n\n"
		"%s %s %s\n"
		"%s",

		"CutterFF",
		str[0],
		"using", "FFmpeg", "and", "GTK+",
		str[1],
		str[2], cutterff_version_str,
		str[3],
		/* Harald F&ouml;rster */
		"Harald F" "\xC3\xB6" "rster",
		"<harald_foerster@users.sourceforge.net>",
		"URL: https://cutterff.sourceforge.io/");

	gtk_label_set_markup(GTK_LABEL(label), markup);
	g_free(markup);

	CFF_CHARCONV_UTF8_FREE(str[3], cff_about_str_3);
	CFF_CHARCONV_UTF8_FREE(str[2], cff_about_str_2);
	CFF_CHARCONV_UTF8_FREE(str[1], cff_about_str_1);
	CFF_CHARCONV_UTF8_FREE(str[0], cff_about_str_0);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	GuiMiscSetPadding(label, 15);

	hbox	= guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	button	= guitl_button_box_add_button(hbox,
					gettext_nop("OK"), NULL, NULL);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, window);

	gtk_widget_show_all(window);

} /* static void cff_gui_help_about(GtkWidget*, GtkWidget**) */


/*
	----------------
	 Menu Functions
	----------------
*/

void cff_gui_adjust_menu(GtkWidget *mni, gpointer data)
{
	guint		item;
	GtkWidget	**menu;
	cff_video_t	*video;
	gboolean	*ptr, menu_sensitive[kMenuItemsTotal],
			popup_sensitive[kPopupItemsTotal];

	menu = &cff_gui_menu_widget[kMenuItemsTotal];
	item = kMenuItemsTotal;

	video = data;

	if (video == NULL)
	{
		do
		{
			if (*(--menu) == mni)
			{
				break;
			}
		}
		while (--item > 0);

		if (item == 0)
		{
			return;
		}

		video = &gVideo;

	} /* if (video == NULL) */

	memset(menu_sensitive, '\0', sizeof(menu_sensitive));
	memset(popup_sensitive, '\0', sizeof(popup_sensitive));

	if (video->inFormatCtx != NULL)
	{
		menu_sensitive[kMenuFileClose] = TRUE;
		menu_sensitive[kMenuFileInfo]  = TRUE;
	}

	if (video->vframe.ic != NULL)
	{
		if (video->vframe.undo[0] != NULL)
		{
			menu_sensitive[kMenuEditUndo]   = TRUE;
			popup_sensitive[kPopupEditUndo] = TRUE;
		}

		if (video->vframe.redo[0] != NULL)
		{
			menu_sensitive[kMenuEditRedo]   = TRUE;
			popup_sensitive[kPopupEditRedo] = TRUE;
		}

		if (video->cut.list == NULL || video->cut.num != 1 ||
			video->cut.list[0].frame != video->vframe.dts_stop)
		{
			menu_sensitive[kMenuFileSave]  = TRUE;
			popup_sensitive[kPopupFileSave] = TRUE;
		}

		menu_sensitive[kMenuFileImage] = TRUE;

		menu_sensitive[kMenuEditLoad]  = TRUE;
		popup_sensitive[kPopupEditLoad] = TRUE;

		if (video->cut.num > 0)
		{
			menu_sensitive[kMenuEditSave]  = TRUE;
			menu_sensitive[kMenuEditClear] = TRUE;
			menu_sensitive[kMenuEditFrame] = TRUE;

			popup_sensitive[kPopupEditSave]  = TRUE;
			popup_sensitive[kPopupEditClear] = TRUE;
			popup_sensitive[kPopupEditFrame] = TRUE;

			if (cff_ffmpeg_video_cut_find_current(video) >= 0)
			{
				menu_sensitive[kMenuEditRemove] = TRUE;
			}
		}

		if (menu_sensitive[kMenuEditRemove] == FALSE)
		{
			menu_sensitive[kMenuEditAdd] = TRUE;
		}

	} /* if (video->ic != NULL) */

	if (video->inFormatCtx != NULL &&
			video->program.index != NULL &&
					video->program.num > 1)
	{
		menu_sensitive[kMenuEditProgram] = TRUE;
	}

	if (mni == cff_gui_popup_menu)
	{
		menu = &cff_gui_popup_widget[kPopupItemsTotal];
		ptr  = &popup_sensitive[kPopupItemsTotal];
		item = kPopupItemsTotal;
	}

	else
	{
		menu = &cff_gui_menu_widget[kMenuItemsTotal];
		ptr  = &menu_sensitive[kMenuItemsTotal];
		item = kMenuItemsTotal;
	}

	do
	{
		gtk_widget_set_sensitive(*(--menu), *(--ptr));
	}
	while (--item != 0);

} /* void cff_gui_adjust_menu(GtkWidget*, gpointer) */

static gboolean _cff_gui_destroy_submenu(GtkWidget *submenu)
{
	GList *children;

	while ((children = GuiMenuShellGetChildren(submenu)) != NULL)
	{
		gtk_widget_destroy(GTK_WIDGET(children->data));
	}

	return Gui_SOURCE_REMOVE;

} /* static gint _cff_gui_destroy_submenu(GtkWidget*) */

static void cff_gui_destroy_submenu(GtkWidget *submenu, gpointer data)
{
	/*
		Destroy a little bit later, else the
		'activate' - signal will not be emitted
	*/

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) _cff_gui_destroy_submenu, submenu, NULL);

} /* static void cff_gui_destroy_submenu(GtkWidget*, gpointer) */

static void cff_gui_build_frame_submenu(GtkWidget *submenu, gpointer data)
{
	unsigned int	i, current;
	cff_video_t	*video;

	video   = data;
	current = cff_ffmpeg_video_cut_find_current(video);

	i = 0;

	do
	{
		unsigned int	frame_number;
		GtkWidget	*mi;
		gchar		buf[64];

		frame_number = cff_ffmpeg_get_cut_frame_number(video, i);

		if (frame_number == (unsigned int) -1)
		{
			continue;
		}

		g_snprintf(buf, sizeof(buf), "%u", frame_number);

		mi = gtk_menu_item_new_with_label(buf);

		if (i == current)
		{
			gtk_widget_set_sensitive(mi, FALSE);
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), mi);
		gtk_widget_show(mi);

		GuiSignalConnect(mi, "activate",
			cff_gui_edit_frame_display, GINT_TO_POINTER(i));
	}
	while (++i < video->cut.num);

} /* static void cff_gui_build_frame_submenu(GtkWidget*, gpointer) */

static void cff_gui_build_program_submenu(GtkWidget *submenu, gpointer data)
{
	unsigned int	i;
	cff_video_t	*video;

	video = data;

	i = 0;

	do
	{
		int		prgnum;
		const char	*name;
		GtkWidget	*mi;
		gchar		buf[64];

		name = cff_ffmpeg_get_program_name(video, i, &prgnum);

		if (name != NULL)
		{
			if (tl_is_ascii_str(name) == FALSE)
			{
				if (strlen(name) >= sizeof(buf))
				{
					name = NULL;
				}

				else
				{
					strcpy(buf, name);
					name = tl_replace_non_ascii_chars_in_str(buf);
				}
			}
		}

		if (name == NULL)
		{
			if (prgnum < 0)
			{
				prgnum = i + 1;
			}

			g_snprintf(buf, sizeof(buf), "%d", prgnum);
			name = buf;
		}

		if (i == video->program.selected)
		{
			mi = gtk_check_menu_item_new_with_label(name);
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), TRUE);
		}

		else
		{
			mi = gtk_menu_item_new_with_label(name);
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), mi);
		gtk_widget_show(mi);

		GuiSignalConnect(mi, "activate",
			cff_gui_edit_program_open, GINT_TO_POINTER(i));
	}
	while (++i < video->program.num);

} /* static void cff_gui_build_program_submenu(GtkWidget*, gpointer) */

static void cff_gui_build_loglevel_submenu(GtkWidget *submenu, gpointer data)
{
	int i, current;

	if (gLogWindow.window == NULL ||
		(current = cff_ffmpeg_get_current_loglevel_index()) < 0)
	{
		current = 0;
	}

	i = 0;

	for (;;)
	{
		GtkWidget	*mi;
		const gchar	*logname;

		logname = cff_ffmpeg_get_loglevel_name(i);

		if (logname == NULL)
		{
			break;
		}

		mi = gtk_menu_item_new_with_label(logname);

		if (i == current)
		{
			gtk_widget_set_sensitive(mi, FALSE);
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), mi);
		gtk_widget_show(mi);

		GuiSignalConnect(mi, "activate",
			cff_gui_opt_logwindow_menu_selected, GINT_TO_POINTER(i));
		i++;

	} /* for (;;) */

} /* static void cff_gui_build_loglevel_submenu(GtkWidget*, gpointer) */

static gboolean cff_gui_mouse_button_pressed(GtkWidget *widget,
					GdkEvent *event, GtkWidget *popup_menu)
{
	guint		button;
	guint32		time;
	GdkEventButton	*bevent;

	bevent = (GdkEventButton *) event;

	if (bevent == NULL)
	{
		time	= GDK_CURRENT_TIME;
		button	= 0;
	}

	else
	{
		button = bevent->button;

		if (bevent->type != GDK_BUTTON_PRESS || button != 3)
		{
			return FALSE;
		}

		time = bevent->time;
	}

	gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, button, time);

	return TRUE;

} /* static gboolean cff_gui_mouse_button_pressed(GtkWidget*,
							GdkEvent*, GtkWidget*) */

static gboolean cff_gui_sf10_show_popup_menu(GtkWidget *widget, GtkWidget *popup_menu)
{
	return cff_gui_mouse_button_pressed(widget, NULL, popup_menu);

} /* static gboolean cff_gui_sf10_show_popup_menu(GtkWidget*, GtkWidget*) */

static void cff_gui_build_menu_item(const cff_guiMenuItem_t *item, GtkWidget *menu, GtkWidget *window)
{
	for (;;)
	{
		int		active, prop;
		GtkWidget	*mi, *submenu;

		switch (item->type)
		{
			case kMTypeItem:
			case kMTypeSubmenuCheckItem:
				mi = guitl_menu_item_new_gettext(item->label);

				if (item->type == kMTypeItem)
				{
					break;
				}

				submenu = gtk_menu_new();
				gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), submenu);
				gtk_widget_realize(submenu);

				if (item->callback)
				{
					GuiSignalConnect(submenu, "show",
						item->callback, item->data);
					GuiSignalConnect(submenu, "hide",
						cff_gui_destroy_submenu, NULL);
				}

				break;

			case kMTypeCheckItem:
				mi = guitl_check_menu_item_new_gettext(item->label);
				prop = GPOINTER_TO_INT(item->data);

				if (cff_gprop_get_bool(prop, &active) == FALSE)
				{
					if (prop == kPropBoolOptDelete ||
							prop == kPropBoolOptCursor ||
								prop == kPropBoolOptSaveConfig)
					{
						active = TRUE;
					}

					else
					{
						active = FALSE;
					}
				}

				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), active);
				break;

			case kMTypeSeparator:
				mi = GuiMenuSeparatorNew();
				break;

			default:
				return;
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);

		if (item->accel_id != kGAccelKeyNone)
		{
			cff_gaccel_bind_widget(item->accel_id, item->label, mi, window);
		}

		if (item->callback && item->type != kMTypeSubmenuCheckItem)
		{
			GuiSignalConnect(mi, "activate", item->callback, item->data);
		}

		if (item->storage)
		{
			*(item->storage) = mi;
		}

		item++;

	} /* for (;;) */

} /* static void cff_gui_build_menu_item(const cff_guiMenuItem_t*, GtkWidget*, GtkWidget*) */

static GtkWidget *cff_gui_build_menu(GtkWidget *window)
{
	static const struct build_menu_array
	{
		const gchar		*label;
		const cff_guiMenuItem_t	*entries;
		gboolean		adjust_menu;

	} build_menu_array[] = {

		{gettext_nop("File"), cff_gui_menuFile, TRUE},
		{gettext_nop("Edit"), cff_gui_menuEdit, TRUE},
		{gettext_nop("Options"), cff_gui_menuOptions, FALSE},
		{gettext_nop("Help"), cff_gui_menuHelp, FALSE},
		{NULL}
	};

	GtkWidget *mbar, *mbb, **mbbptr, *menu;

	const struct build_menu_array *build_tab;

	mbar = gtk_menu_bar_new();

	cff_gui_signal_connect_key_or_mouse_button(mbar, kButtonPressEvent);

#if GTK_MAJOR_VERSION == 2
	GuiMenuBarDisableShadow(mbar);
#endif

	build_tab = build_menu_array;
	mbbptr = cff_gui_menu_bar;

	do
	{
		mbb = guitl_menu_item_new_gettext(build_tab->label);
		GuiSignalConnect(mbb, "destroy", gtk_widget_destroyed, mbbptr);
		*(mbbptr++) = mbb;
		gtk_menu_shell_append(GTK_MENU_SHELL(mbar), mbb);

		if (build_tab->adjust_menu)
		{
			GuiSignalConnect(mbb, "activate",
					cff_gui_adjust_menu, &gVideo);
		}

		menu = gtk_menu_new();
		gtk_widget_realize(menu);

		gtk_menu_item_set_submenu(GTK_MENU_ITEM(mbb), menu);

		cff_gui_build_menu_item(build_tab->entries, menu, window);
	}
	while ((++build_tab)->label);


	/* Popup Menu */

	menu			= gtk_menu_new();
	cff_gui_popup_menu	= menu;

	GuiSignalConnect(menu, "show", cff_gui_adjust_menu, &gVideo);

	gtk_widget_realize(menu);

	cff_gui_build_menu_item(cff_gui_menuPopup, menu, NULL);

	gtk_widget_add_events(window, GDK_BUTTON_PRESS_MASK);

	GuiSignalConnect(window, "button-press-event",
				cff_gui_mouse_button_pressed, menu);
	GuiSignalConnect(window, "popup-menu",
				cff_gui_sf10_show_popup_menu, menu);

	return mbar;

} /* static GtkWidget *cff_gui_build_menu(GtkWidget*) */


/*
	------
	 INIT
	------
*/

static GtkWidget *cff_gui_build_drawable(GtkWidget *box, gint area)
{
	GtkWidget	*widget;
	GCallback	func;
#if GTK_MAJOR_VERSION > 2
	cairo_surface_t	**surface;
#endif

	widget = gtk_drawing_area_new();

	gtk_widget_set_size_request(widget, -1, 16);

	if (area == 0)
	{
		func = (GCallback) cff_gui_draw_bar;
	}
	else
	{
		func = (GCallback) cff_gui_draw_slider;
	}

	GuiVBoxPackStart(box, widget, FALSE, FALSE, 3);
	GuiSignalConnect(widget, cff_expose_event, func, &gVideo);

#if GTK_MAJOR_VERSION > 2

	gtk_widget_set_hexpand(widget, TRUE);

	if (area == 0)
	{
		surface = &gCairo.surfaceBar;
	}
	else
	{
		surface = &gCairo.surfaceSlider;
	}

	GuiSignalConnect(widget, cff_configure_event,
		cff_gui_drawing_area_configure_cb, surface);
#endif

	gtk_widget_add_events(widget, GDK_BUTTON_PRESS_MASK);
	GuiSignalConnect(widget, "button-press-event",
					cff_gui_draw_bar_button, &gVideo);
	return widget;

} /* static GtkWidget *cff_gui_build_drawable(GtkWidget*, gint) */

static void cff_gui_noop(void)
{
	return;

} /* static void cff_gui_noop(void) */


#if GTK_MAJOR_VERSION == 2

static GdkGC *cff_gui_set_gdk_color(const char *str, const GdkWindow *window)
{
	GdkGC    *context;
	GdkColor color;

	gdk_color_parse(str, &color);
	context	= gdk_gc_new((GdkDrawable *) window);
	gdk_gc_set_rgb_fg_color(context, &color);

	return context;

} /* static GdkGC *cff_gui_set_gdk__color(const char*, const GdkWindow*) */

#endif /* #if GTK_MAJOR_VERSION == 2 */


void cff_gui_init(const gchar *fname, const char *loglevel)
{
	gint		i;
	gchar		*str;
	GtkWidget	*menu_bar, *bbox, *dbox, *hbox, *vbox, *button, *frame, *widget, *window;

#ifndef BUILD_MINGW
	GdkPixbuf	*pixbuf;
#endif

	g_log_set_handler("Gdk", G_LOG_LEVEL_MASK, (GLogFunc) cff_gui_noop, NULL);

#if GTK_MAJOR_VERSION > 2
	cff_gui_css_load_from_data(
				"GtkMenuBar,"
				"menubar {"
				"  -GtkMenuBar-shadow-type: None;"
				"}");
#endif

	/* Main window */

	gWindata.topwin = NULL;

	vbox = GuiVBoxNew(FALSE, 0);

	window = guitl_window_new_gettext(NULL, vbox, 2);
	gWindata.mainwin = window;
	guitl_window_setup(window, kPropIntMainWinWidth);

	GuiSignalConnect(window, "delete-event",
				cff_gui_file_quit_delevnt, NULL);
	GuiSignalConnect(window, "destroy",
				gtk_main_quit, NULL);
	GuiSignalConnect(window, "screen-changed",
				cff_gui_mainwindow_cursor_new, &gCursor);

	cff_gui_signal_connect_key_or_mouse_button(window, kButtonPressEvent);
	cff_gui_signal_connect_key_or_mouse_button(window, kKeyPressEvent);

	gtk_widget_realize(window);

	cff_gui_mainwindow_cursor_new(window, NULL, &gCursor);

#ifndef BUILD_MINGW

	/* Application icon */

	pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) cutterff_xpm);
	gtk_window_set_icon(GTK_WINDOW(window), pixbuf);
#endif


	/* Menu and popup */

	menu_bar = cff_gui_build_menu(window);
	GuiVBoxPackStart(vbox, menu_bar, FALSE, FALSE, 0);


	/* Video picture */

	widget			= gtk_drawing_area_new();
	gDrawingAreaFrame	= widget;

	gtk_widget_set_size_request(widget, 400, 225);
	frame = guitl_frame_new_add_child_gettext(NULL, widget);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 5);
	GuiSignalConnect(widget, cff_expose_event,
				cff_gui_draw_frame, &gVideo);
	GuiSignalConnect(widget, cff_configure_event,
				cff_gui_draw_frame_configure_cb, &gVideo);

#if GTK_MAJOR_VERSION > 2
	GuiSignalConnect(widget, cff_configure_event,
		cff_gui_drawing_area_configure_cb, &gCairo.surfaceFrame);
#endif


	/* Info Bar */

	hbox = GuiHBoxNew(TRUE, 2);
	frame = guitl_frame_new_add_child_gettext(NULL, hbox);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 5);

	widget = gtk_label_new(NULL);
	gLabelFrame = GTK_LABEL(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	widget = gtk_label_new(NULL);
	gLabelTime = GTK_LABEL(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	cff_gui_frame_info(&gVideo);


	/* Slider Bar */

	dbox = GuiVBoxNew(TRUE, 2);
	frame = guitl_frame_new_add_child_gettext(NULL, dbox);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 3);

	gDrawingAreaBar    = cff_gui_build_drawable(dbox, 0);
	gDrawingAreaSlider = cff_gui_build_drawable(dbox, 1);


	/* Colors for drawables */

#if GTK_MAJOR_VERSION == 2
	gContext[kColorBlack]	= cff_gui_set_gdk_color("Black", window->window);
	gContext[kColorDarkGrey]= cff_gui_set_gdk_color("DarkGrey", window->window);
	gContext[kColorGrey]	= cff_gui_set_gdk_color("LightGrey", window->window);
	gContext[kColorGreen]	= cff_gui_set_gdk_color("Green", window->window);
	gContext[kColorRed]	= cff_gui_set_gdk_color("Red", window->window);
#else
	gdk_rgba_parse(&gCairo.RGBA[kColorBlack], "Black");
	gdk_rgba_parse(&gCairo.RGBA[kColorDarkGrey], "DarkGrey");
	gdk_rgba_parse(&gCairo.RGBA[kColorGrey], "LightGrey");
	gdk_rgba_parse(&gCairo.RGBA[kColorGreen], "Green");
	gdk_rgba_parse(&gCairo.RGBA[kColorRed], "Red");
#endif


	/* Button Bar */

	hbox = GuiHBoxNew(FALSE, 5);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 5);

	bbox = GuiHBoxNew(TRUE, 5);
	frame = guitl_frame_new_add_child_gettext(NULL, bbox);
	GuiHBoxPackStart(hbox, frame, FALSE, FALSE, 5);


	/* Goto frame */

	i = 0;

	do
	{
		button = gtk_button_new_with_label(gButtonVideoFrameLabel[i]);
		GuiHBoxPackStart(bbox, button, TRUE, TRUE, 1);
		GuiSignalConnect(button, "button-press-event",
					cff_gui_video_frame_button, GINT_TO_POINTER(i));
		gui_button_video_frame[i] = button;
	}
	while (++i < kButtonVideoFrameTotal);

	bbox = GuiHBoxNew(FALSE, 5);
	GuiHBoxPackStart(hbox, bbox, TRUE, TRUE, 5);


	/* Add or remove cutpoints */

	bbox = GuiHBoxNew(FALSE, 5);
	frame = guitl_frame_new_add_child_gettext(NULL, bbox);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 5);

	button = guitl_hbox_add_button(bbox, "+",
				(GCallback) cff_gui_edit_add, &gVideo);
	gui_button_video_cut[kButtonVideoCutAdd] = button;

	button = guitl_hbox_add_button(bbox, "-",
				(GCallback) cff_gui_edit_remove, &gVideo);
	gui_button_video_cut[kButtonVideoCutRemove] = button;


	/* Goto frame */

	GuiSignalConnect(window, "key-press-event",
				cff_gui_video_frame_key, &gVideo);


	/* Test filter extensions */

	cff_gui_file_filter_validate_prop(kPropStrFileVideoFilter);
	cff_gui_file_filter_validate_prop(kPropStrFileImageFilter);


	gtk_widget_show_all(window);


	/* Open the log window if level is not 'quiet' */

	if (loglevel == NULL)
	{
		gchar		*logname;
		const gchar	*logquiet;

		logquiet = cff_ffmpeg_get_loglevel_name(0);
		logname  = cff_gprop_get_str(kPropStrOptLoglevel);

		if (logname != NULL)
		{
			if (strcmp(logname, logquiet) != 0)
			{
				cff_gui_logwindow_open(logname);
			}

			g_free(logname);
		}

	} /* if (loglevel == NULL) */


	/* Open file */

	if (fname == NULL || fname[0] == '\0')
	{
		str = NULL;
	}

	else
	{
		gchar *utf8 = cff_charconv_utf8_from_locale(fname, NULL);

#ifdef BUILD_MINGW
		str = (utf8 == fname) ? g_strdup(utf8) : utf8;
#else
		str = cff_charconv_utf8_to_filename(utf8);

		if (str == utf8 || str == fname)
		{
			if (str == utf8 && str == fname)
			{
				str = g_strdup(fname);
			}
		}

		else
		{
			CFF_CHARCONV_UTF8_FREE(utf8, fname);
		}

#endif /* #ifdef BUILD_MINGW .. #else */

	} /* if (fname == NULL || fname[0] == '\0') .. else */

	if (str == NULL)
	{
		cff_gui_video_button_set_sensitive(FALSE);
		guitl_window_set_title_filename(window, NULL);
	}

	else
	{
		g_idle_add_full(G_PRIORITY_LOW,
			(GSourceFunc) cff_gui_video_open, str, NULL);
	}

} /* void cff_gui_init(const gchar*, const char*) */

/* EOF */
